#include <assert.h>
#include <stdio.h>
#include <math.h>
#include "fitting.h"

/*-----------------------------------------------------*/
/*      definitions for conditional compilation        */
#define __GOPI__
/*-----------------------------------------------------*/

#define N_TRACEPOINTS 100
#define DATATYPE int
#define EPSILON 1e-15

int global_checkDomain = 0;

#ifdef __DEBUG__
#define indent() {int i; for (i=0; i<indentLevel;i++) printf("   ");}
static int indentLevel = 0;
#endif

typedef struct _Node {
  DATATYPE data;
  struct _Node *next;
} Node;

void makeMonotonic(int n,             /* # of points */
		   FiveTuple p[],     /* 1st array of points */
		   FiveTuple q[],     /* 2nd array of points */
		   int *nBreak,       /* number of segments */
		   Node** breakPoints /* starting points of segments */)
{
  int i, pass=0;
  Node *newNode, *marked, *current, *nextCurrent;

  newNode = (Node*) malloc(sizeof(Node));
  newNode->next = 0;
  newNode->data = 0;
  *breakPoints = newNode;
  *nBreak = 1;
  
  if (n < 2) return;

  /*while (pass++ <= 7) { */
  while (pass++ <= 3) {
    current = *breakPoints;
    while (current) {
      int starti, endi;
      
      marked = current;
      nextCurrent = current->next;
      starti=current->data;
      if (current->next)
	endi=current->next->data-1;
      else
	endi=n;
      
      /* subdivide p[starti..endi] */
      for (i=starti+2; i<endi; i++) {
	double sign;

	switch (pass) {
	case 1: sign = (p[i].x - p[i-1].x) * (p[i-1].x - p[i-2].x); break;
	case 2: sign = (p[i].s - p[i-1].s) * (p[i-1].s - p[i-2].s); break;
	case 3: sign = (q[i].s - q[i-1].s) * (q[i-1].s - q[i-2].s); break;

/*
	case 2: sign = (p[i].y - p[i-1].y) * (p[i-1].y - p[i-2].y); break;
	case 3: sign = (p[i].z - p[i-1].z) * (p[i-1].z - p[i-2].z); break;
	case 4: sign = (p[i].s - p[i-1].s) * (p[i-1].s - p[i-2].s); break;
	case 5: sign = (p[i].t - p[i-1].t) * (p[i-1].t - p[i-2].t); break;
	case 6: sign = (q[i].s - q[i-1].s) * (q[i-1].s - q[i-2].s); break;
	case 7: sign = (q[i].t - q[i-1].t) * (q[i-1].t - q[i-2].t); break;
 */
	}
	if (sign < -EPSILON) {
	  newNode = (Node*) malloc(sizeof(Node));
	  newNode->next = nextCurrent;
	  newNode->data = i;
	  /* insert a new breakpoint */
	  marked->next = newNode;
	  marked = newNode;
	  (*nBreak)++;
	  i++;
	}
      }
      current = nextCurrent;
    } /* while (current) */
  } /* while (pass...) */
}

void freeList( Node* p )
{
  Node* next;

  while (p) {
    next = p->next;
    free(p);
    p = next;
  }
}

void printTuple( FiveTuple p )
{
  printf("%f %f %f %f %f\n", p.s, p.t, p.x, p.y, p.z);
}

void fprintTuple(FILE *fp, char *name, int index1, int index2, FiveTuple p )
{
  fprintf(fp, "%s_s(%d,%d)=%f;\n", name, index1, index2, p.s);
  fprintf(fp, "%s_t(%d,%d)=%f;\n", name, index1, index2, p.t);
  fprintf(fp, "%s_x(%d,%d)=%f;\n", name, index1, index2, p.x);
  fprintf(fp, "%s_y(%d,%d)=%f;\n", name, index1, index2, p.y);
  fprintf(fp, "%s_z(%d,%d)=%f;\n", name, index1, index2, p.z);
}

void scaleTuple( FiveTuple in, double s, FiveTuple* out )
{
  out->s = in.s * s;
  out->t = in.t * s;
  out->x = in.x * s;
  out->y = in.y * s;
  out->z = in.z * s;
}

void addTuples(FiveTuple in1, FiveTuple in2,
	       FiveTuple* out)
{
  out->s = in1.s + in2.s;
  out->t = in1.t + in2.t;
  out->x = in1.x + in2.x;
  out->y = in1.y + in2.y;
  out->z = in1.z + in2.z;
}

void subTuples(FiveTuple in1, FiveTuple in2,
	       FiveTuple* out)
{
  out->s = in1.s - in2.s;
  out->t = in1.t - in2.t;
  out->x = in1.x - in2.x;
  out->y = in1.y - in2.y;
  out->z = in1.z - in2.z;
}

void copyTuple(FiveTuple from, FiveTuple *to)
{
  to->s = from.s;
  to->t = from.t;
  to->x = from.x;
  to->y = from.y;
  to->z = from.z;
}

double distancePointToSegment(FiveTuple p, FiveTuple e1, FiveTuple e2)
{
  FiveTuple v1, v2;
  double dotProduct, v1Sqr, v2Sqr, distSqr;

  subTuples(p, e1, &v1);
  subTuples(e2, e1, &v2);
  if (global_checkDomain) {
      dotProduct = v1.s*v2.s + v1.t*v2.t;
      v1Sqr = v1.s*v1.s + v1.t*v1.t;
      v2Sqr = v2.s*v2.s + v2.t*v2.t;
  }
  else {
      dotProduct = v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
      v1Sqr = v1.x*v1.x + v1.y*v1.y + v1.z*v1.z;
      v2Sqr = v2.x*v2.x + v2.y*v2.y + v2.z*v2.z;
  }
  if (dotProduct <= 0) {
      return sqrt(v1Sqr);
  }
  if (dotProduct >= v2Sqr) {
      FiveTuple v3;
      double v3Sqr;
      
      subTuples(p, e2, &v3);
      if (global_checkDomain) {
	  v3Sqr = v3.s*v3.s + v3.t*v3.t;
      }
      else {
	  v3Sqr = v3.x*v3.x + v3.y*v3.y + v3.z*v3.z;
      }
      return sqrt(v3Sqr);
  }
  distSqr = v1Sqr - dotProduct*dotProduct/v2Sqr;
  if (distSqr < 0) {  /* caused by rounding error */
      assert(-distSqr < EPSILON);  /* make sure it's not outrageous */
      return sqrt(-distSqr);
  }
  return sqrt(distSqr);
}

static int nControl=0, nTrace=0;

void resetCount()
{
  nControl = nTrace = 0;
}

void outputLinear(FILE *fp, FiveTuple *p, int index)
{
  int i;
  FiveTuple tmp1, tmp2, tmp3;

  /* output the control points */
  fprintTuple(fp, "control", index, ++nControl, p[0]);
  fprintTuple(fp, "control", index, ++nControl, p[1]);

  /* output large number of sample on the trace */
  for (i=1; i<N_TRACEPOINTS; i++) {
    double t = (double) i/N_TRACEPOINTS;
    scaleTuple(p[0], 1.0-t, &tmp1);
    scaleTuple(p[1], t, &tmp2);
    addTuples(tmp1, tmp2, &tmp3);
    fprintTuple(fp, "trace", index, ++nTrace, tmp3);
  }
}

void outputQuadratic(FILE *fp, FiveTuple *p, int index)
{
  int i;
  FiveTuple tmp1, tmp2, tmp3;

  /* output the control points */
  fprintTuple(fp, "control", index, ++nControl, p[0]);
  fprintTuple(fp, "control", index, ++nControl, p[1]);
  fprintTuple(fp, "control", index, ++nControl, p[2]);

  /* output large number of sample on the trace */
  for (i=1; i<N_TRACEPOINTS; i++) {
    double t = (double) i/N_TRACEPOINTS;
    scaleTuple(p[0], (1.0-t)*(1.0-t), &tmp1);
    scaleTuple(p[1], t*(1.0-t)*2.0, &tmp2);
    addTuples(tmp1, tmp2, &tmp3);
    scaleTuple(p[2], t*t, &tmp1);
    addTuples(tmp3, tmp1, &tmp2);
    fprintTuple(fp, "trace", index, ++nTrace, tmp2);
  }
}

void outputCubic(FILE *fp, FiveTuple *p, int index)
{
  int i;
  FiveTuple tmp1, tmp2, tmp3;

  /* output the control points */
  fprintTuple(fp, "control", index, ++nControl, p[0]);
  fprintTuple(fp, "control", index, ++nControl, p[1]);
  fprintTuple(fp, "control", index, ++nControl, p[2]);
  fprintTuple(fp, "control", index, ++nControl, p[3]);

  /* output large number of sample on the trace */
  for (i=1; i<N_TRACEPOINTS; i++) {
    double t = (double) i/N_TRACEPOINTS;
    scaleTuple(p[0], (1.0-t)*(1.0-t)*(1.0-t), &tmp1);
    scaleTuple(p[1], t*(1.0-t)*(1.0-t)*3.0, &tmp2);
    addTuples(tmp1, tmp2, &tmp3);
    scaleTuple(p[2], t*t*(1.0-t)*3.0, &tmp1);
    addTuples(tmp1, tmp3, &tmp2);
    scaleTuple(p[3], t*t*t, &tmp1);
    addTuples(tmp1, tmp2, &tmp3);
    fprintTuple(fp, "trace", index, ++nTrace, tmp3);
  }
}

void outputCurve(FILE *fp, ParamCurve curve, int index)
{
  fprintf(fp, "degree(%d)=%d;\n", index, curve.degree);
  switch (curve.degree) {
  case 1: outputLinear(fp, curve.control, index); break;
  case 2: outputQuadratic(fp, curve.control, index); break;
  case 3: outputCubic(fp, curve.control, index); break;
  }
}

void evalLinear(FiveTuple p[], double t, FiveTuple* result)
{
  FiveTuple tmp1, tmp2;

  scaleTuple(p[0], 1.0-t, &tmp1);
  scaleTuple(p[1], t, &tmp2);
  addTuples(tmp1, tmp2, result);
}

double distanceLinear(FiveTuple ref, FiveTuple p[])
{
  return distancePointToSegment(ref, p[0], p[1]);
}

void evalQuadratic(FiveTuple p[], double t, FiveTuple* result)
{
  FiveTuple tmp1, tmp2, tmp3;

  scaleTuple(p[0], (1.0-t)*(1.0-t), &tmp1);
  scaleTuple(p[1], t*(1.0-t)*2.0, &tmp2);
  addTuples(tmp1, tmp2, &tmp3);
  scaleTuple(p[2], t*t, &tmp1);
  addTuples(tmp3, tmp1, result);
}

#ifdef __GOPI__
extern double distanceQuadratic(FiveTuple ref, FiveTuple p[]);
#else
double distanceQuadratic(FiveTuple ref, FiveTuple p[])
{
  double dist, minDist;
  char initialized=0;
  int i;
  FiveTuple sample1, sample2;

  copyTuple(p[0], &sample1);
  for (i=1; i<=N_TRACEPOINTS; i++) {
    double t = (double) i/N_TRACEPOINTS;

    evalQuadratic(p, t, &sample2);
    dist = distancePointToSegment(ref, sample1, sample2);
    copyTuple(sample2, &sample1);
    if (! initialized) { minDist = dist; initialized = 1; }
    if (dist < minDist) minDist = dist;
  }
  return minDist;
}
#endif

void evalCubic(FiveTuple p[], double t, FiveTuple* result)
{
  FiveTuple tmp1, tmp2, tmp3;

  scaleTuple(p[0], (1.0-t)*(1.0-t)*(1.0-t), &tmp1);
  scaleTuple(p[1], t*(1.0-t)*(1.0-t)*3.0, &tmp2);
  addTuples(tmp1, tmp2, &tmp3);
  scaleTuple(p[2], t*t*(1.0-t)*3.0, &tmp1);
  addTuples(tmp1, tmp3, &tmp2);
  scaleTuple(p[3], t*t*t, &tmp1);
  addTuples(tmp1, tmp2, result);
}

#ifdef __GOPI__
extern double distanceCubic(FiveTuple ref, FiveTuple p[]);
#else
double distanceCubic(FiveTuple ref, FiveTuple p[])
{
  double dist, minDist;
  char initialized=0;
  int i;
  FiveTuple sample1, sample2;

  copyTuple(p[0], &sample1);
  for (i=1; i<=N_TRACEPOINTS; i++) {
    double t = (double) i/N_TRACEPOINTS;

    evalCubic(p, t, &sample2);
    dist = distancePointToSegment(ref, sample1, sample2);
    copyTuple(sample2, &sample1);
    if (! initialized) { minDist = dist; initialized = 1; }
    if (dist < minDist) minDist = dist;
  }
  return minDist;
}
#endif

double distanceToCurve(FiveTuple ref, ParamCurve curve)
{
  switch (curve.degree) {
  case 1: return distanceLinear(ref, curve.control); break;
  case 2: return distanceQuadratic(ref, curve.control); break;
  case 3: return distanceCubic(ref, curve.control); break;
  }
}

double fittingErrorEstimate( int count, FiveTuple *p, ParamCurve curve )
{
  double dist, maxDist=0.0;
  int i;

  for (i=0; i<count; i++) {
    dist = distanceToCurve(p[i], curve);
    if (dist > maxDist)
      maxDist = dist;
  }
  return maxDist;
}

void createLinear(ParamCurve *curve,
		  FiveTuple p1, FiveTuple p2)
{
  curve->degree = 1;
  curve->control = (FiveTuple*) malloc(sizeof(FiveTuple)*2);
  copyTuple(p1, &curve->control[0]);
  copyTuple(p2, &curve->control[1]);
}

void createQuadratic(ParamCurve *curve,
		     FiveTuple p1, FiveTuple p2, FiveTuple p3)
{
  curve->degree = 2;
  curve->control = (FiveTuple*) malloc(sizeof(FiveTuple)*3);
  copyTuple(p1, &curve->control[0]);
  copyTuple(p2, &curve->control[1]);
  copyTuple(p3, &curve->control[2]);
}

void createCubic(ParamCurve *curve,
		 FiveTuple p1, FiveTuple p2, FiveTuple p3, FiveTuple p4)
{
  curve->degree = 3;
  curve->control = (FiveTuple*) malloc(sizeof(FiveTuple)*4);
  copyTuple(p1, &curve->control[0]);
  copyTuple(p2, &curve->control[1]);
  copyTuple(p3, &curve->control[2]);
  copyTuple(p4, &curve->control[3]);
}

void swapCurve(ParamCurve *a, ParamCurve *b)
{
  ParamCurve tmp;

  tmp.degree = a->degree;
  tmp.control = a->control;
  a->degree = b->degree;
  a->control = b->control;
  b->degree = tmp.degree;
  b->control = tmp.control;
}

/*----------------------------------------------------------------------*/
/* Parameters for fitting():
   Input (pass by value)
       count: number of points in piecewise linear approximation
       p[]  : points in first patch
       q[]  : points in the neighboring patch
       simpleSplit: use 0 for monotonic subdivision
                        1 for bisection (to improve error bound)
		       -1 for no initial subdivision at all
       errorBound: fitting error must be less than this value
       domainError: if 1, check the fitting error in domain value
                    if 0, check in the (x,y,z) 3-space.

   Output (passed by address)
       curveCount: number of curves in Bezier fitting
       pCurves[] : curves description of the 1st patch
       qCurves[] : curves desfription of the 2nd patch                  */
/*----------------------------------------------------------------------*/
void fitting(int count, FiveTuple *p, FiveTuple *q,
	     int *curveCount, ParamCurve **pCurves, ParamCurve **qCurves,
	     char simpleSplit, double errorBound, int checkDomain)
{
  int nBreak, i, tmpCount;
  Node *breakPoints, *current;
  double cubicStep = 1.0/3.0, quadStep = 0.5;
  double diffSum;
  double TOLERANCE = 0.001;
  char qReversed = 0;
  FiveTuple *qLocal;
  ParamCurve *tmpPCurves, *tmpQCurves;

  /* This will influence how distancePointToSegment()
     works while checking the fitting error. */
  global_checkDomain = checkDomain;

  /*****************************************
    Check if p[] == q[] or in reverse order.
    If not, *curveCount = 0 and return
   *****************************************/
  diffSum = 0.0;
  for (i=0; i<count; i++) {
    diffSum += fabs(p[i].x - q[i].x);
    diffSum += fabs(p[i].y - q[i].y);
    diffSum += fabs(p[i].z - q[i].z);
  }
  if (diffSum > TOLERANCE) {
    qReversed = 1;
    diffSum = 0.0;
    for (i=0; i<count; i++) {
      int j = count-i-1;
      diffSum += fabs(p[i].x - q[j].x);
      diffSum += fabs(p[i].y - q[j].y);
      diffSum += fabs(p[i].z - q[j].z);
    }
    if (diffSum > TOLERANCE) {
      *curveCount = 0;
      *pCurves = *qCurves = 0;
      return;
    }
  }

  if (qReversed) {
    /* copy q[] into qLocal[] in reverse order */
    int i, j, len=sizeof(FiveTuple)*count;
    qLocal = (FiveTuple*) malloc(len);
    for (i=0, j=count-1; i<count; i++, j--) {
      copyTuple( q[i], &qLocal[j]);
    }
  }
  else {
    qLocal=q; /* if not in reverse order, then qLocal is just an alias of q */
  }

  switch (simpleSplit) {
    Node *mid;
  case 0:
    makeMonotonic( count, p, qLocal, &nBreak, &breakPoints );
    break;
  case 1:
    breakPoints = (Node*) malloc(sizeof(Node));
    breakPoints->data = 0;
    if (count < 5) {  /* perfect fit can be done, no split necessary */
      breakPoints->next = 0;
      nBreak = 1;
    }
    else {
      mid = (Node*) malloc(sizeof(Node));
      breakPoints->next = mid;
      mid->next = 0;
      mid->data = count/2;
      nBreak = 2;
    }
    break;
  default:
    breakPoints = (Node*) malloc(sizeof(Node));
    breakPoints->data = 0;
    breakPoints->next = 0;
    nBreak = 1;
    break;
  }

#ifdef __DEBUG__
  indent();
  printf("# of segments = %d\n", nBreak);
#endif

  /* Allocate memory for pCurves[] and qCurves[].
   * Right now, the maximal possible space is reserved.
   * The extra space will be free-ed at the end of subroutine
   */
  *pCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*count);
  *qCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*count);
  *curveCount = 0;

  current = breakPoints;
  while (current) {
    int nPoint, degree, prevDegree;
    int i0, i1, i2, i3;
    FiveTuple deriv, derivQ, prevDeriv, prevDerivQ;

    i0 = current->data;
    if (current->next)
      i3 = current->next->data - 1;
    else
      i3 = count-1;
    nPoint = i3 - i0 + 1;

    degree = nPoint - 1;
    if (degree>3) degree = 3;

    if (0==degree && i0>0) {
      FiveTuple tmp1, tmp2;

      scaleTuple( prevDeriv, quadStep/prevDegree, &tmp2 );
      addTuples( p[i0-1], tmp2, &tmp1 );
#ifdef __DEBUG__
      indent();
      printf("bridge point %d to %d with degree 2 curve\n", i0-1, i0);
      indent(); printTuple( p[i0-1] );
      indent(); printTuple( tmp1 );
      indent(); printTuple( p[i0] );
#endif
      createQuadratic( &(*pCurves)[*curveCount], p[i0-1], tmp1, p[i0] );

      scaleTuple( prevDerivQ, quadStep/prevDegree, &tmp2 );
      addTuples( qLocal[i0-1], tmp2, &tmp1 );
      createQuadratic(&(*qCurves)[*curveCount],
		      qLocal[i0-1], tmp1, qLocal[i0] );

      (*curveCount)++;

      /* degree 0 can happen at the last point only,
	 so prevSlopes calculation is not necessary */
    }
    else if (1==degree) {
      FiveTuple tmp1, tmp2, tmp3;

      subTuples( p[i3], p[i0], &deriv );
      subTuples( qLocal[i3], qLocal[i0], &derivQ );
      
      if (i0 > 0) { /* connect p[i0-1] and p[i0] */
	scaleTuple( prevDeriv, cubicStep/prevDegree, &tmp3 );
	addTuples( p[i0-1], tmp3, &tmp1 );
	scaleTuple( deriv, -cubicStep/degree, &tmp3 );
	addTuples( p[i0], tmp3, &tmp2 );
#ifdef __DEBUG__
	indent();
	printf("bridge point %d to %d with degree 3 curve\n", i0-1, i0);
	indent(); printTuple( p[i0-1] );
	indent(); printTuple( tmp1 );
	indent(); printTuple( tmp2 );
	indent(); printTuple( p[i0] );
#endif
	createCubic(&(*pCurves)[*curveCount], p[i0-1], tmp1, tmp2, p[i0]);

	scaleTuple( prevDerivQ, cubicStep/prevDegree, &tmp3 );
	addTuples( qLocal[i0-1], tmp3, &tmp1 );
	scaleTuple( derivQ, -cubicStep/degree, &tmp3 );
	addTuples( qLocal[i0], tmp3, &tmp2 );
	createCubic(&(*qCurves)[*curveCount],
		    qLocal[i0-1], tmp1, tmp2, qLocal[i0]);

	(*curveCount)++;
      }

#ifdef __DEBUG__
      indent();
      printf("approximate point %d to %d with degree %d curve\n",
	     i0, i3, degree);
      indent(); printTuple( p[i0] );
      indent(); printTuple( p[i3] );
#endif
      createLinear( &(*pCurves)[*curveCount], p[i0], p[i3] );
      createLinear( &(*qCurves)[*curveCount], qLocal[i0], qLocal[i3] );
      (*curveCount)++;

      copyTuple( deriv, &prevDeriv );
      copyTuple( derivQ, &prevDerivQ );
    } /* degree 1 */
    else if (2==degree) {
      FiveTuple tmp1, tmpQ1, tmp2, tmp3, tmp4;

      i1 = i0+1;

      /* b1 = 2*p1 - 0.5*p0 - 0.5*p2 */
      scaleTuple( p[i1], 2.0, &tmp1 );
      scaleTuple( p[i0], -0.5, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( p[i3], -0.5, &tmp2 );
      addTuples( tmp3, tmp2, &tmp1 );

      /* repeat for q[] */
      scaleTuple( qLocal[i1], 2.0, &tmpQ1 );
      scaleTuple( qLocal[i0], -0.5, &tmp2 );
      addTuples( tmpQ1, tmp2, &tmp3 );
      scaleTuple( qLocal[i3], -0.5, &tmp2 );
      addTuples( tmp3, tmp2, &tmpQ1 );

      subTuples( tmp1, p[i0], &deriv );
      subTuples( tmpQ1, qLocal[i0], &derivQ );
      
      if (i0 > 0) { /* connect p[i0-1] and p[i0] */
	scaleTuple( prevDeriv, cubicStep/prevDegree, &tmp4 );
	addTuples( p[i0-1], tmp4, &tmp2 );
	scaleTuple( deriv, -cubicStep/degree, &tmp4 );
	addTuples( p[i0], tmp4, &tmp3 );
#ifdef __DEBUG__
	indent();
	printf("bridge point %d to %d with degree 3 curve\n", i0-1, i0);
	indent(); printTuple( p[i0-1] );
	indent(); printTuple( tmp2 );
	indent(); printTuple( tmp3 );
	indent(); printTuple( p[i0] );
#endif
	createCubic(&(*pCurves)[*curveCount], p[i0-1], tmp2, tmp3, p[i0]);

	scaleTuple( prevDerivQ, cubicStep/prevDegree, &tmp4 );
	addTuples( qLocal[i0-1], tmp4, &tmp2 );
	scaleTuple( derivQ, -cubicStep/degree, &tmp4 );
	addTuples( qLocal[i0], tmp4, &tmp3 );
	createCubic(&(*qCurves)[*curveCount],
		    qLocal[i0-1], tmp2, tmp3, qLocal[i0]);

	(*curveCount)++;
      }

#ifdef __DEBUG__
      indent();
      printf("approximate point %d to %d with degree %d curve\n",
	     i0, i3, degree);
      indent(); printTuple( p[i0] );
      indent(); printTuple( tmp1 );
      indent(); printTuple( p[i3] );
#endif
      createQuadratic( &(*pCurves)[*curveCount], p[i0], tmp1, p[i3] );
      createQuadratic( &(*qCurves)[*curveCount],
		       qLocal[i0], tmpQ1, qLocal[i3] );
      (*curveCount)++;

      subTuples( p[i3], tmp1, &prevDeriv );
      subTuples( qLocal[i3], tmpQ1, &prevDerivQ );
    } /* degree 2 */
    else {
      FiveTuple y1, y2, tmp1, tmp2, tmp3;
      FiveTuple pCtrl1, pCtrl2, qCtrl1, qCtrl2; /* control points */
      int s, r, n;
      double a11, a12, a21, a22, det;
      
      n = nPoint-1;
      s = n / 3;
      r = n - s;
      a11 = a22 = 3*s*r*r;
      a12 = a21 = 3*s*s*r;
      det = (double) a11*a22 - a12*a21;
      i1 = i0 + s;
      i2 = i0 + r;

      /* y1 = p1*n^3 - p0*r^3 - p3*s^3 */
      scaleTuple( p[i1], (double) n*n*n, &tmp1 );
      scaleTuple( p[i0], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( p[i3], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y1 );
      /* y2 = p2*n^3 - p0*s^3 - p3*r^3 */
      scaleTuple( p[i2], (double) n*n*n, &tmp1 );
      scaleTuple( p[i3], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( p[i0], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y2 );

      /* pCtrl1 = (a22*y1 - a12*y2)/det */
      scaleTuple( y1, a22, &tmp1 );
      scaleTuple( y2, -a12, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &pCtrl1 );
      /* pCtrl2 = (-a21*y1 + a11*y2)/det */
      scaleTuple( y1, -a21, &tmp1 );
      scaleTuple( y2, a11, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &pCtrl2 );

      /*---------- repeat for q[] ----------*/
      /* y1 = q1*n^3 - q0*r^3 - q3*s^3 */
      scaleTuple( qLocal[i1], (double) n*n*n, &tmp1 );
      scaleTuple( qLocal[i0], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( qLocal[i3], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y1 );
      /* y2 = q2*n^3 - q0*s^3 - q3*r^3 */
      scaleTuple( qLocal[i2], (double) n*n*n, &tmp1 );
      scaleTuple( qLocal[i3], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( qLocal[i0], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y2 );

      /* qCtrl1 = (a22*y1 - a12*y2)/det */
      scaleTuple( y1, a22, &tmp1 );
      scaleTuple( y2, -a12, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &qCtrl1 );
      /* qCtrl2 = (-a21*y1 + a11*y2)/det */
      scaleTuple( y1, -a21, &tmp1 );
      scaleTuple( y2, a11, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &qCtrl2 );

#ifdef __DEBUG__
      indent();
      printf("approximate point %d to %d with degree %d curve\n",
	     i0, i3, degree);
      indent(); printTuple( p[i0] );
      indent(); printTuple( pCtrl1 );
      indent(); printTuple( pCtrl2 );
      indent(); printTuple( p[i3] );
#endif

      /*
       * Now, we have the control points of the cubic curves for p[] and q[].
       * But, instead of putting them in (*pCurves)[] and (*qCurves)[],
       * we will save them in tmpPCurves[] and tmpQCurves[] first.
       * This is because the "bridging curves" (the curve to bridge the
       * last point of the previous partition) should be stored first.
       *
       * However, we can not compute and store the "bridge curve" 
       * before we compute the current curve, because the slope 
       * required for the bridging curves depends on the first control
       * point of current curve, which may not be pCtrl1 or qCtrl1 if the
       * curve fitting is not good enough and requires further partitioning
       */
      tmpPCurves = (ParamCurve*) malloc(sizeof(ParamCurve));
      tmpQCurves = (ParamCurve*) malloc(sizeof(ParamCurve));
      tmpCount = 1;
      createCubic( &tmpPCurves[0],
		   p[i0], pCtrl1, pCtrl2, p[i3] );
      createCubic( &tmpQCurves[0],
		   qLocal[i0], qCtrl1, qCtrl2, qLocal[i3] );

      /********************************************************************
	ERROR BOUND CHECKING !!!
      ********************************************************************/
#ifdef __DEBUG__
      indent();
      printf("error bound for P[] = %.15f for Q[] = %.15f\n",
	     fittingErrorEstimate(nPoint, &p[i0], tmpPCurves[0]),
	     fittingErrorEstimate(nPoint, &qLocal[i0], tmpQCurves[0]));
#endif
      if (nPoint > 4 &&
	(fittingErrorEstimate(nPoint, &p[i0], tmpPCurves[0])>errorBound ||
	 fittingErrorEstimate(nPoint, &qLocal[i0], tmpQCurves[0])>errorBound))
      {
#ifdef __DEBUG__
	indent();
	printf("fitting error is out of bound, calling recursively\n");
	indent();
	printf("point 0 will be correspondent to current point %d\n", i0);
	indent();
	printf("---beginning of nesting level %d output---------------\n",
	       ++indentLevel);
#endif
	free(tmpPCurves);
	free(tmpQCurves);
	switch (simpleSplit) {
	  case 0:
	  case 1:
	    fitting(nPoint, &p[i0], &qLocal[i0],
		    &tmpCount, &tmpPCurves, &tmpQCurves,
		    1, errorBound, checkDomain);
	    break;
	  case -1:
	  default:
	    fitting(nPoint, &p[i0], &qLocal[i0],
		    &tmpCount, &tmpPCurves, &tmpQCurves,
		    0, errorBound, checkDomain);
	    break;
	}
#ifdef __DEBUG__
	indentLevel--;
	indent();
	printf("---end of nesting level %d message--------------------\n",
	       indentLevel+1);
#endif
      }

      subTuples( tmpPCurves[0].control[0], p[i0], &deriv );
      subTuples( tmpQCurves[0].control[0], qLocal[i0], &derivQ );

      if (i0 > 0) { /* connect p[i0-1] and p[i0] */
	scaleTuple( prevDeriv, cubicStep/prevDegree, &tmp3 );
	addTuples( p[i0-1], tmp3, &tmp1 );
	scaleTuple( deriv, -cubicStep/degree, &tmp3 );
	addTuples( p[i0], tmp3, &tmp2 );
	
#ifdef __DEBUG__
	indent();
	printf("bridge point %d to %d with degree 3 curve\n", i0-1, i0);
	indent(); printTuple( p[i0-1] );
	indent(); printTuple( tmp1 );
	indent(); printTuple( tmp2 );
	indent(); printTuple( p[i0] );
#endif
	createCubic(&(*pCurves)[*curveCount], p[i0-1], tmp1, tmp2, p[i0]);

	/* repeat for q[] */
	scaleTuple( prevDerivQ, cubicStep/prevDegree, &tmp3 );
	addTuples( qLocal[i0-1], tmp3, &tmp1 );
	scaleTuple( derivQ, -cubicStep/degree, &tmp3 );
	addTuples( qLocal[i0], tmp3, &tmp2 );

	createCubic(&(*qCurves)[*curveCount],
		    qLocal[i0-1], tmp1, tmp2, qLocal[i0]);
	
	(*curveCount)++;
      }
      
      /* copying tmpPCurves[] to (*pCurves)[], tmpQCurves[] to (*qCurves)[]
         and free the memory */
      for (i=0; i<tmpCount; i++) {
	(*pCurves)[*curveCount].degree = tmpPCurves[i].degree;
	(*pCurves)[*curveCount].control = tmpPCurves[i].control;
	(*qCurves)[*curveCount].degree = tmpQCurves[i].degree;
	(*qCurves)[*curveCount].control = tmpQCurves[i].control;
	(*curveCount)++;
      }

      /* Now, the "bridging" curves are done.
	 Create the slopes for the next "bridging" */
      i = tmpPCurves[tmpCount-1].degree;
      subTuples( p[i3], tmpPCurves[tmpCount-1].control[i], &prevDeriv );
      i = tmpQCurves[tmpCount-1].degree;
      subTuples( qLocal[i3], tmpQCurves[tmpCount-1].control[i], &prevDerivQ );

      free(tmpPCurves);
      free(tmpQCurves);
    } /* degree 3 */
      
    prevDegree = degree;
    current = current->next;
  }

  if (qReversed) {
    assert( q != qLocal);  /* make sure we don't free q[] accidentally */
    free( qLocal );
  }

  /* pack the memory, so the sizes of (*pCurves)[] and (*qCurves)[] 
     == *curveCount */

  tmpPCurves = *pCurves;
  tmpQCurves = *qCurves;
  *pCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*(*curveCount));
  *qCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*(*curveCount));
  for (i=0; i<*curveCount; i++) {
    int j = qReversed ? *curveCount-i-1 : i;
    (*pCurves)[i].degree = tmpPCurves[i].degree;
    (*pCurves)[i].control = tmpPCurves[i].control;
    (*qCurves)[j].degree = tmpQCurves[i].degree;
    (*qCurves)[j].control = tmpQCurves[i].control;
  }
  free(tmpPCurves);
  free(tmpQCurves);
  freeList(breakPoints);
}

void spec_ratio( FiveTuple *p, FiveTuple *q, FiveTuple *ratio, int flag)
{
   ratio->s = p->s/q->s;
  if(flag) ratio->x = p->x/q->x;

}
void spec_length( FiveTuple *p, FiveTuple *len, int flag)
{

  len->s = sqrt(p->s*p->s + p->t*p->t); 
  if(flag) len->x = sqrt(p->x*p->x + p->y*p->y + p->z*p->z); 
   
return;
}


/*----------------------------------------------------------------------*/
/* Parameters for fitting():
   Input (pass by value)
       count: number of points in piecewise linear approximation
       p[]  : points in first patch
       q[]  : points in the neighboring patch
       simpleSplit: use 0 for monotonic subdivision
                        1 for bisection (to improve error bound)
		       -1 for no initial subdivision at all
       errorBound: fitting error must be less than this value
       domainError: if 1, check the fitting error in domain value
                    if 0, check in the (x,y,z) 3-space.

   Output (passed by address)
       curveCount: number of curves in Bezier fitting
       pCurves[] : curves description of the 1st patch
       qCurves[] : curves desfription of the 2nd patch                  */
/*----------------------------------------------------------------------*/
void spec_fitting(int count, FiveTuple *p, FiveTuple *q,
	     int *curveCount, ParamCurve **pCurves, ParamCurve **qCurves,
	     char simpleSplit, double errorBound, int checkDomain)
{
  int nBreak, i, tmpCount;
  Node *breakPoints, *current;
  double cubicStep = 1.0/3.0, quadStep = 0.5;
  double diffSum;
  double TOLERANCE = 0.001;
  char qReversed = 0;
  FiveTuple *qLocal;
  ParamCurve *tmpPCurves, *tmpQCurves;

  /* This will influence how distancePointToSegment()
     works while checking the fitting error. */
  global_checkDomain = checkDomain;

  /*****************************************
    Check if p[] == q[] or in reverse order.
    If not, *curveCount = 0 and return
   *****************************************/
  diffSum = 0.0;
  for (i=0; i<count; i++) {
    diffSum += fabs(p[i].x - q[i].x);
    diffSum += fabs(p[i].y - q[i].y);
    diffSum += fabs(p[i].z - q[i].z);
  }
  if (diffSum > TOLERANCE) {
    qReversed = 1;
    diffSum = 0.0;
    for (i=0; i<count; i++) {
      int j = count-i-1;
      diffSum += fabs(p[i].x - q[j].x);
      diffSum += fabs(p[i].y - q[j].y);
      diffSum += fabs(p[i].z - q[j].z);
    }
    if (diffSum > TOLERANCE) {
      *curveCount = 0;
      *pCurves = *qCurves = 0;
      return;
    }
  }

  if (qReversed) {
    /* copy q[] into qLocal[] in reverse order */
    int i, j, len=sizeof(FiveTuple)*count;
    qLocal = (FiveTuple*) malloc(len);
    for (i=0, j=count-1; i<count; i++, j--) {
      copyTuple( q[i], &qLocal[j]);
    }
  }
  else {
    qLocal=q; /* if not in reverse order, then qLocal is just an alias of q */
  }

  switch (simpleSplit) {
    Node *mid;
  case 0:
    makeMonotonic( count, p, qLocal, &nBreak, &breakPoints );
    break;
  case 1:
    breakPoints = (Node*) malloc(sizeof(Node));
    breakPoints->data = 0;
    if (count < 5) {  /* perfect fit can be done, no split necessary */
      breakPoints->next = 0;
      nBreak = 1;
    }
    else {
      mid = (Node*) malloc(sizeof(Node));
      breakPoints->next = mid;
      mid->next = 0;
      mid->data = count/2;
      nBreak = 2;
    }
    break;
  default:
    breakPoints = (Node*) malloc(sizeof(Node));
    breakPoints->data = 0;
    breakPoints->next = 0;
    nBreak = 1;
    break;
  }

#ifdef __DEBUG__
  indent();
  printf("# of segments = %d\n", nBreak);
#endif

  /* Allocate memory for pCurves[] and qCurves[].
   * Right now, the maximal possible space is reserved.
   * The extra space will be free-ed at the end of subroutine
   */
  *pCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*count);
  *qCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*count);
  *curveCount = 0;

  current = breakPoints;
  while (current) {
    int nPoint, degree, prevDegree;
    int i0, i1, i2, i3;
    FiveTuple deriv, derivQ, prevDeriv, prevDerivQ;

    i0 = current->data;
    if (current->next)
      i3 = current->next->data - 1;
    else
      i3 = count-1;
    nPoint = i3 - i0 + 1;

    degree = nPoint - 1;
    if (degree>3) degree = 3;

    if (0==degree && i0>0) {
      FiveTuple tmp1, tmp2;

      scaleTuple( prevDeriv, quadStep/prevDegree, &tmp2 );
      addTuples( p[i0-1], tmp2, &tmp1 );
#ifdef __DEBUG__
      indent();
      printf("bridge point %d to %d with degree 2 curve\n", i0-1, i0);
      indent(); printTuple( p[i0-1] );
      indent(); printTuple( tmp1 );
      indent(); printTuple( p[i0] );
#endif
      createQuadratic( &(*pCurves)[*curveCount], p[i0-1], tmp1, p[i0] );

      scaleTuple( prevDerivQ, quadStep/prevDegree, &tmp2 );
      addTuples( qLocal[i0-1], tmp2, &tmp1 );
      createQuadratic(&(*qCurves)[*curveCount],
		      qLocal[i0-1], tmp1, qLocal[i0] );

      (*curveCount)++;

      /* degree 0 can happen at the last point only,
	 so prevSlopes calculation is not necessary */
    }
    else if (1==degree) {
      FiveTuple tmp1, tmp2, tmp3;

      subTuples( p[i3], p[i0], &deriv );
      subTuples( qLocal[i3], qLocal[i0], &derivQ );
      
      if (i0 > 0) { /* connect p[i0-1] and p[i0] */
	scaleTuple( prevDeriv, cubicStep/prevDegree, &tmp3 );
	addTuples( p[i0-1], tmp3, &tmp1 );
	scaleTuple( deriv, -cubicStep/degree, &tmp3 );
	addTuples( p[i0], tmp3, &tmp2 );
#ifdef __DEBUG__
	indent();
	printf("bridge point %d to %d with degree 3 curve\n", i0-1, i0);
	indent(); printTuple( p[i0-1] );
	indent(); printTuple( tmp1 );
	indent(); printTuple( tmp2 );
	indent(); printTuple( p[i0] );
#endif
	createCubic(&(*pCurves)[*curveCount], p[i0-1], tmp1, tmp2, p[i0]);

	scaleTuple( prevDerivQ, cubicStep/prevDegree, &tmp3 );
	addTuples( qLocal[i0-1], tmp3, &tmp1 );
	scaleTuple( derivQ, -cubicStep/degree, &tmp3 );
	addTuples( qLocal[i0], tmp3, &tmp2 );
	createCubic(&(*qCurves)[*curveCount],
		    qLocal[i0-1], tmp1, tmp2, qLocal[i0]);

	(*curveCount)++;
      }

#ifdef __DEBUG__
      indent();
      printf("approximate point %d to %d with degree %d curve\n",
	     i0, i3, degree);
      indent(); printTuple( p[i0] );
      indent(); printTuple( p[i3] );
#endif
      createLinear( &(*pCurves)[*curveCount], p[i0], p[i3] );
      createLinear( &(*qCurves)[*curveCount], qLocal[i0], qLocal[i3] );
      (*curveCount)++;

      copyTuple( deriv, &prevDeriv );
      copyTuple( derivQ, &prevDerivQ );
    } /* degree 1 */
    else if (2==degree) {
      FiveTuple tmp1, tmpQ1, tmp2, tmp3, tmp4, len_tmp1, len_tmp2, ratio_tuple, len_sum;

      i1 = i0+1;

      subTuples(p[i1], p[i0], &tmp2);
      subTuples(p[i3], p[i1], &tmp3);
 
      spec_length(&tmp2, &len_tmp1,1);
      spec_length(&tmp3, &len_tmp2,1);

      addTuples(len_tmp1, len_tmp2, &len_sum);
      spec_ratio(&len_tmp1,&len_sum, &ratio_tuple, 1);

     /* b1 = (p1 - (1-u)^2 p0 - u^2 p3)/(2u(1-u) */

      /* for u,v only.. so you need to extract that quantity from the tuple */
      scaleTuple( p[i0], (ratio_tuple.s-1)*(1-ratio_tuple.s) , &tmp2);
      scaleTuple( p[i3], (-ratio_tuple.s)*(ratio_tuple.s) , &tmp3);
      addTuples( tmp2, tmp3, &tmp2);
      addTuples (tmp2, p[i1], &tmp2);
      scaleTuple(tmp2, (1.0/(2*ratio_tuple.s*(1-ratio_tuple.s))), &tmp3);

      tmp1.s = tmp3.s;
      tmp1.t = tmp3.t;

     /* Now for x,y,z */

      scaleTuple( p[i0], (ratio_tuple.x-1)*(1-ratio_tuple.x) , &tmp2);
      scaleTuple( p[i3], (-ratio_tuple.x)*(ratio_tuple.x) , &tmp3);
      addTuples( tmp2, tmp3, &tmp2);
      addTuples (tmp2, p[i1], &tmp2);
      scaleTuple(tmp2, (1.0/(2*ratio_tuple.x*(1-ratio_tuple.x))), &tmp3);

      tmp1.x = tmp3.x;
      tmp1.y = tmp3.y;
      tmp1.z = tmp3.z;

  /* Now tmp1 has the second control point for u-v and x-y-z */

      /* repeat for q[]*/
      subTuples(qLocal[i1], qLocal[i0], &tmp2);
      subTuples(qLocal[i3], qLocal[i1], &tmp3);
 
      spec_length(&tmp2, &len_tmp1,0);
      spec_length(&tmp3, &len_tmp2,0);

      addTuples(len_tmp1, len_tmp2, &len_sum);
      spec_ratio(&len_tmp1,&len_sum, &ratio_tuple, 0);

     /* b1 = (p1 - (1-u)^2 p0 - u^2 p3)/(2u(1-u) */

      /* for u,v only.. so you need to extract that quantity from the tuple */
      scaleTuple( qLocal[i0], (ratio_tuple.s-1)*(1-ratio_tuple.s) , &tmp2);
      scaleTuple( qLocal[i3], (-ratio_tuple.s)*(ratio_tuple.s) , &tmp3);
      addTuples( tmp2, tmp3, &tmp2);
      addTuples (tmp2, qLocal[i1], &tmp2);
      scaleTuple(tmp2, (1.0/(2*ratio_tuple.s*(1-ratio_tuple.s))), &tmp3);

      tmpQ1.s = tmp3.s;
      tmpQ1.t = tmp3.t;

     /* Now for x,y,z; but its same as uv patch */

      tmpQ1.x = tmp1.x;
      tmpQ1.y = tmp1.y;
      tmpQ1.z = tmp1.z;


      subTuples( tmp1, p[i0], &deriv );
      subTuples( tmpQ1, qLocal[i0], &derivQ );
      
      if (i0 > 0) { /* connect p[i0-1] and p[i0] */
	scaleTuple( prevDeriv, cubicStep/prevDegree, &tmp4 );
	addTuples( p[i0-1], tmp4, &tmp2 );
	scaleTuple( deriv, -cubicStep/degree, &tmp4 );
	addTuples( p[i0], tmp4, &tmp3 );
#ifdef __DEBUG__
	indent();
	printf("bridge point %d to %d with degree 3 curve\n", i0-1, i0);
	indent(); printTuple( p[i0-1] );
	indent(); printTuple( tmp2 );
	indent(); printTuple( tmp3 );
	indent(); printTuple( p[i0] );
#endif
	createCubic(&(*pCurves)[*curveCount], p[i0-1], tmp2, tmp3, p[i0]);

	scaleTuple( prevDerivQ, cubicStep/prevDegree, &tmp4 );
	addTuples( qLocal[i0-1], tmp4, &tmp2 );
	scaleTuple( derivQ, -cubicStep/degree, &tmp4 );
	addTuples( qLocal[i0], tmp4, &tmp3 );
	createCubic(&(*qCurves)[*curveCount],
		    qLocal[i0-1], tmp2, tmp3, qLocal[i0]);

	(*curveCount)++;
      }

#ifdef __DEBUG__
      indent();
      printf("approximate point %d to %d with degree %d curve\n",
	     i0, i3, degree);
      indent(); printTuple( p[i0] );
      indent(); printTuple( tmp1 );
      indent(); printTuple( p[i3] );
#endif
      createQuadratic( &(*pCurves)[*curveCount], p[i0], tmp1, p[i3] );
      createQuadratic( &(*qCurves)[*curveCount],
		       qLocal[i0], tmpQ1, qLocal[i3] );
      (*curveCount)++;

      subTuples( p[i3], tmp1, &prevDeriv );
      subTuples( qLocal[i3], tmpQ1, &prevDerivQ );
    } /* degree 2 */
    else {
      FiveTuple y1, y2, tmp1, tmp2, tmp3, const1, const2;
      FiveTuple pCtrl1, pCtrl2, qCtrl1, qCtrl2; /* control points */
      FiveTuple len1, len2, len3, len_sum, resultant1, resultant2;
      FiveTuple len_sum1, len_sum2, len_sum3;
      int s, r, n,i;
      double a11, a12, a21, a22, det;
 /*     if(4 == nPoint) */
      if(1) 
       {
        a11 =  (int)((float)nPoint/3.0);
         i1 = i0+(int)a11;
         i2 = i0+2*(int)a11;

         /* Initialize */
           len_sum1.s = len_sum2.s = len_sum3.s =0.0;
           len_sum1.t = len_sum2.t = len_sum3.t =0.0;
           len_sum1.x = len_sum2.x = len_sum3.x =0.0;
           len_sum1.y = len_sum2.y = len_sum3.y =0.0;
           len_sum1.z = len_sum2.z = len_sum3.z =0.0;

         /* Find the euclidean distance between the successive data points  */
         for(i=i0+1;i<=i1;i++)
          {
               subTuples(p[i],p[i-1],&tmp1);
               spec_length(&tmp1, &len1, 1);
               addTuples(len1, len_sum1,&len_sum1);
           }
        
         for(i=i1+1; i<=i2; i++)
          {
               subTuples(p[i],p[i-1],&tmp1);
               spec_length(&tmp1, &len1, 1);
               addTuples(len1, len_sum2,&len_sum2);
          }
        
         for(i=i2+1; i<=i3; i++)
          {
               subTuples(p[i],p[i-1],&tmp1);
               spec_length(&tmp1, &len1, 1);
               addTuples(len1, len_sum3,&len_sum3);
          }

         /* Find each distance's contribution to the sum */
         addTuples(len_sum1, len_sum2, &len_sum);
         addTuples(len_sum, len_sum3, &len_sum);

        /* find the ratio and thus the parameter value : for first control point */
         spec_ratio(&len_sum1, &len_sum, &tmp1, 1);
         spec_ratio(&len_sum2, &len_sum, &tmp2, 1);

         /* for second control point */
         addTuples(tmp1, tmp2, &tmp2);

         /* We have a system of linear equations to be solved with the second and
            third data points on one side  and the bernstein polynomial in the other */
         /* set of unknowns is the set of two control points in the bernstein polyn. */
   
         /* the a11,.. are the constants in the cofactor matrix to find the inverse
            of the constant matrix */
         a22 = 3*tmp1.s*(1-tmp1.s)*(1-tmp1.s);
         a21 = -3*tmp2.s*(1-tmp2.s)*(1-tmp2.s);

         a12 = -3*tmp1.s*tmp1.s*(1-tmp1.s);
         a11 = 3*tmp2.s*tmp2.s*(1-tmp2.s);
         
         det = a11*a22 - a12*a21;

         /* finding RHS of the set of linear equations (LHS contains the unknown ctrl pts.) */

        /* const1 =  i1 - (1-u1)^3*i0 - u1^3*i3 */
        /* const2 =  i2 - (1-u2)^3*i0 - u2^3*i3 */
         
         scaleTuple(p[i0],(1-tmp1.s)*(1-tmp1.s)*(1-tmp1.s), &tmp3);
         scaleTuple(p[i3],tmp1.s*tmp1.s*tmp1.s, &const1);
         addTuples(tmp3, const1, &const1);
         subTuples(p[i1], const1, &const1);
         
         scaleTuple(p[i0],(1-tmp2.s)*(1-tmp2.s)*(1-tmp2.s), &tmp3);
         scaleTuple(p[i3],tmp2.s*tmp2.s*tmp2.s, &const2);
         addTuples(tmp3, const2, &const2);
         subTuples(p[i2], const2, &const2);

         /* Finding the control points */
         /* Ctrl1 = (a11*const1 + a12*const2)/det */

         scaleTuple(const1, a11, &resultant1);
         scaleTuple(const2, a12, &resultant2);
         addTuples(resultant1, resultant2, &resultant1);
         scaleTuple(resultant1, 1.0/det, &resultant2);

         pCtrl1.s = resultant2.s;
         pCtrl1.t = resultant2.t;
           
         /* Ctrl2 = (a21*const1 + a22*const2)/det */

         scaleTuple(const1, a21, &resultant1);
         scaleTuple(const2, a22, &resultant2);
         addTuples(resultant1, resultant2, &resultant1);
         scaleTuple(resultant1, 1.0/det, &resultant2);

         pCtrl2.s = resultant2.s;
         pCtrl2.t = resultant2.t;
           
       /* Now repeat the same for the object space curve */
         
         a22 = 3*tmp1.x*(1-tmp1.x)*(1-tmp1.x);
         a21 = -3*tmp2.x*(1-tmp2.x)*(1-tmp2.x);

         a12 = -3*tmp1.x*tmp1.x*(1-tmp1.x);
         a11 = 3*tmp2.x*tmp2.x*(1-tmp2.x);
         
         det = a11*a22 - a12*a21;

         /* finding RHS of the set of linear equations (LHS contains the unknown ctrl pts.) */
         
        /* const1 =  i1 - (1-x1)^3*i0 - x1^3*i3 */
        /* const2 =  i2 - (1-x2)^3*i0 - x2^3*i3 */
         
         scaleTuple(p[i0],(1-tmp1.x)*(1-tmp1.x)*(1-tmp1.x), &tmp3);
         scaleTuple(p[i3],tmp1.x*tmp1.x*tmp1.x, &const1);
         addTuples(tmp3, const1, &const1);
         subTuples(p[i1], const1, &const1);
         
         scaleTuple(p[i0],(1-tmp2.x)*(1-tmp2.x)*(1-tmp2.x), &tmp3);
         scaleTuple(p[i3],tmp2.x*tmp2.x*tmp2.x, &const2);
         addTuples(tmp3, const2, &const2);
         subTuples(p[i2], const2, &const2);

         /* Finding the control points */
         /* Ctrl1 = (a11*const1 + a12*const2)/det */
         /* Ctrl2 = (a21*const1 + a22*const2)/det */

         scaleTuple(const1, a11, &resultant1);
         scaleTuple(const2, a12, &resultant2);
         addTuples(resultant1, resultant2, &resultant1);
         scaleTuple(resultant1, 1.0/det, &resultant2);

         pCtrl1.x = resultant2.x;
         pCtrl1.y = resultant2.y;
         pCtrl1.z = resultant2.z;

         /* Ctrl2 = (a21*const1 + a22*const2)/det */

         scaleTuple(const1, a21, &resultant1);
         scaleTuple(const2, a22, &resultant2);
         addTuples(resultant1, resultant2, &resultant1);
         scaleTuple(resultant1, 1.0/det, &resultant2);

         pCtrl2.x = resultant2.x;
         pCtrl2.y = resultant2.y;
         pCtrl2.z = resultant2.z;

     /* Now repeat it for the qCurve.. */
           
         /* Find the euclidean distance between the successive data points  */
         subTuples(qLocal[i1],qLocal[i0],&tmp1);
         spec_length(&tmp1, &len1, 0);
        
         subTuples(qLocal[i2], qLocal[i1], &tmp1);
         spec_length(&tmp1, &len2, 0);
        
         subTuples(qLocal[i3], qLocal[i2], &tmp1);
         spec_length(&tmp1, &len3, 0);

         /* Find each distance's contribution to the sum */
         addTuples(len1, len2, &len_sum);
         addTuples(len_sum, len3, &len_sum);

        /* find the ratio and thus the parameter value : for first control point */
         spec_ratio(&len1, &len_sum, &tmp1, 0);
         spec_ratio(&len2, &len_sum, &tmp2, 0);

         /* for second control point */
         addTuples(tmp1, tmp2, &tmp2);

         /* We have a system of linear equations to be solved with the second and
            third data points on one side  and the bernstein polynomial in the other */
         /* set of unknowns is the set of two control points in the bernstein polyn. */
   
         /* the a11,.. are the constants in the cofactor matrix to find the inverse
            of the constant matrix */
         a22 = 3*tmp1.s*(1-tmp1.s)*(1-tmp1.s);
         a21 = -3*tmp2.s*(1-tmp2.s)*(1-tmp2.s);

         a12 = -3*tmp1.s*tmp1.s*(1-tmp1.s);
         a11 = 3*tmp2.s*tmp2.s*(1-tmp2.s);
         
         det = a11*a22 - a12*a21;

         /* finding RHS of the set of linear equations (LHS contains the unknown ctrl pts.) */
         
         scaleTuple(qLocal[i0],(1-tmp1.s)*(1-tmp1.s)*(1-tmp1.s), &tmp3);
         scaleTuple(qLocal[i3],tmp1.s*tmp1.s*tmp1.s, &const1);
         addTuples(tmp3, const1, &const1);
         subTuples(qLocal[i1], const1, &const1);
         
         scaleTuple(qLocal[i0],(1-tmp2.s)*(1-tmp2.s)*(1-tmp2.s), &tmp3);
         scaleTuple(qLocal[i3],tmp2.s*tmp2.s*tmp2.s, &const2);
         addTuples(tmp3, const2, &const2);
         subTuples(qLocal[i2], const2, &const2);

         /* Finding the control points */
         /* Ctrl1 = (a11*const1 + a12*const2)/det */
         /* Ctrl2 = (a21*const1 + a22*const2)/det */

         scaleTuple(const1, a11, &resultant1);
         scaleTuple(const2, a12, &resultant2);
         addTuples(resultant1, resultant2, &resultant1);
         scaleTuple(resultant1, 1.0/det, &resultant2);

         qCtrl1.s = resultant2.s;
         qCtrl1.t = resultant2.t;
         qCtrl1.x = pCtrl1.x ; 
         qCtrl1.y = pCtrl1.y ; 
         qCtrl1.z = pCtrl1.z ; 
           

         scaleTuple(const1, a21, &resultant1);
         scaleTuple(const2, a22, &resultant2);
         addTuples(resultant1, resultant2, &resultant1);
         scaleTuple(resultant1, 1.0/det, &resultant2);

         qCtrl2.s = resultant2.s;
         qCtrl2.t = resultant2.t;
         qCtrl2.x = pCtrl2.x ; 
         qCtrl2.y = pCtrl2.y ; 
         qCtrl2.z = pCtrl2.z ; 
       }
     else 
       {
      
      n = nPoint-1;
      s = n / 3;
      r = n - s;
      a11 = a22 = 3*s*r*r;
      a12 = a21 = 3*s*s*r;
      det = (double) a11*a22 - a12*a21;
      i1 = i0 + s;
      i2 = i0 + r;

      /* y1 = p1*n^3 - p0*r^3 - p3*s^3 */
      scaleTuple( p[i1], (double) n*n*n, &tmp1 );
      scaleTuple( p[i0], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( p[i3], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y1 );
      /* y2 = p2*n^3 - p0*s^3 - p3*r^3 */
      scaleTuple( p[i2], (double) n*n*n, &tmp1 );
      scaleTuple( p[i3], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( p[i0], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y2 );

      /* pCtrl1 = (a22*y1 - a12*y2)/det */
      scaleTuple( y1, a22, &tmp1 );
      scaleTuple( y2, -a12, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &pCtrl1 );
      /* pCtrl2 = (-a21*y1 + a11*y2)/det */
      scaleTuple( y1, -a21, &tmp1 );
      scaleTuple( y2, a11, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &pCtrl2 );

      /*---------- repeat for q[] ----------*/
      /* y1 = q1*n^3 - q0*r^3 - q3*s^3 */
      scaleTuple( qLocal[i1], (double) n*n*n, &tmp1 );
      scaleTuple( qLocal[i0], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( qLocal[i3], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y1 );
      /* y2 = q2*n^3 - q0*s^3 - q3*r^3 */
      scaleTuple( qLocal[i2], (double) n*n*n, &tmp1 );
      scaleTuple( qLocal[i3], (double) -r*r*r, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( qLocal[i0], (double) -s*s*s, &tmp2 );
      addTuples( tmp3, tmp2, &y2 );

      /* qCtrl1 = (a22*y1 - a12*y2)/det */
      scaleTuple( y1, a22, &tmp1 );
      scaleTuple( y2, -a12, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &qCtrl1 );
      /* qCtrl2 = (-a21*y1 + a11*y2)/det */
      scaleTuple( y1, -a21, &tmp1 );
      scaleTuple( y2, a11, &tmp2 );
      addTuples( tmp1, tmp2, &tmp3 );
      scaleTuple( tmp3, (double) 1/det, &qCtrl2 );

#ifdef __DEBUG__
      indent();
      printf("approximate point %d to %d with degree %d curve\n",
	     i0, i3, degree);
      indent(); printTuple( p[i0] );
      indent(); printTuple( pCtrl1 );
      indent(); printTuple( pCtrl2 );
      indent(); printTuple( p[i3] );
#endif
     }

      /*
       * Now, we have the control points of the cubic curves for p[] and q[].
       * But, instead of putting them in (*pCurves)[] and (*qCurves)[],
       * we will save them in tmpPCurves[] and tmpQCurves[] first.
       * This is because the "bridging curves" (the curve to bridge the
       * last point of the previous partition) should be stored first.
       *
       * However, we can not compute and store the "bridge curve" 
       * before we compute the current curve, because the slope 
       * required for the bridging curves depends on the first control
       * point of current curve, which may not be pCtrl1 or qCtrl1 if the
       * curve fitting is not good enough and requires further partitioning
       */
      tmpPCurves = (ParamCurve*) malloc(sizeof(ParamCurve));
      tmpQCurves = (ParamCurve*) malloc(sizeof(ParamCurve));
      tmpCount = 1;
      createCubic( &tmpPCurves[0],
		   p[i0], pCtrl1, pCtrl2, p[i3] );
      createCubic( &tmpQCurves[0],
		   qLocal[i0], qCtrl1, qCtrl2, qLocal[i3] );

      /********************************************************************
	ERROR BOUND CHECKING !!!
      ********************************************************************/
#ifdef __DEBUG__
      indent();
      printf("error bound for P[] = %.15f for Q[] = %.15f\n",
	     fittingErrorEstimate(nPoint, &p[i0], tmpPCurves[0]),
	     fittingErrorEstimate(nPoint, &qLocal[i0], tmpQCurves[0]));
#endif
      if (nPoint > 4 &&
	(fittingErrorEstimate(nPoint, &p[i0], tmpPCurves[0])>errorBound ||
	 fittingErrorEstimate(nPoint, &qLocal[i0], tmpQCurves[0])>errorBound))
      {
#ifdef __DEBUG__
	indent();
	printf("fitting error is out of bound, calling recursively\n");
	indent();
	printf("point 0 will be correspondent to current point %d\n", i0);
	indent();
	printf("---beginning of nesting level %d output---------------\n",
	       ++indentLevel);
#endif
	free(tmpPCurves);
	free(tmpQCurves);
	switch (simpleSplit) {
	  case 0:
	  case 1:
	    fitting(nPoint, &p[i0], &qLocal[i0],
		    &tmpCount, &tmpPCurves, &tmpQCurves,
		    1, errorBound, checkDomain);
	    break;
	  case -1:
	  default:
	    fitting(nPoint, &p[i0], &qLocal[i0],
		    &tmpCount, &tmpPCurves, &tmpQCurves,
		    0, errorBound, checkDomain);
	    break;
	}
#ifdef __DEBUG__
	indentLevel--;
	indent();
	printf("---end of nesting level %d message--------------------\n",
	       indentLevel+1);
#endif
      }

      subTuples( tmpPCurves[0].control[0], p[i0], &deriv );
      subTuples( tmpQCurves[0].control[0], qLocal[i0], &derivQ );

      if (i0 > 0) { /* connect p[i0-1] and p[i0] */
	scaleTuple( prevDeriv, cubicStep/prevDegree, &tmp3 );
	addTuples( p[i0-1], tmp3, &tmp1 );
	scaleTuple( deriv, -cubicStep/degree, &tmp3 );
	addTuples( p[i0], tmp3, &tmp2 );
	
#ifdef __DEBUG__
	indent();
	printf("bridge point %d to %d with degree 3 curve\n", i0-1, i0);
	indent(); printTuple( p[i0-1] );
	indent(); printTuple( tmp1 );
	indent(); printTuple( tmp2 );
	indent(); printTuple( p[i0] );
#endif
	createCubic(&(*pCurves)[*curveCount], p[i0-1], tmp1, tmp2, p[i0]);

	/* repeat for q[] */
	scaleTuple( prevDerivQ, cubicStep/prevDegree, &tmp3 );
	addTuples( qLocal[i0-1], tmp3, &tmp1 );
	scaleTuple( derivQ, -cubicStep/degree, &tmp3 );
	addTuples( qLocal[i0], tmp3, &tmp2 );

	createCubic(&(*qCurves)[*curveCount],
		    qLocal[i0-1], tmp1, tmp2, qLocal[i0]);
	
	(*curveCount)++;
      }
      
      /* copying tmpPCurves[] to (*pCurves)[], tmpQCurves[] to (*qCurves)[]
         and free the memory */
      for (i=0; i<tmpCount; i++) {
	(*pCurves)[*curveCount].degree = tmpPCurves[i].degree;
	(*pCurves)[*curveCount].control = tmpPCurves[i].control;
	(*qCurves)[*curveCount].degree = tmpQCurves[i].degree;
	(*qCurves)[*curveCount].control = tmpQCurves[i].control;
	(*curveCount)++;
      }

      /* Now, the "bridging" curves are done.
	 Create the slopes for the next "bridging" */
      i = tmpPCurves[tmpCount-1].degree;
      subTuples( p[i3], tmpPCurves[tmpCount-1].control[i], &prevDeriv );
      i = tmpQCurves[tmpCount-1].degree;
      subTuples( qLocal[i3], tmpQCurves[tmpCount-1].control[i], &prevDerivQ );

      free(tmpPCurves);
      free(tmpQCurves);
    } /* degree 3 */
      
    prevDegree = degree;
    current = current->next;
  }

  if (qReversed) {
    assert( q != qLocal);  /* make sure we don't free q[] accidentally */
    free( qLocal );
  }

  /* pack the memory, so the sizes of (*pCurves)[] and (*qCurves)[] 
     == *curveCount */

  tmpPCurves = *pCurves;
  tmpQCurves = *qCurves;
  *pCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*(*curveCount));
  *qCurves = (ParamCurve*) malloc(sizeof(ParamCurve)*(*curveCount));
  for (i=0; i<*curveCount; i++) {
    int j = qReversed ? *curveCount-i-1 : i;
    (*pCurves)[i].degree = tmpPCurves[i].degree;
    (*pCurves)[i].control = tmpPCurves[i].control;
    (*qCurves)[j].degree = tmpQCurves[i].degree;
    (*qCurves)[j].control = tmpQCurves[i].control;
  }
  free(tmpPCurves);
  free(tmpQCurves);
  freeList(breakPoints);
}


