#include <tree.h>
#include <macro.h>

#define SIGN(n) (((n)%2 == 0) ? 1 : -1)

#define CLOSE(a, b)  ((fabs((a) - (b)) <= 5.0e-3))

extern void copy_xyz_spline();

static const double UV_EPS = 0.05;

static int C(int n, int r)
{
  register int i;
  int product = 1;
  
  if (r > n/2)			/* Use C(n, r) = C(n, n - r) */
    r = n - r;
  
  for (i = 0; i < r; i++)
    product *= (n - i);

  for (i = 1; i <= r; i++)
    product /= i;
  
  return product;
}


/* Input: power coefficients in array 'a' */
/* Output: Bezeir coefficients in array 'b' */
/* Space must already be allocated to both the arrays before calling */
/* this routine */
/* Use back-substitution to get the values of the Bezier coeffs. (the */
/* first and last Bezier points are already known from the data). Use */
/* these values to determine other coeffs */

int power_to_bez(int *ord, vector_t *a, vector_t *b)
{
  register int i, j, k;
  double sum;
  int n;
  
  /* First, try to eliminate the zero coefficients to see if we can */
  /* get away with a lower degree bezier representation */
   
  for (i = *ord - 1; i > 0; i--)
    if (!((a[i][0] == 0.0) && (a[i][1] == 0.0) && (a[i][2] == 0.0)))
      break;
  
  *ord = i + 1;
  n = *ord - 1;			/* n is the degree of the curve */
  
  /* Now get the Bezier control points for the curve */
  
  for (k = 0; k < 3; k++)	/* for X, Y, Z do */
    for (i = 0; i <= n; i++)
      {				/* Calculate b[i][k] */
	sum = 0.0;
	for (j = 0; j < i; j++)
	  sum += C(n, j) * C(n - j, i - j) * SIGN(i - j) * b[j][k];  
	b[i][k] = (a[i][k] - sum)/C(n, i);
      }
  
  for (i = 0; i <= n; i++)
    b[i][3] = 1.0;		/* w = 1 for each point */

  return *ord;
}

#define JOHNSSUBDIVISION
#ifdef JOHNSSUBDIVISION
/* Subdivide the tensor-product Bezier patch within the region  */
/* 0 <= u0 < u1 <= 1, and 0 <= v0 < v1 <= 1 */
/* The resulting patch is obtained in tpout. Space must already */
/* allocated in tpout to hold all the control points */

/* I have removed the static definition because I am using it - SK */
/* Modified version by JK to speed up 5/30/96 */
/* Note: this version is faster (by a factor of 2 or more) than */
/*       the other version, but it requires more space- proportional */
/*       to MAX_CURVEDEG^2 instead of 2*MAX_CURVEDEG */

int subdivide_bezier_patch(TENSOR_PATCH tp, double u0,
                           double u1, double v0, double v1,
                           TENSOR_PATCH *tpout)
{
  int i, j, k, l;
  POINT4D b[MAX_CURVEDEG][MAX_CURVEDEG];
  int n;

  tpout->iDeg0 = tp.iDeg0;
  tpout->iDeg1 = tp.iDeg1;

  /* Subdivide along u first */
  n = tp.iDeg0;
  for (l=0; l<= tp.iDeg1; l++) {
     for(i=0;i<=n;i++) {
        b[0][i].dCoord[0] = tp.coeffs[i][l].dCoord[0];
        b[0][i].dCoord[1] = tp.coeffs[i][l].dCoord[1];
        b[0][i].dCoord[2] = tp.coeffs[i][l].dCoord[2];
        b[0][i].dCoord[3] = tp.coeffs[i][l].dCoord[3];
     }

     for(i=1;i<=n;i++) {
        for(k=0; k <= (n-i); k++) {
           b[i][k].dCoord[0] = (1.0-u0)*b[i-1][k].dCoord[0] +
                                     u0*b[i-1][k+1].dCoord[0];
           b[i][k].dCoord[1] = (1.0-u0)*b[i-1][k].dCoord[1] +
                                     u0*b[i-1][k+1].dCoord[1];
           b[i][k].dCoord[2] = (1.0-u0)*b[i-1][k].dCoord[2] +
                                     u0*b[i-1][k+1].dCoord[2];
           b[i][k].dCoord[3] = (1.0-u0)*b[i-1][k].dCoord[3] +
                                     u0*b[i-1][k+1].dCoord[3];
        }
     }
     tpout->coeffs[0][l].dCoord[0] = b[n][0].dCoord[0];
     tpout->coeffs[0][l].dCoord[1] = b[n][0].dCoord[1];
     tpout->coeffs[0][l].dCoord[2] = b[n][0].dCoord[2];
     tpout->coeffs[0][l].dCoord[3] = b[n][0].dCoord[3];

     for (j=1;j<=n;j++) {
        for (i=(n-j+1); i<=n; i++) {
           for (k=0; k <= (n-i); k++) {
              b[i][k].dCoord[0] = (1.0-u1)*b[i-1][k].dCoord[0] +
                                        u1*b[i-1][k+1].dCoord[0];
              b[i][k].dCoord[1] = (1.0-u1)*b[i-1][k].dCoord[1] +
                                        u1*b[i-1][k+1].dCoord[1];
              b[i][k].dCoord[2] = (1.0-u1)*b[i-1][k].dCoord[2] +
                                        u1*b[i-1][k+1].dCoord[2];
              b[i][k].dCoord[3] = (1.0-u1)*b[i-1][k].dCoord[3] +
                                        u1*b[i-1][k+1].dCoord[3];
           }
        }
        tpout->coeffs[j][l].dCoord[0] = b[n][0].dCoord[0];
        tpout->coeffs[j][l].dCoord[1] = b[n][0].dCoord[1];
        tpout->coeffs[j][l].dCoord[2] = b[n][0].dCoord[2];
        tpout->coeffs[j][l].dCoord[3] = b[n][0].dCoord[3];
     }
  }

  /* now subdivide along v */
  n = tp.iDeg1;
  for (l=0; l<= tp.iDeg0; l++) {
     for(i=0;i<=n;i++) {
        b[0][i].dCoord[0] = tpout->coeffs[l][i].dCoord[0];
        b[0][i].dCoord[1] = tpout->coeffs[l][i].dCoord[1];
        b[0][i].dCoord[2] = tpout->coeffs[l][i].dCoord[2];
        b[0][i].dCoord[3] = tpout->coeffs[l][i].dCoord[3];
     }

     for(i=1;i<=n;i++) {
        for(k=0; k <= (n-i); k++) {
           b[i][k].dCoord[0] = (1.0-v0)*b[i-1][k].dCoord[0] +
                                     v0*b[i-1][k+1].dCoord[0];
           b[i][k].dCoord[1] = (1.0-v0)*b[i-1][k].dCoord[1] +
                                     v0*b[i-1][k+1].dCoord[1];
           b[i][k].dCoord[2] = (1.0-v0)*b[i-1][k].dCoord[2] +
                                     v0*b[i-1][k+1].dCoord[2];
           b[i][k].dCoord[3] = (1.0-v0)*b[i-1][k].dCoord[3] +
                                     v0*b[i-1][k+1].dCoord[3];
        }
     }
     tpout->coeffs[l][0].dCoord[0] = b[n][0].dCoord[0];
     tpout->coeffs[l][0].dCoord[1] = b[n][0].dCoord[1];
     tpout->coeffs[l][0].dCoord[2] = b[n][0].dCoord[2];
     tpout->coeffs[l][0].dCoord[3] = b[n][0].dCoord[3];

     for (j=1;j<=n;j++) {
        for (i=(n-j+1); i<=n; i++) {
           for (k=0; k <= (n-i); k++) {
              b[i][k].dCoord[0] = (1.0-v1)*b[i-1][k].dCoord[0] +
                                        v1*b[i-1][k+1].dCoord[0];
              b[i][k].dCoord[1] = (1.0-v1)*b[i-1][k].dCoord[1] +
                                        v1*b[i-1][k+1].dCoord[1];
              b[i][k].dCoord[2] = (1.0-v1)*b[i-1][k].dCoord[2] +
                                        v1*b[i-1][k+1].dCoord[2];
              b[i][k].dCoord[3] = (1.0-v1)*b[i-1][k].dCoord[3] +
                                        v1*b[i-1][k+1].dCoord[3];
           }
        }
        tpout->coeffs[l][j].dCoord[0] = b[n][0].dCoord[0];
        tpout->coeffs[l][j].dCoord[1] = b[n][0].dCoord[1];
        tpout->coeffs[l][j].dCoord[2] = b[n][0].dCoord[2];
        tpout->coeffs[l][j].dCoord[3] = b[n][0].dCoord[3];
     }
  }
  return 0;
}

#else
/* Subdivide the tensor-product Bezier patch within the region  */
/* 0 <= u0 < u1 <= 1, and 0 <= v0 < v1 <= 1 */
/* The resulting patch is obtained in tpout. Space must already */
/* allocated in tpout to hold all the control points */

/* I have removed the static definition because I am using it - SK */

int subdivide_bezier_patch(TENSOR_PATCH tp, double u0, 
			   double u1, double v0, double v1,
			   TENSOR_PATCH *tpout)
{
  register int i, j;
  int k, l, m;
  vector_t b[MAX_CURVEDEG], temp[MAX_CURVEDEG];
  int n;
  double c;

  tpout->iDeg0 = tp.iDeg0;
  tpout->iDeg1 = tp.iDeg1;

  /* Subdivide along u first */
  n = tp.iDeg0;
  for (j = 0; j <= tp.iDeg1; j++)
    {
      for (l = 0; l <= n; l++)	/* evaluate each bezier point */
	{
	  for (i = 0; i <= n; i++)
	    {
	      yourpt_to_mypt(tp.coeffs[i][j], b[i]);
	    }
	  
	  c = u0;
	  for (i = 0; i < n; i++)
	    {
	      if (i >= (n - l))
		c = u1;
	      for (k = 0; k < (n - i); k++)
	      {
		b[k][0] = (1 - c)*b[k][0] + c*b[k + 1][0];
		b[k][1] = (1 - c)*b[k][1] + c*b[k + 1][1];
		b[k][2] = (1 - c)*b[k][2] + c*b[k + 1][2];
		b[k][3] = (1 - c)*b[k][3] + c*b[k + 1][3]; 
	      }
	    }
	  
	  /* Now the new point is present in b[0] */
	  
	  for (m = 0; m < 4; m++)
	    temp[l][m] = b[0][m];
	}

      /* copy back the new points in 'temp' into the patch */
      for (i = 0; i <= n; i++)
	{
	  mypt_to_yourpt_no_w(temp[i], tpout->coeffs[i][j]);
	  tpout->coeffs[i][j].dCoord[3] = temp[i][3];
	}
    }

  /* Now perform the same operation along the V direction */
  
  n = tp.iDeg1;
  for (j = 0; j <= tp.iDeg0; j++)
    {
      for (l = 0; l <= n; l++)	/* evaluate each bezier point */
	{
	  for (i = 0; i <= n; i++)
	    {
	      yourpt_to_mypt(tpout->coeffs[j][i], b[i]);
	    }

	  c = v0;
	  for (i = 0; i < n; i++)
	    {
	      if (i >= (n - l))
		c = v1;
	      for (k = 0; k < (n - i); k++)
	      {
		b[k][0] = (1 - c)*b[k][0] + c*b[k + 1][0];
		b[k][1] = (1 - c)*b[k][1] + c*b[k + 1][1];
		b[k][2] = (1 - c)*b[k][2] + c*b[k + 1][2];
		b[k][3] = (1 - c)*b[k][3] + c*b[k + 1][3]; 
	      }
	    }
	  
	  /* Now the new point is present in b[0] */
	  for (m = 0; m < 4; m++)
	    temp[l][m] = b[0][m];
	}

      /* copy back the new points in 'temp' into the patch */
      for (i = 0; i <= n; i++)
	{
	  mypt_to_yourpt_no_w(temp[i], tpout->coeffs[j][i]);
	  tpout->coeffs[j][i].dCoord[3] = temp[i][3];	  
	}
    }
  
  return 0;
}
#endif

int subdivide_bezier_patch_using_trim_curve(TRIM_PATCH_INFO *tpiin,
					    int regnum,
					    TRIM_PATCH_INFO *tpiout)
{
  TENSOR_PATCH *tpin, *tpout;
  double (*regin)[2], (*regout)[2];
  LIST_POINT2D *trim_curveout;
  int i, j, n;
  double umax, umin, vmax, vmin;
  double tmpu, tmpv;
  double udenom, vdenom;
  LIST_POINT2D *p, *q, *r;
  TRIM_PORTION_SPLINE_TYPE *tpstin, *tpstout;

  tpin = &tpiin->tp;
  tpout = &tpiout->tp;
  tpiout->NumTrimPortions = 1;
  tpiout->csg_status = IN_CSG;
  tpiout->primNum = tpiin->primNum; /* propagate the primitive number */
  tpiout->SplineReprn = tpiin->SplineReprn;

  if (!tpiout->SplineReprn) {
    n =  tpiin->TrimPortion[regnum].PiecePortion.NumOfVertices; 
                                           /* #vertices forming */
                                           /* region */
    regin = tpiin->TrimPortion[regnum].PiecePortion.poly_vertex;
    regout = tpiout->TrimPortion[0].PiecePortion.poly_vertex;
    tpiout->TrimPortion[0].PiecePortion.NumOfVertices = n;  
    regout = (double (*)[2]) new(MAX_VERTEX_COUNT*2*sizeof(double));
    umax = umin = regin[0][0];
    vmax = vmin = regin[0][1];
    
    for (i = 1; i <= n; i++)
      {
	if (regin[i][0] > umax)
	  umax = regin[i][0];
	else if (regin[i][0] < umin)
	  umin = regin[i][0];
	
	if (regin[i][1] > vmax)
	  vmax = regin[i][1];
	else if (regin[i][1] < vmin)
	  vmin = regin[i][1];
      }
    
    
    /* Allow a little tolerance in the scale so that the trim curve does */
    /* not graze at the border ?? */
    
    if ((umin -= UV_EPS) < 0.0)
      umin = 0.0;
    if ((umax += UV_EPS) > 1.0)
      umax = 1.0;
    if ((vmin -= UV_EPS) < 0.0)
      vmin = 0.0;
    if ((vmax += UV_EPS) > 1.0)
      vmax = 1.0;
  
    /* Now traverse the list for the trim curve again and scale */
    /* all the points appropriately */
  
    udenom = umax - umin;
    vdenom = vmax - vmin;
    
    for (i = 0; i <= n; i++)
      {
	regout[i][0] = (regin[i][0] - umin)/udenom;
	regout[i][1] = (regin[i][1] - vmin)/vdenom;
      }

    /* Change pt inside region */
    tpiout->TrimPortion[0].PiecePortion.ptInsideRegion.u = 
      (tpiin->TrimPortion[regnum].PiecePortion.ptInsideRegion.u - umin)/udenom;
    tpiout->TrimPortion[0].PiecePortion.ptInsideRegion.v = 
      (tpiin->TrimPortion[regnum].PiecePortion.ptInsideRegion.v - vmin)/vdenom;
  } else {
    tpstin = &tpiin->TrimPortion[regnum].SplinePortion;
    tpstout = &tpiout->TrimPortion[0].SplinePortion;
    n =  tpstout->NumOfCurves = tpstin->NumOfCurves; 

    tpstout->UV_Splines = (UV_SPLINE_TYPE *)
                             new(n * sizeof(UV_SPLINE_TYPE));
    tpstout->XYZ_Splines = (XYZ_SPLINE_TYPE *)
                             new(n * sizeof(XYZ_SPLINE_TYPE));
    tpstout->AdjInfo = (int *) new(n * sizeof(int));

    umax = umin = (tpstin->UV_Splines[0].CntrlPts[0].dCoord[0] /
		   tpstin->UV_Splines[0].CntrlPts[0].dCoord[2]);
    vmax = vmin = (tpstin->UV_Splines[0].CntrlPts[0].dCoord[1] /
		   tpstin->UV_Splines[0].CntrlPts[0].dCoord[2]);
    
    for (i = 0; i < n; i++)
      for (j=0; j<=tpstin->UV_Splines[i].Deg; j++) {
	tmpu = (tpstin->UV_Splines[i].CntrlPts[j].dCoord[0] /
		tpstin->UV_Splines[i].CntrlPts[j].dCoord[2]);
	tmpv = (tpstin->UV_Splines[i].CntrlPts[j].dCoord[1] /
		tpstin->UV_Splines[i].CntrlPts[j].dCoord[2]);
	if (tmpu > umax)
	  umax = tmpu;
	else if (tmpu < umin)
	  umin = tmpu;
	
	if (tmpv > vmax)
	  vmax = tmpv;
	else if (tmpv < vmin)
	  vmin = tmpv;
      }
    
    
    /* Allow a little tolerance in the scale so that the trim curve does */
    /* not graze at the border ?? */
    
    if ((umin -= UV_EPS) < 0.0)
      umin = 0.0;
    if ((umax += UV_EPS) > 1.0)
      umax = 1.0;
    if ((vmin -= UV_EPS) < 0.0)
      vmin = 0.0;
    if ((vmax += UV_EPS) > 1.0)
      vmax = 1.0;
  
    /* Now traverse the list for the trim curve again and scale */
    /* all the points appropriately */

    udenom = umax - umin;
    vdenom = vmax - vmin;
    
    for (i = 0; i < n; i++) {
      tpstout->UV_Splines[i].Deg = tpstin->UV_Splines[i].Deg;
      tpstout->UV_Splines[i].CntrlPts = (POINT3D *) 
	new((tpstout->UV_Splines[i].Deg+1) * sizeof(POINT3D));
      for (j=0; j<=tpstout->UV_Splines[i].Deg; j++) {
	tpstout->UV_Splines[i].CntrlPts[j].dCoord[0] = 
	  (tpstin->UV_Splines[i].CntrlPts[j].dCoord[0] - 
	   umin*tpstin->UV_Splines[i].CntrlPts[j].dCoord[2]) / udenom;
	tpstout->UV_Splines[i].CntrlPts[j].dCoord[1] = 
	  (tpstin->UV_Splines[i].CntrlPts[j].dCoord[1] - 
	   vmin*tpstin->UV_Splines[i].CntrlPts[j].dCoord[2]) / vdenom;
	tpstout->UV_Splines[i].CntrlPts[j].dCoord[2] = 
	  tpstin->UV_Splines[i].CntrlPts[j].dCoord[2];
      }
      copy_xyz_spline(&tpstin->XYZ_Splines[i], &tpstout->XYZ_Splines[i]);
      /* The true adjacency will be figured out outside this procedure */
      tpstout->AdjInfo[i] = -1;
    }

    /* Change pt inside region */
    tpstout->ptInsideRegion.u = (tpstin->ptInsideRegion.u - umin)/udenom;
    tpstout->ptInsideRegion.v = (tpstin->ptInsideRegion.v - vmin)/vdenom;
  }

  /* Now subdivide the tensor patch to retain only the portion within */
  /* the values (umin, umax) and (vmin, vmax) */
  
  tpout->coeffs = (POINT4D **) new((tpin->iDeg0 + 1)*sizeof(POINT4D *));
  for (i = 0; i <= tpin->iDeg0; i++)
    tpout->coeffs[i] = (POINT4D *) new((tpin->iDeg1 +
					1)*sizeof(POINT4D)); 

  tpiout->umin = tpiin->umin + umin * (tpiin->umax - tpiin->umin);
  tpiout->umax = tpiin->umin + umax * (tpiin->umax - tpiin->umin);
  tpiout->vmin = tpiin->vmin + vmin * (tpiin->vmax - tpiin->vmin);
  tpiout->vmax = tpiin->vmin + vmax * (tpiin->vmax - tpiin->vmin);
/*
  printf("OLD : %lf %lf %lf %lf ::: NEW : %lf %lf %lf %lf\n", 
		tpiin->umin,tpiin->umax,
		tpiin->vmin,tpiin->vmax,tpiout->umin,tpiout->umax,
		tpiout->vmin,tpiout->vmax);
*/

  /* Copy the original, undeformed tensor patch to the the new */
  /* structure. This is maintained to remove crcks while rendering the */
  /* whole solid */
  copy_ten_p(&tpiin->orig_tp, &tpiout->orig_tp);

  subdivide_bezier_patch(*tpin, umin, umax, vmin, vmax, tpout);
  
  return 0;
}


int subdivide_trimmed_patches(node_t *np)
{
  register int i, j;
  TRIM_PATCH_INFO *trim_patches;
  int new_npatches = 0, k;

  if (np->npatches == 0)
    return 0;

  for (i = 0; i < np->npatches; i++)
    new_npatches += np->trim[i].NumTrimPortions;
  
  trim_patches = (TRIM_PATCH_INFO *) new(new_npatches*
					 sizeof(TRIM_PATCH_INFO));
  
  k = 0;
  for (i = 0; i < np->npatches; i++)
    for (j = 0; j < np->trim[i].NumTrimPortions; j++)
      subdivide_bezier_patch_using_trim_curve(&np->trim[i], j,
					      &trim_patches[k++]);

  
  /* FREE the patches later !! */
#ifdef __FREE_ALL
  for (i = 0; i < np->npatches; i++)
    free_trim_patch(&np->trim[i]);
  myfree(np->trim);
#endif

  np->trim = trim_patches;
  np->npatches = new_npatches;
  
  return 0;
}

int read_patch_primitive(node_t *np, char *filename, int pnum)
{
  FILE *fp;
  register int i, j, k, m, n;
  int iDeg0, iDeg1, l;
  point_typ p;
  TRIM_PORTION_TYPE *tpor;
  int IsAdj;
  
  if ((fp = fopen(filename, "r")) == NULL)
    {
      perror("fopen");
      exit(2);
    }
  
  /* Read in no of patches */
  fscanf(fp, "%d", &np->npatches);
  np->trim = (TRIM_PATCH_INFO *)
    new(np->npatches*sizeof(TRIM_PATCH_INFO));

  for (i = 0; i < np->npatches; i++)
    {
      fscanf(fp, "%d%d", &iDeg0, &iDeg1);

      alloc_tensor_patch(&np->trim[i].tp, iDeg0, iDeg1);
      
      for (j = 0; j <= np->trim[i].tp.iDeg0; j++)
	for (k = 0; k <= np->trim[i].tp.iDeg1; k++)
	  fscanf(fp, "%lf%lf%lf%lf", 
		  &np->trim[i].tp.coeffs[j][k].dCoord[0],
		  &np->trim[i].tp.coeffs[j][k].dCoord[1],
		  &np->trim[i].tp.coeffs[j][k].dCoord[2],
		  &np->trim[i].tp.coeffs[j][k].dCoord[3]);
      
      /* read in number of trim portions, if spline reprn and if adj reprn*/

      fscanf(fp, "%d %d", &np->trim[i].NumTrimPortions, 
	     &np->trim[i].SplineReprn);

      if (!np->trim[i].SplineReprn) {
	for (j = 0; j < np->trim[i].NumTrimPortions; j++)
	  {
	    fscanf(fp, "%d%d",
		   &np->trim[i].TrimPortion[j].PiecePortion.NumOfVertices, 
		   &iDeg0);

	    np->trim[i].TrimPortion[j].PiecePortion.poly_vertex = 
	      (double (*)[2]) new(MAX_VERTEX_COUNT*2*sizeof(double));

	    /* I am commenting out all if (isAdj)'s because there is */
	    /* no adjacency info with piecewise rep right now */
/*
	    if (IsAdj)
	      np->trim[i].TrimPortion[j].PiecePortion.AdjInfo = 
		(int *) new(MAX_VERTEX_COUNT*sizeof(int));
	    else
*/
	    np->trim[i].TrimPortion[j].PiecePortion.AdjInfo = NULL;

	    for (k = 0; 
		 k <= np->trim[i].TrimPortion[j].PiecePortion.NumOfVertices;
		 k++)
	      fscanf(fp, "%lf%lf",
		     &np->trim[i].TrimPortion[j].PiecePortion.poly_vertex[k][0],
		     &np->trim[i].TrimPortion[j].PiecePortion.poly_vertex[k][1]);
	    
	    /* I am doing < and not <= in the loop because adjacency */
	    /* info is segment based */
/*
	    if (IsAdj)
	      for (k = 0; 
		   k < np->trim[i].TrimPortion[j].PiecePortion.NumOfVertices;
		   k++)
		fscanf(fp, "%d",
		       &np->trim[i].TrimPortion[j].PiecePortion.AdjInfo[k]);
	    else {
*/
	    /* Preprocess the trimming ........... */

	    tpor = &np->trim[i].TrimPortion[j];
	    k = 1;
	    while (k < tpor->PiecePortion.NumOfVertices)
	      {
		if (CLOSE(tpor->PiecePortion.poly_vertex[k][0],
			  tpor->PiecePortion.poly_vertex[k-1][0]) &&
		    CLOSE(tpor->PiecePortion.poly_vertex[k][1],
			  tpor->PiecePortion.poly_vertex[k-1][1]))
		  {
		    /* delete pt[k] */
		    for (l = k; l < tpor->PiecePortion.NumOfVertices; l++)
		      {
			tpor->PiecePortion.poly_vertex[l][0] =
			  tpor->PiecePortion.poly_vertex[l+1][0];
			tpor->PiecePortion.poly_vertex[l][1] =
			  tpor->PiecePortion.poly_vertex[l+1][1];
		      }
		    tpor->PiecePortion.NumOfVertices--;
		    k--;
		  }
		
		k++;
	      }
	    /* Now get the point-inside region for each trim-region */
/*
   Commented out because I think we can read it directly from the file
	    find_point_inside_polygon(
				      tpor->PiecePortion.NumOfVertices,
				      tpor->PiecePortion.poly_vertex,
				      &p);
*/	  
	    fscanf(fp, "%lf %lf",
		   &tpor->PiecePortion.ptInsideRegion.u,
		   &tpor->PiecePortion.ptInsideRegion.v);
	  }
      } else {
	for (j = 0; j < np->trim[i].NumTrimPortions; j++)
	  {
	    tpor = &np->trim[i].TrimPortion[j];
	    fscanf(fp, "%d%d",
		   &tpor->SplinePortion.NumOfCurves, 
		   &iDeg0);

	    tpor->SplinePortion.UV_Splines = 
	      (UV_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION * 
				     sizeof(UV_SPLINE_TYPE));

	    tpor->SplinePortion.XYZ_Splines = 
	      (XYZ_SPLINE_TYPE *) new(MAX_SPLINES_PER_TRIM_REGION * 
				     sizeof(XYZ_SPLINE_TYPE));

	    /* if (IsAdj) */
	    tpor->SplinePortion.AdjInfo = 
	      (int *) new(MAX_SPLINES_PER_TRIM_REGION * sizeof(int));
/*
	    else 
	      tpor->SplinePortion.AdjInfo = NULL;
*/

	    for (k = 0; 
		 k < tpor->SplinePortion.NumOfCurves;
		 k++) {
	      fscanf(fp, "%d %d", 
		     &tpor->SplinePortion.UV_Splines[k].Deg,
		     &tpor->SplinePortion.XYZ_Splines[k].Deg);

	      tpor->SplinePortion.UV_Splines[k].CntrlPts =
		(POINT3D *) new((tpor->SplinePortion.UV_Splines[k].Deg+1) *
				sizeof(POINT3D));
	      tpor->SplinePortion.XYZ_Splines[k].CntrlPts =
		(POINT4D *) new((tpor->SplinePortion.XYZ_Splines[k].Deg+1) *
				sizeof(POINT4D));
	      
	      for (m=0; m<=tpor->SplinePortion.UV_Splines[k].Deg; m++)
		for (n=0; n<3; n++)
		  fscanf(fp, "%lf", 
		     &tpor->SplinePortion.UV_Splines[k].CntrlPts[m].dCoord[n]);
	      for (m=0; m<=tpor->SplinePortion.XYZ_Splines[k].Deg; m++)
		for (n=0; n<4; n++)
		  fscanf(fp, "%lf", 
		     &tpor->SplinePortion.XYZ_Splines[k].CntrlPts[m].dCoord[n]);

	      /* if (IsAdj) */
	      fscanf(fp, "%d", &tpor->SplinePortion.AdjInfo[k]);
	    }

	    /* Now get the point-inside region for each trim-region */
	  
/*
	    find_point_inside_polygon(
				      tpor->NumOfVertices,
				      tpor->poly_vertex,
				      &p);
	  
	    tpor->ptInsideRegion.u = p.x;
	    tpor->ptInsideRegion.v = p.y;
*/
	    fscanf(fp, "%lf %lf",
		   &tpor->SplinePortion.ptInsideRegion.u,
		   &tpor->SplinePortion.ptInsideRegion.v);
	  }
      }
      
      copy_ten_p(&np->trim[i].tp, &np->trim[i].orig_tp);
      
      np->trim[i].umin = np->trim[i].vmin = 0.0;
      np->trim[i].umax = np->trim[i].vmax = 1.0;
      np->trim[i].csg_status = IN_CSG;
      np->trim[i].primNum = pnum;
    }

  fclose(fp);
  return np->npatches;
}




#ifdef CTEST

main()
{
  TENSOR_PATCH tp0, tp1;
  double u0, u1, v0, v1;
  int i, j;

  tp0.coeffs = (POINT4D **) malloc(3*sizeof(POINT4D *));
  for (i = 0; i <= 2; i++)
    tp0.coeffs[i] = (POINT4D *) malloc(3*sizeof(POINT4D));

  tp1.coeffs = (POINT4D **) malloc(3*sizeof(POINT4D *));
  for (i = 0; i <= 2; i++)
    tp1.coeffs[i] = (POINT4D *) malloc(3*sizeof(POINT4D));

  for (i = 0; i <= 2; i++)
    for (j = 0; j <= 2; j++)
      scanf("%lf%lf%lf%lf", &tp0.coeffs[i][j].dCoord[0],
	    &tp0.coeffs[i][j].dCoord[1], &tp0.coeffs[i][j].dCoord[2],
	    &tp0.coeffs[i][j].dCoord[3]); 

  tp0.iDeg0 = tp0.iDeg1 = 2;
  tp1.iDeg0 = tp1.iDeg1 = 2;

  u0 = 0.5;
  u1 = 1.0;
  v0 = 0.0;
  v1 = 1.0;

  subdivide_bezier_patch(tp0, u0, u1, v0, v1, &tp1);
}

#endif
