#include <tree.h>
#include <math.h>
#include <macro.h>

#define GLUE_SHIFT_FACTOR 0.02


/* Detect glueing condition. If two planar patches are in identical */
/* planes and they overlap, then displace the right solid by a small */
/* amount in the other direction */

int detect_glue_condition(TRIM_PATCH_INFO *tpil, int nl,
			  TRIM_PATCH_INFO *tpir, int nr)
{
  register int i, j;
  int k, l, m, bflag;
  POINT4D pl, pr;		/* two planes to be compared */
  vector_t v0, v1, v2, vec0l, vec1l, vec0r, vec1r, vn1, vn2;
  TENSOR_PATCH *tpl, *tpr;
  double len, prod;
  
  for (k = 0; k < nl; k++)	/* for every patch in l-solid */
    if ((tpil[k].tp.iDeg0 == 1) && (tpil[k].tp.iDeg1 == 1))
      {
	/* patch is a plane . Get the normalised plane */
	tpl = &(tpil[k].tp);
	yourpt_to_mypt(tpl->coeffs[0][0], v0);
	yourpt_to_mypt(tpl->coeffs[1][0], v1);
	yourpt_to_mypt(tpl->coeffs[0][1], v2);
	
	for (i = 0; i < 3; i++)
	  vec0l[i] = v1[i] - v0[i];
	for (i = 0; i < 3; i++)
	  vec1l[i] = v2[i] - v0[i];

	CROSS(vec0l, vec1l, vn1);
	vn1[3] = -DOT(vn1, v0);

	len = sqrt(DOT(vn1, vn1));
	NORMALISE(vn1, len);

	/* Now we have obtained the normal vector for the first plane */
	/* Look for identical plane in the other solid */

	for (l = 0; l < nr; l++)
	  if ((tpir[l].tp.iDeg0 == 1) && (tpir[l].tp.iDeg1 == 1))
	    {
	      /* patch is a plane . Get the normalised plane */
	      tpr = &(tpir[l].tp);
	      yourpt_to_mypt(tpr->coeffs[0][0], v0);
	      yourpt_to_mypt(tpr->coeffs[1][0], v1);
	      yourpt_to_mypt(tpr->coeffs[0][1], v2);
	      
	      for (i = 0; i < 3; i++)
		vec0r[i] = v1[i] - v0[i];
	      for (i = 0; i < 3; i++)
		vec1r[i] = v2[i] - v0[i];
	      
	      CROSS(vec0r, vec1r, vn2);
	      vn2[3] = -DOT(vn2, v0);
	      
	      len = sqrt(DOT(vn2, vn2));
	      NORMALISE(vn2, len);
	    
	      /* Now compare if the two planes are identical */
	      
	      if (FP_EQUAL(fabs(DOT(vn1, vn2)), 1.0) && 
		  FP_EQUAL(DOT(vn1, v0) + vn1[3], 0.0))
		{
		  /* Check if the patches actually overlap. If so, */
		  /* shift the right solid away slightly in the */
		  /* direction of the normal vector. */
		  
		  if (glue_if_patches_overlap(&tpil[k], vn1, vec0l,
					      &tpir[l]))
		    {
		      /* find which side of the plane to shift the */
		      /* right solid. Choose any other point on the */
		      /* solid and find which side of the plane it */
		      /* lies */
		      prod = 0.0;
		      m = (l == 0) ? 1 : l-1; /* Choose a different */
					      /* patch */
		      bflag = FALSE;
		      for (i = 0; i <= tpir[m].tp.iDeg0; i++)
			{
			  for (j = 0; j <= tpir[m].tp.iDeg1; j++)
			    {
			      yourpt_to_mypt(tpir[m].tp.coeffs[i][j],
					     v0);
			      prod = DOT(v0, vn2) + vn2[3];
			      if (FP_EQUAL(prod, 0.0))
				continue;
			      else
				{
				  bflag = TRUE;
				  break;
				}
			    }
			  
			  if (bflag)
			    break;
			}
		      
		      if (prod > 0)
			glue_shift_solid(tpir, nr, vn2,
					 GLUE_SHIFT_FACTOR); 
		      else
			glue_shift_solid(tpir, nr, vn2,
					 -GLUE_SHIFT_FACTOR);
		      return 0;
		    }
		}
	    } /* inner for-if */
      }	/* outer for-if */
  
  return -1;			/* No glue detected */
}


int glue_shift_solid(TRIM_PATCH_INFO tpir[], int npatches, 
		     vector_t vn, double sfactor)
{
  register int i, j, l;
  int n;
  TENSOR_PATCH *tp, *orig_tp;

  printf("Glue detected: SHifting one solid\n");
  fflush(stdout);
  for (n = 0; n < npatches; n++)
    {
      tp = &tpir[n].tp;
      orig_tp = &tpir[n].orig_tp;
      for (i = 0; i <= tp->iDeg0; i++)
	for (j = 0; j <= tp->iDeg1; j++)
	  for (l = 0; l < 3; l++)
	    {
	      tp->coeffs[i][j].dCoord[l] = 
		(tp->coeffs[i][j].dCoord[l]/tp->coeffs[i][j].dCoord[3] +
		 sfactor*vn[l]) * tp->coeffs[i][j].dCoord[3];

	      orig_tp->coeffs[i][j].dCoord[l] = 
		(orig_tp->coeffs[i][j].dCoord[l]/
		 orig_tp->coeffs[i][j].dCoord[3] + sfactor*vn[l]) *
		   orig_tp->coeffs[i][j].dCoord[3];
	    }
    }
  
  return 0;
}


/* Return True or False depending on whether the planar patches */
/* actually overlap or not */
int glue_if_patches_overlap(TRIM_PATCH_INFO *tpil, vector_t vn1,
			    vector_t vec0l, TRIM_PATCH_INFO *tpir)
{
  register int i, j;
  int k, itype, overlap, count;
  vector_t v0, v1, v0l;
  matrix_t xmat;
  double len;
  point_typ poly1[5], poly2[5], tmp_pt, q0, q1, pp0, pp1;
  double s0, s1, t0, t1;
  
  /* Normalise vec0l and vec0r first */
  len = sqrt(DOT(vec0l, vec0l));
  NORMALISE(vec0l, len);
  
  /* First project the patches so that they lie in the X-Y plane */ 
  
  CROSS(vn1, vec0l, v0l);	/* v is also a unit vector */
  
  memcpy((void *) &xmat[0], (void *) vec0l, 3*sizeof(double));
  memcpy((void *) &xmat[4], (void *) v0l, 3*sizeof(double));
  memcpy((void *) &xmat[8], (void *) vn1, 3*sizeof(double));
  xmat[3] = xmat[7] = xmat[11] = xmat[12] = xmat[13] = xmat[14] = 0.0;
  xmat[15] = 1.0;
  
  
  k = 0;
  for (i = 0; i < 2; i++)
    for (j = 0; j < 2; j++)
      {
	yourpt_to_mypt(tpil->tp.coeffs[j][i], v0);
	matvec_mult(xmat, v0, v1); /* get result in v1 */
	poly1[k].x = v1[0];
	poly1[k++].y = v1[1];
      }
  tmp_pt = poly1[2];
  poly1[2] = poly1[3];
  poly1[3] = tmp_pt;
  poly1[4] = poly1[0];
    
  /* similarly, get the second polygon */
  
  k = 0;
  for (i = 0; i < 2; i++)
    for (j = 0; j < 2; j++)
      {
	yourpt_to_mypt(tpir->tp.coeffs[j][i], v0);
	matvec_mult(xmat, v0, v1); /* get result in v1 */
	poly2[k].x = v1[0];
	poly2[k++].y = v1[1];
      }
  tmp_pt = poly2[2];
  poly2[2] = poly2[3];
  poly2[3] = tmp_pt;
  poly2[4] = poly2[0];
  
  /* At this stage, we have two quadrilaterals in X-Y plane. Check if */
  /* they overlap by using simple intersection tests (avoid */
  /* triangulation .. which might me more reliable) */
  
  overlap = FALSE;
  for (i = 0; i < 4; i++)
    {
      for (j = 0; j < 4; j++)
	if (segments_intersect(&poly1[i], &poly1[i+1], &poly2[j],
			       &poly2[j+1], &itype,
			       &s0, &s1, &t0, &t1, &pp0, &pp1) > 0)
	  {
	    overlap = TRUE;
	    break;
	  }
      if (overlap) break;
    }

  if (overlap)
    return TRUE;
  
  /* Now check for completely inside/outside condition */
  
  q0 = poly1[0];
  q1.x = 1000.0;
  q1.y = 1100.0;
  count = 0;
  for (i = 0; i < 4; i++)
    if (segments_intersect(&q0, &q1, &poly2[i], &poly2[i+1], &itype,
			   &s0, &s1, &t0, &t1, &pp0, &pp1) > 0)
      count++;
  
  if (count%2)			/* odd number of intersections */
    return TRUE;
  
  /* similar check for other polygon */
  
  q0 = poly2[0];
  q1.x = 1000.0;
  q1.y = 1100.0;
  count = 0;
  for (i = 0; i < 4; i++)
    if (segments_intersect(&q0, &q1, &poly1[i], &poly1[i+1], &itype,
			   &s0, &s1, &t0, &t1, &pp0, &pp1) > 0)
      count++;
  
  if (count%2)			/* odd number of intersections */
    return TRUE;
  
  return FALSE;
}


#define NUM_MAX_PATCHES 50
#define NUM_MAX_PRIM    25

/* Check for identical sub-objects in the two solids and eliminate the */
/* common subobject */
int eliminate_identical_subobjects(node_t *nl, node_t *nr)
{
  register int i, j;
  int k, l, pnum, bs_idx0, bs_idx1;
  int pcount;
  TENSOR_PATCH *tp0, *tp1;
  struct {
    int primnum;
    int npatches;
    BOUNDINGBOX bbox;
  } bs[2][NUM_MAX_PRIM];
  int is_scanned[2][NUM_MAX_PATCHES];
  
  for (i = 0 ; i < NUM_MAX_PRIM; i++)
    {
      bs[0][i].primnum = -1;
      bs[1][i].primnum = -1;
    }
  memset(is_scanned, 0, 2*NUM_MAX_PATCHES*sizeof(int));
  bs_idx0 = bs_idx1 = 0;

  /* Compute the bounding boxes for each primitive comprising the */
  /* solid */
  for (i = 0; i < nl->npatches; i++)
    {
      if (is_scanned[0][i])
	continue;
      
      pnum = nl->trim[i].primNum;
      compute_bbox_for_prim(nl, i, is_scanned[0], 
			    &pcount, pnum, &bs[0][bs_idx0].bbox);
      bs[0][bs_idx0].primnum = pnum;
      bs[0][bs_idx0++].npatches = pcount;
    }

  /* do the same thing for the second solid */
  
  for (i = 0; i < nr->npatches; i++)
    {
      if (is_scanned[1][i])
	continue;
      
      pnum = nr->trim[i].primNum;
      compute_bbox_for_prim(nr, i, is_scanned[1], 
			    &pcount, pnum, &bs[1][bs_idx1].bbox);
      bs[1][bs_idx1].primnum = pnum;
      bs[1][bs_idx1++].npatches = pcount;
    }
  
  /* Now check the bs data-structure to detect identical sub-solids */

  for (i = 0; i < bs_idx0; i++)
    for (j = 0; j < bs_idx1; j++)
      if (/* (bs[0][i].npatches == bs[1][j].npatches) && */
	  are_bounding_boxes_same(&bs[0][i].bbox, &bs[1][j].bbox))
	{
	  pnum = bs[1][j].primnum;
	  if (bs[0][i].npatches != bs[1][j].npatches)
	      printf("elim_dup_subobj: Partial overlap detected\n");

	  /* Now remove all the pathces in the second solid which */
	  /* belong to this particular primitive. Shift all the other */
	  /* trim-patch infos as necessary */
	  
	  for (k = 0; k < nr->npatches; k++)
	    if (nr->trim[k].primNum == pnum)
	      {
		for (l = k; l < nr->npatches - 1; l++)
		  nr->trim[l] = nr->trim[l+1];
		nr->npatches--;
		k--;
	      }
	}
  
  return 0;
}
	  	  

int are_bounding_boxes_same(BOUNDINGBOX *b0, BOUNDINGBOX *b1)
{
  return (FP_EQUAL(b0->i3Dbox[0].high, b1->i3Dbox[0].high) && 
	  FP_EQUAL(b0->i3Dbox[0].low, b1->i3Dbox[0].low) && 
	  FP_EQUAL(b0->i3Dbox[1].high, b1->i3Dbox[1].high) && 
	  FP_EQUAL(b0->i3Dbox[1].low, b1->i3Dbox[1].low) && 
	  FP_EQUAL(b0->i3Dbox[2].high, b1->i3Dbox[2].high) && 
	  FP_EQUAL(b0->i3Dbox[2].low, b1->i3Dbox[2].low));
}


int compute_bbox_for_prim(node_t *np, int fpos, int is_scanned[], 
			  int *pcount, int pnum, BOUNDINGBOX *bb)
{
  int m, n;
  int i, j, k;
  double minx, maxx, miny, maxy, minz, maxz;
  double wval;
  TENSOR_PATCH *tp;

  if (np->npatches == 0)
    return 0;

  *pcount = 1;
  tp = &np->trim[fpos].tp;
  minx = maxx = tp->coeffs[0][0].dCoord[0] / tp->coeffs[0][0].dCoord[3];
  miny = maxy = tp->coeffs[0][0].dCoord[1] / tp->coeffs[0][0].dCoord[3]; 
  minz = maxz = tp->coeffs[0][0].dCoord[2] / tp->coeffs[0][0].dCoord[3];
  
  for (k = fpos + 1; k < np->npatches; k++)
    {
      if (np->trim[k].primNum != pnum)
	continue;
      
      is_scanned[k] = TRUE;
      (*pcount)++;
      tp = &np->trim[k].tp;
      m = tp->iDeg0;
      n = tp->iDeg1;
      
      for (i=0; i<=m; i++)
	for (j=0; j<=n; j++)
	  {
	    wval = tp->coeffs[i][j].dCoord[3];
	    if ((tp->coeffs[i][j].dCoord[0]/wval) < minx) 
	      minx = tp->coeffs[i][j].dCoord[0] / wval;
	    if ((tp->coeffs[i][j].dCoord[0]/wval) > maxx) 
	      maxx = tp->coeffs[i][j].dCoord[0] / wval;
	    if ((tp->coeffs[i][j].dCoord[1]/wval) < miny) 
	      miny = tp->coeffs[i][j].dCoord[1] / wval;
	    if ((tp->coeffs[i][j].dCoord[1]/wval) > maxy) 
	      maxy = tp->coeffs[i][j].dCoord[1] / wval;
	    if ((tp->coeffs[i][j].dCoord[2]/wval) < minz) 
	      minz = tp->coeffs[i][j].dCoord[2] / wval;
	    if ((tp->coeffs[i][j].dCoord[2]/wval) > maxz) 
	      maxz = tp->coeffs[i][j].dCoord[2] / wval;
	  }
    }
  
  (bb->i3Dbox[0]).low = minx;
  (bb->i3Dbox[0]).high = maxx;
  (bb->i3Dbox[1]).low = miny;
  (bb->i3Dbox[1]).high = maxy;
  (bb->i3Dbox[2]).low = minz;
  (bb->i3Dbox[2]).high = maxz;  
}

/* Detect the special case where a cylinder/other object is drilling a */
/* hole through the bigger object, and the hoels are exactly the same */
/* size. In this case, wnlongate the drilling object by a small amount */

int detect_cyl_degenerate_overlap(TRIM_PATCH_INFO *tpil, int nl,
			  TRIM_PATCH_INFO *tpir, int nr)
{
  register int i, j, h;
  int k, l, m, bflag;
  POINT4D pl, pr, tmppt;		/* two planes to be compared */
  vector_t v0, v1, v2, vec0l, vec1l, vec0r, vec1r, vn1, vn2;
  TENSOR_PATCH *tpl, *tpr;
  double len, prod, prod_sav;
  vector_t v_sav, tmp_v;
  int first_time;
  
  first_time = TRUE;
  for (k = 0; k < nl; k++)	/* for every patch in l-solid */
    if ((tpil[k].tp.iDeg0 == 1) && (tpil[k].tp.iDeg1 == 1))
      {
	/* patch is a plane . Get the normalised plane */
	tpl = &(tpil[k].tp);
	yourpt_to_mypt(tpl->coeffs[0][0], v0);
	yourpt_to_mypt(tpl->coeffs[1][0], v1);
	yourpt_to_mypt(tpl->coeffs[0][1], v2);
	
	for (i = 0; i < 3; i++)
	  vec0l[i] = v1[i] - v0[i];
	for (i = 0; i < 3; i++)
	  vec1l[i] = v2[i] - v0[i];

	CROSS(vec0l, vec1l, vn1);
	vn1[3] = -DOT(vn1, v0);

	len = sqrt(DOT(vn1, vn1));
	NORMALISE(vn1, len);

	/* Now we have obtained the normal vector for the first plane */
	/* Look for identical plane in the other solid */

	for (l = 0; l < nr; l++)
	  if ((tpir[l].tp.iDeg0 == 1) && (tpir[l].tp.iDeg1 == 1))
	    {
	      /* patch is a plane . Get the normalised plane */
	      tpr = &(tpir[l].tp);
	      yourpt_to_mypt(tpr->coeffs[0][0], v0);
	      yourpt_to_mypt(tpr->coeffs[1][0], v1);
	      yourpt_to_mypt(tpr->coeffs[0][1], v2);
	      
	      for (i = 0; i < 3; i++)
		vec0r[i] = v1[i] - v0[i];
	      for (i = 0; i < 3; i++)
		vec1r[i] = v2[i] - v0[i];
	      
	      CROSS(vec0r, vec1r, vn2);
	      vn2[3] = -DOT(vn2, v0);
	      
	      len = sqrt(DOT(vn2, vn2));
	      NORMALISE(vn2, len);
	    
	      /* Now compare if the two planes are identical */
	      
	      if (FP_EQUAL(fabs(DOT(vn1, vn2)), 1.0) && 
		  FP_EQUAL(DOT(vn1, v0) + vn1[3], 0.0))
		{
		  /* Check if the patches actually overlap. If so, */
		  /* shift the right solid away slightly in the */
		  /* direction of the normal vector. */
		  
		  if (glue_if_patches_overlap(&tpil[k], vn1, vec0l,
					      &tpir[l]))
		    {
		      /* find which side of the plane to shift the */
		      /* right solid. Choose any other point on the */
		      /* solid and find which side of the plane it */
		      /* lies */

		      if (first_time)
			{
			  for (i = 0; i < 4; i++)
			    v_sav[i] = vn2[i];
			  first_time = FALSE;

			  prod = 0.0;
			  m = (l == 0) ? 1 : l-1; /* Choose a different */
			  /* patch */
			  bflag = FALSE;
			  for (i = 0; i <= tpir[m].tp.iDeg0; i++)
			    {
			      for (j = 0; j <= tpir[m].tp.iDeg1; j++)
				{
				  yourpt_to_mypt(tpir[m].tp.coeffs[i][j],
						 v0);
				  prod = DOT(v0, vn2) + vn2[3];
				  if (FP_EQUAL(prod, 0.0))
				    continue;
				  else
				    {
				      bflag = TRUE;
				      break;
				    }
				}
			      
			      if (bflag)
				break;
			    }
			  prod_sav = prod/fabs(prod);
			}
		      else
			if (FP_EQUAL(fabs(DOT(v_sav, vn1)), 1.0))
			  {
			    /* Detected opposing planes which */
			    /* intersect. Enlarge the cylinder */
			    
			       
			    prod = 0.0;
			    m = (l == 0) ? 1 : l-1; /* Choose a different */
			    /* patch */
			    bflag = FALSE;
			    for (i = 0; i <= tpir[m].tp.iDeg0; i++)
			      {
				for (j = 0; j <= tpir[m].tp.iDeg1; j++)
				  {
				    yourpt_to_mypt(tpir[m].tp.coeffs[i][j],
						   v0);
				    prod = DOT(v0, vn2) + vn2[3];
				    if (FP_EQUAL(prod, 0.0))
				      continue;
				    else
				      {
					bflag = TRUE;
					break;
				      }
				  }
				
				if (bflag)
				  break;
			      }
			    
			    prod /= fabs(prod);

			    /* Enlarge the cylinder along the vectors*/
			    
			    for (i = 0; i < nr; i++) /* #patches */
			      for (j = 0; j <= tpir[i].tp.iDeg0; j++)
				for (h = 0; h <= tpir[i].tp.iDeg1; h++)
				  {
				    tmppt = tpir[i].tp.coeffs[j][h];
				    tmppt.dCoord[0] /=
				      tmppt.dCoord[3];
				    tmppt.dCoord[1] /=
				      tmppt.dCoord[3];
				    tmppt.dCoord[2] /=
				      tmppt.dCoord[3]; 
				    
				    yourpt_to_mypt(tmppt, tmp_v);
				    if (FP_EQUAL(DOT(tmp_v, vn1) +
						 vn1[3], 0.0))
				      {
					tmppt.dCoord[0] -=
					  0.5*prod*vn2[0];
					tmppt.dCoord[1] -=
					  0.5*prod*vn2[1];
					tmppt.dCoord[2] -=
					  0.5*prod*vn2[2]; 
					
					tmppt.dCoord[0] *=
					  tmppt.dCoord[3];
					tmppt.dCoord[1] *=
					  tmppt.dCoord[3];
					tmppt.dCoord[2] *=
					  tmppt.dCoord[3];

					tpir[i].tp.coeffs[j][h] = tmppt;
				      }
				    else
				      {
					tmppt.dCoord[0] -=
					  0.5*prod_sav*v_sav[0];
					tmppt.dCoord[1] -=
					  0.5*prod_sav*v_sav[1];
					tmppt.dCoord[2] -=
					  0.5*prod_sav*v_sav[2]; 
					
					tmppt.dCoord[0] *=
					  tmppt.dCoord[3];
					tmppt.dCoord[1] *=
					  tmppt.dCoord[3];
					tmppt.dCoord[2] *=
					  tmppt.dCoord[3];

					tpir[i].tp.coeffs[j][h] = tmppt;
				      }
				  }
			    
			    return 0;
			  }
			else
			  {
			    printf("Too many glues\n");
			    return -1;
			  }
			  
		    }
		}
	    } /* inner for-if */
      }	/* outer for-if */
  
  return -1;			/* No glue detected */
}


/* Return TRUE if the planar patches tpA and tpB lie in the same plane */
int planar_patches_overlap(TRIM_PATCH_INFO *tpA, TRIM_PATCH_INFO *tpB,
			   POINT4D *plA, POINT4D *plB)
{
  register int i, j;
  int k, l, m, bflag;
  POINT4D pl, pr;		/* two planes to be compared */
  vector_t v0, v1, v2, vec0l, vec1l, vec0r, vec1r, vn1, vn2;
  TENSOR_PATCH *tpl, *tpr;
  double len, prod;

  tpl = &tpA->tp;
  yourpt_to_mypt(tpl->coeffs[0][0], v0);
  yourpt_to_mypt(tpl->coeffs[1][0], v1);
  yourpt_to_mypt(tpl->coeffs[0][1], v2);
  
  for (i = 0; i < 3; i++)
    vec0l[i] = v1[i] - v0[i];
  for (i = 0; i < 3; i++)
    vec1l[i] = v2[i] - v0[i];
  
  CROSS(vec0l, vec1l, vn1);
  vn1[3] = -DOT(vn1, v0);
  
  len = sqrt(DOT(vn1, vn1));
  NORMALISE(vn1, len);
  
  /* Now we have obtained the normal vector for the first plane */
  /* Obtain normal vector for other plane */

  tpr = &tpB->tp;
  yourpt_to_mypt(tpr->coeffs[0][0], v0);
  yourpt_to_mypt(tpr->coeffs[1][0], v1);
  yourpt_to_mypt(tpr->coeffs[0][1], v2);
  
  for (i = 0; i < 3; i++)
    vec0r[i] = v1[i] - v0[i];
  for (i = 0; i < 3; i++)
    vec1r[i] = v2[i] - v0[i];
  
  CROSS(vec0r, vec1r, vn2);
  vn2[3] = -DOT(vn2, v0);
  
  len = sqrt(DOT(vn2, vn2));
  NORMALISE(vn2, len);
  
  /* Now compare if the two planes are identical */
  
  mypt_to_yourpt(vn1, (*plA));
  mypt_to_yourpt(vn2, (*plB));
  if (FP_EQUAL(fabs(DOT(vn1, vn2)), 1.0) && 
      FP_EQUAL(DOT(vn1, v0) + vn1[3], 0.0))
    return TRUE;
  else
    return FALSE;
}

