#include <tree.h>
#include <macro.h>
#include <math.h>

#ifdef __LINK

int gen_cone(POINT4D, POINT4D, double, double, double,
	     TENSOR_PATCH **);
int gen_cylinder(POINT4D, POINT4D, double, double, TENSOR_PATCH **);
int gen_sphere(POINT4D, double, TENSOR_PATCH **);
int gen_torus(POINT4D, POINT4D, double, double, TENSOR_PATCH **);
extern void scale_tensor_patch(TENSOR_PATCH *, int);
extern double Scale;

extern int compute_bbox_for_node();
 
#endif

static statistics_t statistics;	/* structure to store some debugging */
				/* information  */

extern int tpatch_idx;		/* global variable to keep a count of */
				/* the patches generated while reading */
				/* arbitrary curves. Must be */
				/* initialised to zero everytime a */
				/* primitive is read */

/* Skip n-lines in the data-file */
static int skip_lines(FILE *infile, int n)
{
  char s[BUFSIZ];
  
  while (n-- >= 0)
    fgets(s, BUFSIZ, infile);

  return n;
}


/* skip all the contents of the file till the string 's' is */
/* encountered */
int skip_lines_until(FILE *infile, char *s)
{
  char buf[BUFSIZ];
  char *s1;

#ifdef READ_STDIN
  fgets(buf, BUFSIZ, infile);

  while (((s1 = strtok(buf, " \t")) == NULL) || (strcmp(s1, s)))
    fgets(buf, BUFSIZ, infile);

#else

  while (1)
    {
      fgets(buf, BUFSIZ, infile);
      s1 = strtok(buf, " \t\n");
      if ((s1 != NULL) && !strcmp(s1, s))
	break;
    }

#endif   
  return 0;
}


/* Read the transformation matrix from the file. Although the */
/* documentation does say that all the objects are modelled at their */
/* finel positions, I seriously doubt it!! */
int read_transformation_matrix(FILE *infile, matrix_t xmat)
{
  register int i, j;
  
  for (j = 0; j < 4; j++)
    for (i = 0; i < 3; i++)
      fscanf(infile, "%lf", &xmat[i*4 + j]);
  xmat[12] = xmat[13] = xmat[14] = 0.0;
  xmat[15] = 1.0;
  
  return 0;
}


/* Read the header for the file. Do some primitive checks */
int read_header(FILE *infile, int *nsol, int *nisosol, int *nstruc)
{
  char s[BUFSIZ];
  int dummy;

  fscanf(infile, "%s", s);
  if (strcmp(s, "nexus"))
    {
      fprintf(stderr, "File not is NEXUS format\n");
      exit(1);
    }

  skip_lines_until(infile, "model:");
  fscanf(infile, "%d%d%d%d\n", nsol, nisosol, nstruc, &dummy);
  
  statistics.nsol += *nsol;
  statistics.nisosol += *nisosol;
  statistics.nstruc += *nstruc;

  return 0;
}


/* read a single primitve */
static int read_primitive(FILE *infile, node_t *np, int pnum)
{
  char s[BUFSIZ];
  int prim_xform, parent_xform, d0, d1;
  primitive_t *p = (primitive_t *) new(sizeof(primitive_t));
  TRIM_PORTION_TYPE *trimP; /* This is passed to gen_<prim> and */
			    /* returned */
  vector_t v1, v2;
  POINT4D pv1, pv2;
  matrix_t xmat, dummy_xmat;

  
  np->prim = p;
  skip_lines_until(infile, "primitive:");
  fscanf(infile, "%d", &p->ptype);

  np->left = np->right = NULL;
  np->nodetype = T_LEAF;

  switch (p->ptype)
    {
    case P_CONE:
      {
	cone_t cone;
	double cone_len;

	/* We do not need the transform matrices at all. So if at all */
	/* they are specified, skip over them */

	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%lf%lf%lf", &cone.c0.x, &cone.c0.y, 
	       &cone.c0.z);
	fscanf(infile, "%lf%lf%lf", &cone.c1.x, &cone.c1.y,
	       &cone.c1.z); 

	cone.c1.x -= cone.c0.x; 
	cone.c1.y -= cone.c0.y;
	cone.c1.z -= cone.c0.z;
	
	fscanf(infile, "%lf%lf%d", &cone.r0, &cone.r1, &d0);
	
#ifdef DEBUG
	printf("\tcone: rin = %lf, rout = %lf\n", cone.r0,
	       cone.r1);
	fflush(stdout);
#endif

#ifdef __LINK
	point_to_4D(cone.c0, v1);
	point_to_4D(cone.c1, v2);
	cone_len = _LENGTH(v2);
	mypt_to_yourpt(v1, pv1);
	mypt_to_yourpt(v2, pv2);
	np->npatches = gen_cone(pv1, pv2, cone_len, cone.r0, cone.r1,
				&np->patch);
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_CONE, 0, 0, 0, pnum);
	compute_bbox_for_node(np);
#endif
	p->ptype = P_CONE;
	break;
      }

    case P_CYLINDER:
      {
	cylinder_t cyl;
	double cyl_len;

	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%lf%lf%lf", &cyl.c0.x, &cyl.c0.y, &cyl.c0.z);
	fscanf(infile, "%lf%lf%lf", &cyl.c1.x, &cyl.c1.y, &cyl.c1.z);

	cyl.c1.x -= cyl.c0.x; 
	cyl.c1.y -= cyl.c0.y;
	cyl.c1.z -= cyl.c0.z;
	
	fscanf(infile, "%lf%lf%d%d", &cyl.r0, &cyl.r1, &d0, &d1);
	
#ifdef DEBUG
	printf("\tcylinder: rin = %lf, rout = %lf\n", cyl.r0,
	       cyl.r1);
	fflush(stdout);
#endif

#ifdef __LINK
	point_to_4D(cyl.c0, v1); /* centre */
	point_to_4D(cyl.c1, v2); /* axis vector */
	cyl_len = _LENGTH(v2);
	mypt_to_yourpt(v1, pv1);
	mypt_to_yourpt(v2, pv2);
	np->npatches = gen_cylinder(pv1, pv2, cyl_len, cyl.r1,
				    &np->patch); 
	calc_scale(np->patch, np->npatches);	
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_CYLINDER, 0, 0, 0, pnum);
	compute_bbox_for_node(np);
#endif
	p->ptype = P_CYLINDER;
	break;
      }

    case P_ISOLATED:
      {
	int totpatches;

#ifdef DEBUG
	printf("\tIsolated-solid:\n");
#endif
	
	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
	
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif
	
	if ((totpatches = read_isolated_solid(np, infile, pnum)) < 0) 
	  {
	    fprintf(stderr, "Iso-solid: read error\n");
	    totpatches = 0;
	  }
	np->npatches = totpatches;
	
#ifdef __LINK
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
	if (parent_xform)
	  fprintf(stderr, "Parent-Xform in Isolated solid: ignored\n");
#endif
	compute_bbox_for_node(np);
#endif
	
	p->ptype = P_ISOLATED;
	break;
      }
      
    case P_PATCH:		/* read trimmed-patch representation */
				/* of the primitive from the file */
      {
	char fname[20];

	fscanf(infile, "%s", fname);
#ifdef DEBUG
	printf("\tpatches from file:: %s\n", fname);
#endif
	read_patch_primitive(np, fname, pnum);
	compute_bbox_for_node(np);
	p->ptype = P_PATCH;

	break;
      }
      
    case P_PIPE:
      {
#ifdef DEBUG
	printf("\tpipe:\n");
#endif

	break;
      }
      
    case P_PPPED:
      {
	ppped_t pp;
	register int i;
	double fdum;
	int totpatches, sense;
	extern TENSOR_PATCH tpatch[];
	
#ifdef DEBUG
	printf("\tparallelopiped:\n");
#endif
	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%lf%lf%lf", &pp.c0.x, &pp.c0.y, &pp.c0.z);
	fscanf(infile, "%lf%lf%lf", &pp.v0[0], &pp.v0[1], &pp.v0[2]);
	fscanf(infile, "%lf%lf%lf", &pp.v1[0], &pp.v1[1], &pp.v1[2]);
	fscanf(infile, "%lf%lf%lf", &pp.v2[0], &pp.v2[1], &pp.v2[2]);
	fscanf(infile, "%lf%lf%lf", &pp.len0, &pp.len1, &pp.len2);
	
	tpatch_idx = 0;
	sense = generate_patches_for_ppped(&pp);
	totpatches = 6;

	np->patch = (TENSOR_PATCH *) new(totpatches *
					 sizeof(TENSOR_PATCH));
	memcpy((void *) np->patch, (void *) tpatch,
	       totpatches * sizeof(TENSOR_PATCH));
	np->npatches = totpatches;

#ifdef __LINK
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_PPPED, sense, 0, 0, pnum);
	compute_bbox_for_node(np);
#endif
	
	p->ptype = P_PPPED;
	break;
      }

    case P_PLANE:
      {
	vector_t v0, v1;
	primplane_t *pp;

#ifdef DEBUG
	printf("\tplane:\n");
#endif
	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	pp = (primplane_t *) new(sizeof(primplane_t));
	read_plane(infile, &pp->pl, pp->v0, pp->v1);
	
	/* The plane is treated specially. Save the entire */
	/* data-structure for future use later when traversing the */
	/* tree */
	np->npatches = 0;
	np->trim = NULL;
	np->patch = NULL;

	p->u.prpl_p = (primplane_t *) pp;
	p->ptype = P_PLANE;
	printf("P_PLANE: Dont know what exact format is\n");
	
	break;
      }

    case P_PRISM:
      {
	prism_t psm;
	vector_t pl1_v0, pl1_v1, pl2_v0, pl2_v1;
	int i, dtype, npatches, totpatches = 0, j, nl, nl1, nl2;
	int *ndomcurves;
	double dummy;
	extern TENSOR_PATCH tpatch[];
#ifdef DEBUG
	printf("\tprism:\n");
	fflush(stdout);
#endif
	
	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%d%d", &psm.ncrv, &psm.ndomains);
	ndomcurves = (int *) new(psm.ndomains*sizeof(int));
	for (i = 0; i < psm.ndomains; i++)
	  fscanf(infile, "%d", &ndomcurves[i]);

	fscanf(infile, "%lf%lf%lf", &psm.v[0], &psm.v[1], &psm.v[2]);
	read_plane(infile, &psm.pl1, pl1_v0, pl1_v1);
	read_plane(infile, &psm.pl2, pl2_v0, pl2_v1);
	fscanf(infile, "%lf", &dummy);
	
	if ((psm.ndomains > 1) && (ndomcurves[1] > 1))
	  {
	    np->npatches = 0;
	    break;
	  }

	/* For all the curves, obtain the patches and store them in */
	/* the global tpatch table */
	dtype = DOM_OUTER;
	tpatch_idx = 0;
	for (i = 0; i < psm.ndomains; i++) /* for every domain */
	  for (j = 0; j < ndomcurves[i]; j++) /* for every curve */
	    {
	      npatches = read_curve(infile);
	      totpatches += npatches;
	      dtype = DOM_INNER;
	    }

#ifdef TRY_PARENT_XFORM
	if (parent_xform)
	  transform_patches2(tpatch, totpatches, dummy_xmat);
#endif

	/* Now project all the control points of the curves in the */
	/* appropriate direction and obtain the true patches */
	for (i = 0; i < totpatches; i++)
	    project_control_points(i, psm.v, &psm.pl1, &psm.pl2);

	/* Now generate the top and bottom patches for the prism */
	if (psm.ndomains > 1)
	  {
	    ndomcurves[0] = totpatches - 4;
	    ndomcurves[1] = 4;
	  }
	else
	  {
	    ndomcurves[0] = totpatches;
	    ndomcurves[1] = 0;
	  }
	nl = generate_top_and_bottom_lids_for_prism(np, pl1_v0, pl1_v1,
					       psm.ndomains,
					       ndomcurves, &nl1, &nl2);

	totpatches += 2;
	np->patch = (TENSOR_PATCH *) new(totpatches *
					 sizeof(TENSOR_PATCH));
	memcpy((void *) np->patch, (void *) tpatch,
	       totpatches * sizeof(TENSOR_PATCH));
	np->npatches = totpatches;

#ifdef __LINK
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_PRISM, psm.ndomains, nl1, nl2, pnum);
#endif
	/* Invoke patch-subdivision to gererate two patches incase we */
	/* have broken the loop at the top of the prism */
	if (psm.ndomains > 1)
	  subdivide_trimmed_patches(np);

	compute_bbox_for_node(np);

	p->ptype = P_PRISM;
	break;
      }

    case P_PYRAMID:
      {
	pyramid_t pyr;
	double fdum;
	int idum, i, npatches, totpatches;
	extern TENSOR_PATCH tpatch[];

#ifdef DEBUG
	printf("\tpyramid:\n");
	fflush(stdout);
#endif
	
	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%lf%lf%lf", &pyr.v[0], &pyr.v[1], &pyr.v[2]);
	fscanf(infile, "%lf%d", &fdum, &pyr.ncrv);
	
	tpatch_idx = totpatches = 0;
	for (i = 0; i < pyr.ncrv; i++)
	  {
	    npatches = read_curve(infile);
	    totpatches += npatches;
	  }
	
	for (i = 0; i < totpatches; i++)
	  generate_pyramid_control_points(i, pyr.v);

	np->patch = (TENSOR_PATCH *) new(totpatches *
					 sizeof(TENSOR_PATCH));
	memcpy((void *) np->patch, (void *) tpatch,
	       totpatches * sizeof(TENSOR_PATCH));
	np->npatches = totpatches;

#ifdef __LINK
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_PYRAMID, 0, 0, 0, pnum);
	compute_bbox_for_node(np);
#endif

	p->ptype = P_PYRAMID;
	break;
      }
      
      case P_SPHERE:
      {
	sphere_t sph;

	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%lf%lf%lf", &sph.c.x, &sph.c.y, &sph.c.z);
	fscanf(infile, "%lf%lf%d%d", &sph.rin, &sph.rout, &d0, &d1);
	
#ifdef DEBUG
	printf("\tsphere: rin = %lf, rout = %lf\n", sph.rin,
	       sph.rout);
	fflush(stdout);
#endif
	
#ifdef __LINK
	point_to_4D(sph.c, v1);
	mypt_to_yourpt(v1, pv1);
	np->npatches = gen_sphere(pv1, sph.rout, &np->patch);
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_SPHERE, 0, 0, 0, pnum);
	compute_bbox_for_node(np);
#endif
	
	p->ptype = P_SPHERE;
	
	break;
      }
      
    case P_TORUS:
      {
	torus_t tor;
	
	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%lf%lf%lf", &tor.c.x, &tor.c.y, &tor.c.z);
	fscanf(infile, "%lf%lf%lf", &tor.v[0], &tor.v[1], &tor.v[2]);
	tor.v[3] = 0.0;

	fscanf(infile, "%lf%lf%d%d", &tor.rg, &tor.rs, &d0, &d1);
	
#ifdef DEBUG
	printf("\ttorus: rsmaller = %lf, rgreater = %lf\n", tor.rs,
	       tor.rg);
	fflush(stdout);
#endif

#ifdef __LINK
	point_to_4D(tor.c, v1);
	mypt_to_yourpt(v1, pv1);
	mypt_to_yourpt(tor.v, pv2);
	np->npatches = gen_torus(pv1, pv2, tor.rg - tor.rs,
				 tor.rg + tor.rs, &np->patch);
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_TORUS, 0, 0, 0, pnum);
	compute_bbox_for_node(np);
#endif
	p->ptype = P_TORUS;
	break;
      }

    case P_VASE:
      {
	vase_t vs;
	double fdum;
	int idum, i, npatches, totpatches;
	extern TENSOR_PATCH tpatch[];

#ifdef DEBUG
	printf("\tvase:\n");
#endif

	fscanf(infile, "%lf%lf%lf%d%d", &p->r, &p->g, &p->b, 
	       &prim_xform, &parent_xform);
#ifdef WEIRD
	if (prim_xform)
	  read_transformation_matrix(infile, xmat);
	if (parent_xform)
	  read_transformation_matrix(infile, dummy_xmat);
#else	
	skip_lines(infile, (prim_xform + parent_xform)*4); 
#endif

	fscanf(infile, "%d%d", &vs.ncrv, &idum);
	fscanf(infile, "%lf%lf%lf", &vs.pt1[0], &vs.pt1[1], &vs.pt1[2]);
	fscanf(infile, "%lf%lf%lf", &vs.v[0], &vs.v[1], &vs.v[2]);
	fscanf(infile, "%lf%lf%lf", &vs.a0, &vs.a1, &fdum);

	
	tpatch_idx = totpatches = 0;
	for (i = 0; i < vs.ncrv; i++)
	  {
	    npatches = read_curve(infile);
	    totpatches += npatches;
	  }

#ifdef TRY_PARENT_XFORM
	if (parent_xform)
	  transform_patches2(tpatch, totpatches, dummy_xmat);
#endif

	totpatches = generate_vase_representation(&vs, totpatches);
	np->patch = (TENSOR_PATCH *) new(totpatches *
					 sizeof(TENSOR_PATCH));
	memcpy((void *) np->patch, (void *) tpatch,
	       totpatches * sizeof(TENSOR_PATCH));
	
	np->npatches = totpatches;

#ifdef __LINK
	calc_scale(np->patch, np->npatches);
#ifdef WEIRD
	if (prim_xform)
	  transform_patches(np, xmat);
#endif
	add_trim_patch_info(np, P_VASE, 0, 0, 0, pnum);
	compute_bbox_for_node(np);
#endif

	p->ptype = P_VASE;	
	break;
      }

    default:
      printf("\tdefault prim\n");
      break;
    }
  
  return 0;
}

/* Read one solid from the input file and construct the CSG */
/* represenation for it. Return the root of the CSG tree */

node_t *read_solid(FILE *infile)
{
  char s[BUFSIZ];
  int i, histlen, op, op1, op2;
  solinfo_t *si;
  node_t *nodetab;
  int offset;

  skip_lines_until(infile, "solid:");
  
  si = &(statistics.currinfo);

#ifndef _MIPS
  si->name = strdup(fgets(s, BUFSIZ, infile));
#else
  fgets(s, BUFSIZ, infile);
  si->name = (char *) new(strlen(s) + 1);
  strcpy(si->name, s);
#endif

  fscanf(infile, "%d%d%lf%lf%lf", &histlen, &si->nprim,
	 &si->r, &si->g, &si->b); 

#ifdef DEBUG
  printf("read_solid: histlen = %d, nprim = %d\n", histlen, si->nprim);
#endif
  
  /* Allocte enough space to hold the entire CSG tree. This is equal */
  /* to the number of primitives + the number of lines in the history */
  /* table (each lines corrensponds to one intermediate node in the */
  /* tree */ 
  nodetab = (node_t *) new((histlen + si->nprim + 2)*sizeof(node_t)); 
  
  /* Read in all the primitives first */
  for (i = 1; i <= si->nprim; i++)
    {
      read_primitive(infile, &nodetab[i], i);
      preprocess_solid(&nodetab[i]);

#ifndef VALIDATE_PRIMITIVE
      switch (nodetab[i].prim->ptype)
	{
	case P_PRISM:
	case P_PPPED:
	  validate_solid(&nodetab[i]);
	  break;
	}
#endif
       /* dump_patches(stdout, &nodetab[i]); */
    }
  
  /* Scale all the patches by Scale */

#ifdef __LINK 
#ifdef _SCALE
  for (i = 1; i <= si->nprim; i++)
    scale_tensor_patch(nodetab[i].patch, nodetab[i].npatches);
#else 
  Scale = 1.0;
#endif
#endif

  /* Read in the history table and construct the CSG tree */
 
  offset = si->nprim;
  skip_lines_until(infile, "history:");
  for (i = offset + 1; i < histlen + offset + 1; i++)
    {
      fscanf(infile, "%d%d%d", &op, &op1, &op2);
      
      if (op == 0)
	{
	  if (histlen == 1)
	    {
	      nodetab[op1].op = op;
	      nodetab[op1].nodetype = T_LEAF;
	      nodetab[op1].info = (solinfo_t *) new(sizeof(solinfo_t)); 
	      memcpy((void *) nodetab[op1].info, (void *) &statistics.currinfo, 
		     sizeof(solinfo_t)); 
	      return &nodetab[op1];
	    }
	  else			/* Error */
	    {
	      fprintf(stderr, "\tInvalid history table: opcode 0\n");
	      exit(1);
	    }
	}
      
      nodetab[i].op = op;
      nodetab[i].nodetype = T_NONLEAF;
      if (op1 > 0) 
	nodetab[i].left = &nodetab[op1];
      else
	nodetab[i].left = &nodetab[offset - op1];
      
      if (op2 > 0) 
	nodetab[i].right = &nodetab[op2];
      else
	nodetab[i].right = &nodetab[offset - op2];
      
    }
  
  /* return the root of the tree */
  nodetab[histlen + offset].info = (solinfo_t *)
    new(sizeof(solinfo_t)); 
  memcpy((void *) nodetab[histlen + offset].info, (void *) &statistics.currinfo,
	 sizeof(solinfo_t)); 
  return &nodetab[histlen + offset];
}  







