#include <math.h>
#include <tree.h>
#include <macro.h>

TENSOR_PATCH tpatch[MAX_TPNUM];	/* global table used to temporarily */
				/* store al the generated patches */
				/* which can then be stored in the */
				/* node for the primitive */

int tpatch_idx;			/* global variable. must be */
				/* initialised to zero everytime a */
				/* primitive is read */

int top_is_zero;		/* Boolean flag to store which row */
				/* represents the top of the prism */

/* vdeg is always 1. udeg is varying. Hence, allocate a 2D array of */
/* size udeg X 2 in the patch */
int get_new_patch(int udeg, int vdeg)
{
  int i = tpatch_idx, j;
  
  tpatch[i].iDeg0 = udeg;
  tpatch[i].iDeg1 = vdeg;

  tpatch[i].coeffs = (POINT4D **) new((udeg + 1)*sizeof(POINT4D *));
  for (j = 0; j < udeg + 1; j++)
    tpatch[i].coeffs[j] = (POINT4D *) new((vdeg + 1)*sizeof(POINT4D));

  return tpatch_idx++;
}


/* Read a plane. Return the four coeffs. of the plane in *p. v0 and v1 */
/* are also returned to be used later to determine the co-od system */
/* for the circle, ellipse, etc. */
int read_plane(FILE *infile, plane_typ *p, vector_t v0, vector_t v1)
{
  vector_t pt;

  fscanf(infile, "%lf%lf%lf", &pt[0], &v0[0], &v1[0]);
  fscanf(infile, "%lf%lf%lf", &pt[1], &v0[1], &v1[1]);
  fscanf(infile, "%lf%lf%lf", &pt[2], &v0[2], &v1[2]);
  
  v0[3] = v1[3] = 1.0;
  CROSS(v0, v1, p->coeff);
  p->coeff[3] = -1.0*DOT(p->coeff, pt);
  
  return 0;
}


/* read a line from the file. Return the two endpoints of the line in */
/* v0 and v1 respectively */
static int read_line(FILE *infile, vector_t v0, vector_t v1)
{
  line_t l;

  fscanf(infile, "%lf%lf", &l.u0, &l.u1);
  fscanf(infile, "%lf%lf%lf", &l.v[0], &l.v[1], &l.v[2]);
  fscanf(infile, "%lf%lf%lf", &l.o[0], &l.o[1], &l.o[2]);

  v0[0] = l.o[0] + l.u0 * l.v[0];
  v0[1] = l.o[1] + l.u0 * l.v[1];
  v0[2] = l.o[2] + l.u0 * l.v[2];

  v1[0] = l.o[0] + l.u1 * l.v[0];
  v1[1] = l.o[1] + l.u1 * l.v[1];
  v1[2] = l.o[2] + l.u1 * l.v[2];
}


/* whenever the quadrant of the circle is subdivided, calculate the */
/* c-value from the angle (which lies between zero and 90). Basically, */
/* solve the quadriatic equation in t using the fact that the point */
/* (r*c(theta), r*s(theta)) lies on the circle */
static double calculate_c_for_circle(double ang)
{
  double theta, s;
  double del, retval;
  
  if (ang == 90.0)
    return 1.0;

  theta = ang*M_PI/180.0;
  s = 1.0/cos(theta);
  
  del = 2*(s - 1)*(s - 1) - 4*(s + sqrt(2.0))*(s - 1)/(1 - sqrt(2.0));
  del = sqrt(del);
  
  retval = (-1.0*(s -1)*sqrt(2.0) + del)/2.0/(s + sqrt(2.0));

  return retval;
}


/* Perform the similar procedure for the ellipse. But none of the data */
/* in the torpedo model seems to require sub-division */
static double calculate_c_for_ellipse(double ang, double w)
{
  double theta, s, p, q;
  double del, retval, a, b, c;
  
  if (ang == 90.0)
    return 1.0;

  theta = ang*M_PI/180.0;
  p = 1.0 - cos(theta);
  q = 1 - w;

  a = 1 - 2*p*q;
  b = 2*p*q;
  c = -p;

  retval = (-b + sqrt(b*b - 4*a*c))/2/a;

  return retval;
}

/* Get weight for the middle control point for the ellipse. Turns out */
/* to be 1/sqrt(2) everytime. The method employed is fairly general, */
/* and should work for ANY conic. */
/* (R.C. Beech -- Intro to curves and surfaces of CAD */
static double get_weight_for_ellipse_control_point(double rmaj, 
						   double rmin)
{
  double theta, beta, t, x;
  double tmprt, wt;
  
  theta = M_PI/4.0;
  
  t = (1 - cos(theta))/(2 - sin(theta) - cos(theta));
  
  x = rmaj * (1 - t);
  tmprt = sqrt((rmaj/x) - 1.0);

  t = tmprt/(1 + tmprt);
  beta = (2*(1 - t)*t)/((1 - t)*(1 - t) + t*t);
  
  wt = (rmaj*cos(theta) - x)/(beta*rmaj*(1 - cos(theta)));

  return wt;
}
  

/* Rotate the first row of control points in each of the quadriatic */
/* tensor patch by angle 'ang'. */
static int rotate_control_points(TENSOR_PATCH *tp, double ang)
{
  double theta, x, y;
  int i;

  if (ang == 0.0)
    return 0;

  theta = ang*M_PI/180.0;
  for (i = 0; i < 3; i++)
    {
      x = tp->coeffs[i][0].dCoord[0];
      y = tp->coeffs[i][0].dCoord[1];
      
      tp->coeffs[i][0].dCoord[0] = x*cos(theta) - y*sin(theta);
      tp->coeffs[i][0].dCoord[1] = x*sin(theta) + y*cos(theta);
    }
  
  return 0;
}


int matvec_mult(matrix_t xmat, vector_t v, vector_t vv)
{
  register int i, j, k;

  for (i = 0; i < 4; i++)
    {
      vv[i] = 0.0;
      for (j = 0; j < 4; j++)
	vv[i] += xmat[i*4 + j]*v[j];
    }

  return 0;
}


int transpose_mat(matrix_t xmat)
{
  register int i, j;
  double temp;
  
  for (i = 0; i < 4; i++)
    for (j = i+1; j < 4; j++)
      {
	temp = xmat[i*4 + j];
	xmat[i*4 + j] = xmat[j*4 + i];
	xmat[j*4 + i] = temp;
      }
  
  return 0;
}


/* Generate the required no. of patches for the circular arc. (1-4) */
/* The two vectors v0 and v1 provide the frame of reference */
static int generate_patches_for_circle(circle_t *cc, vector_t v0,
				vector_t v1)
{
  int tpnum, i, npatches, j;
  TENSOR_PATCH *tp;
  double old_a0, old_a1, a0, a1, rad;
  double ang, amax, c, wt;
  vector_t b[3], v, vv;
  matrix_t xmat_tr, xmat_rot;

  npatches = 0;
  if (cc->a1 > 360.0)
    {				/* process the patches separately if */
				/* a1 is greater than 360*/
      old_a1 = cc->a1;
      cc->a1 = 360.0;
      npatches = generate_patches_for_circle(cc, v0, v1);
      cc->a0 = 0.0;
      cc->a1 = old_a1 - 360.0;
      npatches += generate_patches_for_circle(cc, v0, v1);
    }
  else
    {
      a1 = cc->a1 - cc->a0;	/* first generate an arc from 0 to (a1 */
				/* - a0). Then rotate the control */
				/* points for this by a0. Done to */
				/* avoid two subdivisions in the */
				/* middle of the circle */
      a0 = 0.0;
      ang = 0.0;
      rad = cc->r;

      while (ang + 0.5 < a1)	/* 0.5 ::: tolerance */
	{
	  tpnum = get_new_patch(2, 1);
	  npatches++;
	  tp = &tpatch[tpnum];
	  
	  /* Now generate the three control points describing the */
	  /* 90 degree arc of the circle from 0-90. Subdivide if */
	  /* necessary. Then rotate the conrtol points by 'ang' */
	  tp->coeffs[0][0].dCoord[0] = rad;
	  tp->coeffs[0][0].dCoord[1] = 0.0;
	  tp->coeffs[0][0].dCoord[2] = 0.0;
	  tp->coeffs[0][0].dCoord[3] = 1.0;
	  
	  tp->coeffs[1][0].dCoord[0] = rad*1.0/sqrt(2.0);
	  tp->coeffs[1][0].dCoord[1] = rad*1.0/sqrt(2.0);
	  tp->coeffs[1][0].dCoord[2] = 0.0;
	  tp->coeffs[1][0].dCoord[3] = 1.0/sqrt(2.0);
	  
	  tp->coeffs[2][0].dCoord[0] = 0.0;
	  tp->coeffs[2][0].dCoord[1] = rad;
	  tp->coeffs[2][0].dCoord[2] = 0.0;
	  tp->coeffs[2][0].dCoord[3] = 1.0;
	  
	  /* Now check if we need to sub-divide the curve */
	  if (a1 >= (ang + 90.0))
	    ;
	  else			/* subdivide */
	    {
	      for (i = 0; i < 3; i++)
		{
		  yourpt_to_mypt(tp->coeffs[i][0], b[i]);
		}
	      
	      /* Bezier curve subdivision. Use De Casteljau algorithm */
	      /* to obtain the new points (Farin) */
	      c = calculate_c_for_circle(a1 - ang);
	      for (i = 1; i <= 2; i++)
		{
		  for (j = 0; j <= 2 - i; j++)
		    {
		      b[j][0] = (1 - c)*b[j][0] + c*b[j + 1][0];
		      b[j][1] = (1 - c)*b[j][1] + c*b[j + 1][1];
		      b[j][2] = (1 - c)*b[j][2] + c*b[j + 1][2];
		      b[j][3] = (1 - c)*b[j][3] + c*b[j + 1][3]; 
		    }
		  mypt_to_yourpt_no_w(b[0], tp->coeffs[i][0]);
		  tp->coeffs[i][0].dCoord[3] = b[0][3];
		}
	    }

	  rotate_control_points(tp, ang);

	  ang += 90.0;
	} /* end-while */

      /* Now all the patches have benn generated. Transform the coeff. */
      /* properly using the vectors v0, v1, and the plane eqn. Before */
      /* that, apply the additional rotation == cc->a0 */

      for (i = tpnum - npatches + 1; i <=tpnum; i++)
	rotate_control_points(&tpatch[i], cc->a0);
      

      for (i = 0; i < 16; i++)
	xmat_tr[i] = 0.0;
      xmat_tr[3] = cc->c[0];	/* Translation first. Translate the */
				/* circle to the new centre */
      xmat_tr[7] = cc->c[1];
      xmat_tr[11] = cc->c[2];
      xmat_tr[0] = xmat_tr[5] = xmat_tr[10] = xmat_tr[15] = 1.0;

      /* Rotation next so that (X, Y, Z) becomes coincident with */
      /* (v0, v1, cross(v0, v1)). Since v0 and v1 are orthonormal, we */
      /* do not have to normalise their cross-product */

      xmat_rot[0] = v0[0];
      xmat_rot[4] = v0[1];	/* x-axis */
      xmat_rot[8] = v0[2];

      xmat_rot[1] = v1[0];
      xmat_rot[5] = v1[1];	/* y-axis */
      xmat_rot[9] = v1[2];

      CROSS(v0, v1, v);
      
      xmat_rot[2] = v[0];
      xmat_rot[6] = v[1];	/* z-axis */
      xmat_rot[10] = v[2];

      xmat_rot[3] = xmat_rot[7] = xmat_rot[11] = 0.0;
      xmat_rot[12] = xmat_rot[13] = xmat_rot[14] = 0.0;
      xmat_rot[15] = 1.0;


      for (i = tpnum - npatches + 1; i <=tpnum; i++)
	{
	  tp =&tpatch[i];
	  for (j = 0; j < 3; j++)
	    {
	      yourpt_to_mypt(tp->coeffs[j][0], v);
	      matvec_mult(xmat_rot, v, vv);
	      matvec_mult(xmat_tr, vv, v);
	      mypt_to_yourpt_no_w(v, tp->coeffs[j][0]);
	      tp->coeffs[j][0].dCoord[3] = v[3];
	    }
	}
    }
  
  return npatches;
}


/* Similar to generating patches fot the circle */
static int generate_patches_for_ellipse(ellipse_t *el, vector_t v0,
				vector_t v1)
{
  int tpnum, i, npatches, j;
  TENSOR_PATCH *tp;
  double old_a0, old_a1, a0, a1, rmaj, rmin, wt;
  double ang, amax, c;
  vector_t b[3], v, vv;
  matrix_t xmat_tr, xmat_rot;

  npatches = 0;
  if (el->a1 > 360.0)
    {				/* process the patches separately */
      old_a1 = el->a1;
      el->a1 = 360.0;
      npatches = generate_patches_for_ellipse(el, v0, v1);
      el->a0 = 0.0;
      el->a1 = old_a1 - 360.0;
      npatches += generate_patches_for_ellipse(el, v0, v1);
    }
  else
    {
      a1 = el->a1 - el->a0;
      a0 = 0.0;
      ang = 0.0;
      rmaj = el->rmaj;
      rmin = el->rmin;

      while (ang < a1)
	{
	  tpnum = get_new_patch(2, 1);
	  npatches++;
	  tp = &tpatch[tpnum];

	  if (ang == 0.0)
	    {
	      /* Now generate the three control points describing the */
	      /* 90 degree arc of the circle from 0-90 */
	      tp->coeffs[0][0].dCoord[0] = rmaj;
	      tp->coeffs[0][0].dCoord[1] = 0.0;
	      tp->coeffs[0][0].dCoord[2] = 0.0;
	      tp->coeffs[0][0].dCoord[3] = 1.0;
	      
	      wt = get_weight_for_ellipse_control_point(el->rmaj,
							el->rmin);
	      
	      tp->coeffs[1][0].dCoord[0] = rmaj * wt;
	      tp->coeffs[1][0].dCoord[1] = rmin * wt;
	      tp->coeffs[1][0].dCoord[2] = 0.0;
	      tp->coeffs[1][0].dCoord[3] = wt;
	      
	      tp->coeffs[2][0].dCoord[0] = 0.0;
	      tp->coeffs[2][0].dCoord[1] = rmin;
	      tp->coeffs[2][0].dCoord[2] = 0.0;
	      tp->coeffs[2][0].dCoord[3] = 1.0;
	    }
	  else if (ang == 90.0)
	    {
	      tp->coeffs[0][0].dCoord[0] = 0.0;
	      tp->coeffs[0][0].dCoord[1] = rmin;
	      tp->coeffs[0][0].dCoord[2] = 0.0;
	      tp->coeffs[0][0].dCoord[3] = 1.0;
	      
	      wt = get_weight_for_ellipse_control_point(el->rmaj,
							el->rmin);
	      
	      tp->coeffs[1][0].dCoord[0] = -1.0 * rmaj * wt;
	      tp->coeffs[1][0].dCoord[1] = rmin * wt;
	      tp->coeffs[1][0].dCoord[2] = 0.0;
	      tp->coeffs[1][0].dCoord[3] = wt;
	      
	      tp->coeffs[2][0].dCoord[0] = -1.0*rmaj;
	      tp->coeffs[2][0].dCoord[1] = 0.0;
	      tp->coeffs[2][0].dCoord[2] = 0.0;
	      tp->coeffs[2][0].dCoord[3] = 1.0;
	    }	    
	  else if (ang == 180.0)
	    {
	      tp->coeffs[0][0].dCoord[0] = -1.0 * rmaj;
	      tp->coeffs[0][0].dCoord[1] = 0.0;
	      tp->coeffs[0][0].dCoord[2] = 0.0;
	      tp->coeffs[0][0].dCoord[3] = 1.0;
	      
	      wt = get_weight_for_ellipse_control_point(el->rmaj,
							el->rmin);
	      
	      tp->coeffs[1][0].dCoord[0] = -1.0 * rmaj * wt;
	      tp->coeffs[1][0].dCoord[1] = -1.0 * rmin * wt;
	      tp->coeffs[1][0].dCoord[2] = 0.0;
	      tp->coeffs[1][0].dCoord[3] = wt;
	      
	      tp->coeffs[2][0].dCoord[0] = 0.0;
	      tp->coeffs[2][0].dCoord[1] = -1.0 * rmin;
	      tp->coeffs[2][0].dCoord[2] = 0.0;
	      tp->coeffs[2][0].dCoord[3] = 1.0;
	    }	    
	  else if (ang == 270.0)
	    {
	      tp->coeffs[0][0].dCoord[0] = 0.0;
	      tp->coeffs[0][0].dCoord[1] = -1.0 * rmin;
	      tp->coeffs[0][0].dCoord[2] = 0.0;
	      tp->coeffs[0][0].dCoord[3] = 1.0;
	      
	      wt = get_weight_for_ellipse_control_point(el->rmaj,
							el->rmin);
	      
	      tp->coeffs[1][0].dCoord[0] = rmaj * wt;
	      tp->coeffs[1][0].dCoord[1] = -1.0 * rmin * wt;
	      tp->coeffs[1][0].dCoord[2] = 0.0;
	      tp->coeffs[1][0].dCoord[3] = wt;
	      
	      tp->coeffs[2][0].dCoord[0] = rmaj;
	      tp->coeffs[2][0].dCoord[1] = 0.0;
	      tp->coeffs[2][0].dCoord[2] = 0.0;
	      tp->coeffs[2][0].dCoord[3] = 1.0;
	    }	    
	  
	  
	  /* Now check if we need to sub-divide the curve */
	  if (a1 >= (ang + 90.0))
	    ;
	  else			/* subdivide */
	    {
	      for (i = 0; i < 3; i++)
		{
		  yourpt_to_mypt(tp->coeffs[i][0], b[i]);
		}
	      
	      /* Bezier curve subdivision. (Farin) */
	      c = calculate_c_for_ellipse(a1 - ang, wt);
	      for (i = 1; i <= 2; i++)
		{
		  for (j = 0; j <= 2 - i; j++)
		    {
		      b[j][0] = (1 - c)*b[j][0] + c*b[j + 1][0];
		      b[j][1] = (1 - c)*b[j][1] + c*b[j + 1][1];
		      b[j][2] = (1 - c)*b[j][2] + c*b[j + 1][2];
		      b[j][3] = (1 - c)*b[j][3] + c*b[j + 1][3]; 
		    }
		  mypt_to_yourpt_no_w(b[0], tp->coeffs[i][0]);
		  tp->coeffs[i][0].dCoord[3] = b[0][3];
		}
	    }

	  ang += 90.0;
	} /* end-while */

      /* Now all the patches have benn generated. Transform the coeff. */
      /* properly using the vectors v0, v1, and the plane eqn. Before */
      /* that, apply the additional rotation == el->a0  + el->ang*/

      for (i = tpnum - npatches + 1; i <=tpnum; i++)
	rotate_control_points(&tpatch[i], el->a0 + el->ang);
      
      for (i = 0; i < 16; i++)
	xmat_tr[i] = 0.0;
      xmat_tr[3] = el->c[0];	/* Translation first */
      xmat_tr[7] = el->c[1];
      xmat_tr[11] = el->c[2];
      xmat_tr[0] = xmat_tr[5] = xmat_tr[10] = xmat_tr[15] = 1.0;

				/* Rotation next */
      xmat_rot[0] = v0[0];
      xmat_rot[4] = v0[1];	/* x-axis */
      xmat_rot[8] = v0[2];

      xmat_rot[1] = v1[0];
      xmat_rot[5] = v1[1];	/* y-axis */
      xmat_rot[9] = v1[2];

      CROSS(v0, v1, v);
      
      xmat_rot[2] = v[0];
      xmat_rot[6] = v[1];	/* z-axis */
      xmat_rot[10] = v[2];

      xmat_rot[3] = xmat_rot[7] = xmat_rot[11] = 0.0;
      xmat_rot[12] = xmat_rot[13] = xmat_rot[14] = 0.0;
      xmat_rot[15] = 1.0;


      for (i = tpnum - npatches + 1; i <=tpnum; i++)
	{
	  tp =&tpatch[i];
	  for (j = 0; j < 3; j++)
	    {
	      yourpt_to_mypt(tp->coeffs[j][0], v);
	      matvec_mult(xmat_rot, v, vv);
	      matvec_mult(xmat_tr, vv, v);
	      mypt_to_yourpt_no_w(v, tp->coeffs[j][0]);
	      tp->coeffs[j][0].dCoord[3] = v[3];
	    }
	}
    }
  
  return npatches;
}


/* Read a single curve from the data-file. Generate the required */
/* number of patches in the global table tpatch[]. Return the number */
/* of patches generated. */
int read_curve(FILE *infile)
{
  int sp_type, npatches;
  int tpnum, i, j;
  vector_t v0, v1;
  TENSOR_PATCH *tp;
  

  fscanf(infile, "%d", &sp_type);	/* read in curve type */
  
  switch (sp_type)
    {
    case SP_LINE:		/* line segment */
      {
	tpnum = get_new_patch(1, 1);	/* allocate space */
	tp = &tpatch[tpnum];
	
	read_line(infile, v0, v1);
	mypt_to_yourpt(v0, tp->coeffs[0][0]);
	tp->coeffs[0][0].dCoord[3] = 1.0;
	mypt_to_yourpt(v1, tp->coeffs[1][0]);
	tp->coeffs[1][0].dCoord[3] = 1.0;
	
#ifdef DEBUG
	printf("\t\tline:\n");
#endif
	npatches = 1;
	break;
      }
      
    case SP_CIRCLE:		/* arc of a circle */
      {
	circle_t cc;
	read_plane(infile, &cc.pl, v0, v1);
	fscanf(infile, "%lf%lf%lf", &cc.c[0], &cc.c[1], &cc.c[2]);
	fscanf(infile, "%lf%lf%lf", &cc.r, &cc.a0, &cc.a1);
	
#ifdef DEBUG
	printf("\t\tcircle:\n");
#endif      
	npatches = generate_patches_for_circle(&cc, v0, v1);
	break;
      }

    case SP_ELLIPSE:		/* arc of an ellipse */
      {
	ellipse_t el;

	read_plane(infile, &el.pl, v0, v1);
	fscanf(infile, "%lf%lf%lf", &el.c[0], &el.c[1], &el.c[2]);
	fscanf(infile, "%lf%lf", &el.rmaj, &el.rmin);
	fscanf(infile, "%lf%lf%lf", &el.ang, &el.a0, &el.a1);
	
#ifdef DEBUG
	printf("\t\tellipse:\n");
#endif      
	npatches = generate_patches_for_ellipse(&el, v0, v1);
	break;
      }
				/* generic curve (in power basis) */
    case SP_CURVE:
      {
	gcurve_t gc;
	int ord, dummy, ntab;
	vector_t *a, *b;

	fscanf(infile, "%d%d%d%d", &gc.iarc1, &gc.iarcn, &gc.narc,
	       &gc.maxord);
	fscanf(infile, "%lf%lf", &gc.w0, &gc.w1);
	
#ifdef DEBUG
	printf("\t\tgeneric curve:\n");
#endif      

	for (i = gc.iarc1; i <= gc.iarcn; i++)
	  {
	    fscanf(infile, "%d%d", &ntab, &dummy);
	    a = (vector_t *)new(ntab*sizeof(vector_t));
	    b = (vector_t *)new(ntab*sizeof(vector_t));

#ifdef CONFUSED
	    for (j = 0; j < ntab; j++)
	      fscanf(infile, "%lf%lf%lf", &a[j][0], &a[j][1],
		     &a[j][2]);
#else
	    /* Read in all the coeffs. for the power basis. 'ndeg' */
	    /* coeffs for X first, followed by Y, and Z. The */
	    /* hard-copy documentation is wrong about this format */
	    for (j = 0; j < ntab; j++)
	      fscanf(infile, "%lf", &a[j][0]);
	    for (j = 0; j < ntab; j++)
	      fscanf(infile, "%lf", &a[j][1]);
	    for (j = 0; j < ntab; j++)
	      fscanf(infile, "%lf", &a[j][2]);
#endif

	    /* Convert the power basis to Bezier basis */
	    ord = ntab;
	    power_to_bez(&ord, a, b);
	
#ifdef DEBUG
	    printf("\t\t\tord = %d\n", ord);
#endif      	    
	    tpnum = get_new_patch(ord - 1, 1);	/* allocate space */
	    tp = &tpatch[tpnum];	    

	    for (j = 0; j <= ord - 1; j++)
	      {
		mypt_to_yourpt(b[j], tp->coeffs[j][0]);
		tp->coeffs[j][0].dCoord[3] = 1.0;
	      }
	    
	    free(a);
	    free(b);
	  }
	
	npatches = gc.iarcn - gc.iarc1 + 1;
	break;
      }
      
    default:
      printf("Not implemented yet!!!\n\n");
      break;
    }
  
  return npatches;
}


/* Input: top portion of the curve for the prism primitive. Project */
/* these points along the axis vector and clip it off with the two */
/* planes specified. w remains unchanged in this process (which is */
/* basically a parallel projection) */

int project_control_points(int idx, vector_t v, plane_typ *pl1, 
			   plane_typ *pl2)
{
  TENSOR_PATCH *tp = &tpatch[idx];
  register int i, j;
  double t, m, w, t0;
  
  m = _LENGTH(v);
  v[0] /= m; v[1] /= m; v[2] /= m;

  for (j = 0; j <= tp->iDeg0; j++)
    {
      w = tp->coeffs[j][0].dCoord[3];
      tp->coeffs[j][0].dCoord[0] /= w;
      tp->coeffs[j][0].dCoord[1] /= w;
      tp->coeffs[j][0].dCoord[2] /= w;
      tp->coeffs[j][0].dCoord[3] /= w;
      
      /* Get intersection with pl1 */
      t = -1.0 * (pl1->coeff[3] + DOT(pl1->coeff, tp->coeffs[j][0].dCoord))/
	DOT(pl1->coeff, v);
      t0 = t;

      tp->coeffs[j][1].dCoord[0] = tp->coeffs[j][0].dCoord[0] + t*v[0];
      tp->coeffs[j][1].dCoord[1] = tp->coeffs[j][0].dCoord[1] + t*v[1];
      tp->coeffs[j][1].dCoord[2] = tp->coeffs[j][0].dCoord[2] + t*v[2];
      tp->coeffs[j][1].dCoord[3] = 1.0;

      /* Get intersection with pl2 */
      t = -1.0 * (pl2->coeff[3] + DOT(pl2->coeff, tp->coeffs[j][0].dCoord))/
	DOT(pl2->coeff, v);

      if (t > t0)		/* 0th row is the top of the prism */
	top_is_zero = TRUE;
      else
	top_is_zero = FALSE;
	  
      tp->coeffs[j][0].dCoord[0] = tp->coeffs[j][0].dCoord[0] + t*v[0];
      tp->coeffs[j][0].dCoord[1] = tp->coeffs[j][0].dCoord[1] + t*v[1];
      tp->coeffs[j][0].dCoord[2] = tp->coeffs[j][0].dCoord[2] + t*v[2];
      tp->coeffs[j][0].dCoord[3] = 1.0;

      tp->coeffs[j][1].dCoord[0] *= w;
      tp->coeffs[j][1].dCoord[1] *= w;
      tp->coeffs[j][1].dCoord[2] *= w;
      tp->coeffs[j][1].dCoord[3] *= w;

      tp->coeffs[j][0].dCoord[0] *= w;
      tp->coeffs[j][0].dCoord[1] *= w;
      tp->coeffs[j][0].dCoord[2] *= w;
      tp->coeffs[j][0].dCoord[3] *= w;
      
    }
  
  return 0;
}


/* Easy one!!!! All the upper control points of all the patches are */
/* the same. */
int generate_pyramid_control_points(int idx, vector_t v)
{
  register int i;
  TENSOR_PATCH *tp = &tpatch[idx];

  /* The control points for the curve are generated in the first row */
  /* of the coefficients array. Copy them to the second row, and make */
  /* all the points in the first row same as v */

  for (i = 0; i <= tp->iDeg0; i++)
    {
      mypt_to_yourpt(v, tp->coeffs[i][1]);
      tp->coeffs[i][1].dCoord[3] = tp->coeffs[i][0].dCoord[3] = 1.0;
    }
  
  return 0;
}


/* Generate the patches for parallelopiped */
int generate_patches_for_ppped(ppped_t *pp)
{
  vector_t v, v0, v1;
  double len, len0, len1, len2;
  double sc_tr_prod;
  int sense;
  int tpnum;
  TENSOR_PATCH *tp;

  /* Normalise the direction vectors first!! */

  len = _LENGTH(pp->v0);
  pp->v0[0] /= len;
  pp->v0[1] /= len;
  pp->v0[2] /= len;

  len = _LENGTH(pp->v1);
  pp->v1[0] /= len;
  pp->v1[1] /= len;
  pp->v1[2] /= len;

  len = _LENGTH(pp->v2);
  pp->v2[0] /= len;
  pp->v2[1] /= len;
  pp->v2[2] /= len;

  sc_tr_prod = pp->v0[0]*(pp->v1[1]*pp->v2[2] - pp->v2[1]*pp->v1[2]) -
               pp->v0[1]*(pp->v1[0]*pp->v2[2] - pp->v2[0]*pp->v1[2]) +
	       pp->v0[2]*(pp->v1[0]*pp->v2[1] - pp->v2[0]*pp->v1[1]);

  if (sc_tr_prod > 0.0) /* no epsilon used here because I want a */
                        /* good positive scalar triple product */
    sense = 1;
  else
    sense = 0;
			     
  len0 = pp->len0;
  len1 = pp->len1;
  len2 = pp->len2;

  if (sense) {
    /* face #1 - bottom */

    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[0][1].dCoord[0] = pp->c0.x;
    tp->coeffs[0][1].dCoord[1] = pp->c0.y;
    tp->coeffs[0][1].dCoord[2] = pp->c0.z;
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len1 * pp->v1[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len1 * pp->v1[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len1 * pp->v1[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len1 * pp->v1[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len1 * pp->v1[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len1 * pp->v1[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    /* face #2 - front */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[0][0].dCoord[0] = pp->c0.x;
    tp->coeffs[0][0].dCoord[1] = pp->c0.y;
    tp->coeffs[0][0].dCoord[2] = pp->c0.z;
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len2 * pp->v2[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len2 * pp->v2[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len2 * pp->v2[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len2 * pp->v2[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len2 * pp->v2[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len2 * pp->v2[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    /* face #3 - left */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[1][1].dCoord[0] = pp->c0.x;
    tp->coeffs[1][1].dCoord[1] = pp->c0.y;
    tp->coeffs[1][1].dCoord[2] = pp->c0.z;
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len2 * pp->v2[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len2 * pp->v2[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len2 * pp->v2[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len1 * pp->v1[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len1 * pp->v1[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len1 * pp->v1[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len1 * pp->v1[0] + len2 * pp->v2[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len1 * pp->v1[1] + len2 * pp->v2[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len1 * pp->v1[2] + len2 * pp->v2[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    /* face #4 - right */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len1 * pp->v1[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len1 * pp->v1[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len1 * pp->v1[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len2 * pp->v2[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len2 * pp->v2[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len2 * pp->v2[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + 
      len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + 
      len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + 
      len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;

    /* face #5  - top */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len2 * pp->v2[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len2 * pp->v2[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len2 * pp->v2[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len2 * pp->v2[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len2 * pp->v2[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len2 * pp->v2[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + 
      len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + 
      len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + 
      len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    /* face #6 (at last!!) - back */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len1 * pp->v1[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len1 * pp->v1[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len1 * pp->v1[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len1 * pp->v1[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len1 * pp->v1[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len1 * pp->v1[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + 
      len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + 
      len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + 
      len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
  } else {
    /* face #1 - top */

    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[0][0].dCoord[0] = pp->c0.x;
    tp->coeffs[0][0].dCoord[1] = pp->c0.y;
    tp->coeffs[0][0].dCoord[2] = pp->c0.z;
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len1 * pp->v1[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len1 * pp->v1[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len1 * pp->v1[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len1 * pp->v1[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len1 * pp->v1[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len1 * pp->v1[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    /* face #2 - front */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[0][0].dCoord[0] = pp->c0.x;
    tp->coeffs[0][0].dCoord[1] = pp->c0.y;
    tp->coeffs[0][0].dCoord[2] = pp->c0.z;
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len2 * pp->v2[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len2 * pp->v2[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len2 * pp->v2[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len2 * pp->v2[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len2 * pp->v2[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len2 * pp->v2[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    /* face #3 - left */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[0][0].dCoord[0] = pp->c0.x;
    tp->coeffs[0][0].dCoord[1] = pp->c0.y;
    tp->coeffs[0][0].dCoord[2] = pp->c0.z;
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len2 * pp->v2[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len2 * pp->v2[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len2 * pp->v2[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len1 * pp->v1[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len1 * pp->v1[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len1 * pp->v1[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len1 * pp->v1[0] + len2 * pp->v2[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len1 * pp->v1[1] + len2 * pp->v2[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len1 * pp->v1[2] + len2 * pp->v2[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    /* face #4 - right */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len1 * pp->v1[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len1 * pp->v1[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len1 * pp->v1[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len2 * pp->v2[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len2 * pp->v2[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len2 * pp->v2[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + 
      len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + 
      len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + 
      len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;

    /* face #5  - bottom */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len2 * pp->v2[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len2 * pp->v2[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len2 * pp->v2[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
    
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len2 * pp->v2[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len2 * pp->v2[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len2 * pp->v2[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + 
      len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + 
      len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + 
      len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    /* face #6 (at last!!) - back */
    
    tpnum = get_new_patch(1, 1);
    tp = &tpatch[tpnum];
    tp->coeffs[1][0].dCoord[0] = pp->c0.x + len1 * pp->v1[0];
    tp->coeffs[1][0].dCoord[1] = pp->c0.y + len1 * pp->v1[1];
    tp->coeffs[1][0].dCoord[2] = pp->c0.z + len1 * pp->v1[2];
    tp->coeffs[1][0].dCoord[3] = 1.0;
    
    tp->coeffs[0][0].dCoord[0] = pp->c0.x + len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[0][0].dCoord[1] = pp->c0.y + len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[0][0].dCoord[2] = pp->c0.z + len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[0][0].dCoord[3] = 1.0;
    
    tp->coeffs[1][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + len1 * pp->v1[0];
    tp->coeffs[1][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + len1 * pp->v1[1];
    tp->coeffs[1][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + len1 * pp->v1[2];
    tp->coeffs[1][1].dCoord[3] = 1.0;
    
    tp->coeffs[0][1].dCoord[0] = pp->c0.x + len0 * pp->v0[0] + 
      len2 * pp->v2[0] + len1 * pp->v1[0];
    tp->coeffs[0][1].dCoord[1] = pp->c0.y + len0 * pp->v0[1] + 
      len2 * pp->v2[1] + len1 * pp->v1[1];
    tp->coeffs[0][1].dCoord[2] = pp->c0.z + len0 * pp->v0[2] + 
      len2 * pp->v2[2] + len1 * pp->v1[2];
    tp->coeffs[0][1].dCoord[3] = 1.0;
  }
  
  return sense;
}
