#include <intersect.h>
#include <math.h>

#define STR_TOLERANCE 1.0e-6
#define FP_EQUAL_STR(s, t) (fabs(s - t) <= STR_TOLERANCE)

extern node_t qs[];		/* Query structure */
extern trap_t tr[];		/* Trapezoid structure */
extern segment_t seg[];		/* Segment table */

extern struct global_s global;
extern int is_planar_overlap;

static polychain_t poly[POLYCHAIN_SIZE]; /* a linked list to hold the */
					 /* most current polygon */
					 /* structure  */
pchain_t *pchain[MAX_POLYCHAINS]; /* array to hold all the piecewise */
				  /* chains  */

static sinfo_t ch[MAX_INTERSECTIONS_PER_SEGMENT]; /* used to store all */
						  /* the intersections */
						  /* of one segment in */
						  /* sorted order */
static stack_t stk[MAX_STACKSIZE]; /* stack!! */
static int sp, poly_idx, whichside;

einfo_t edge[SEGSIZE];		/* GLOBAL table to hold the endpoints */
				/* of the polygon and lots of other */
				/* information  */

extern int shady_segments_intersect();

/* normalise vector: used for angle calculations */
static int normalise(vector_t *v)
{
  double denom;

  denom = sqrt(v->x*v->x + v->y*v->y);
  v->x /= denom;
  v->y /= denom;

  return 0;
}


/* Initialise all the s-values i the edge-intersections array so that */
/* insertion in=s always possible */
static int init_tables(void)
{
  register int i, j;
  
  for (i = 0; i < SEGSIZE; i++)
    {
      edge[i].nintr = 0;
      for (j = 0; j < MAX_INTERSECTIONS_PER_EDGE; j++)
	edge[i].intr[j].s = 2.00;
    }
  return 0;
}

/* return TRUE if p2 is to the left of (p0 -> p1) */
static int intr_is_left_of(point_typ *p0, point_typ *p1,
		    point_typ *p2)
{
  double area;
  area = CROSS(*p0, *p1, *p2);
  return (area > 0.0);
}



/* Add one more intersection to the edge. Maintain this list in sorted */
/* order. */
static int insert_iinfo(einfo_t *e, int nch, int segno, int inum,
			double s, double t)
{
  register int i;
  int j;
  int which_enters, which_ch;

#ifdef OLD_VERSION  
  for (i = 0; i < MAX_INTERSECTIONS_PER_EDGE; i++)
    if (e->intr[i].s > s)
      break;
#else
  for (i = 0; i < MAX_INTERSECTIONS_PER_EDGE; i++)
    if (FP_EQUAL(e->intr[i].s, s) || (e->intr[i].s > s))
      break;
  
  if (FP_EQUAL(e->intr[i].s, s)) /* Same intersection */
    {
      /* Get the segment which is entering. Only one of these wil be */
      /* entering */
      if (FP_EQUAL(e->intr[i].t, 1.0)) /* seg# i was entering */
	{			       /* i.e. current is leaving */
	  which_enters = e->intr[i].nseg;
	  which_ch = e->intr[i].nch;
	  
	  if (intr_is_left_of(&pchain[which_ch][which_enters].v,
			      &pchain[which_ch][which_enters+1].v,
			      &pchain[nch][segno+1].v))
	    i = i+1;
	}
      else
	i++;
    }
#endif

  for (j = e->nintr; j > i; j--)
    e->intr[j] = e->intr[j - 1];

  e->intr[i].nch = nch;
  e->intr[i].nseg = segno;
  e->intr[i].inum = inum;
  e->intr[i].s = s;
  e->intr[i].t = t;
      
  e->nintr++;
  return i;
}


/* Add one more intersection to the list of all intersections of a */
/* particular segment. */
static int insert_sinfo(sinfo_t ch[], int nch, point_typ *p,
			int nedge, double s, double t, int h)
{
  register int i;
  int j;
  
  for (i = 0; i < nch; i++)
    if ((ch[i].t > t) || (FP_EQUAL(ch[i].t, t) && (ch[i].h > h)))
      break;
  
  /* The extra h-information added to take care of the case when a */
  /* segment just grazes the commonpoint of 2 edges and lies inside. */
  /* In this case, we consider it to be two intersections */

  for (j = nch; j > i; j--)
    ch[j] = ch[j - 1];

  ch[i].s = s;
  ch[i].t = t;
  ch[i].h = h;
  ch[i].nedge = nedge;
  ch[i].v = *p;
      
  return 0;
}


/* This routine is used to break a chain into 2 when we detect the */
/* condition that some segment of the chain partially/fully overlaps */
/* an edge of the polygon. Just make the remaining points as members */
/* of a new chain */
static int make_new_polychain(int nchain, int k, int j, point_typ *v)
{
  register int i;
  int first_posn;

  pchain[nchain] = (pchain_t *) new(MAX_CHAIN_LENGTH*
				       sizeof(pchain_t));

  /* Take care that the first two points are not the same. This can */
  /* happen if the segment on the chain lies completely within an edge */
  /* of the polygon, so that the intersection, and the next point on */
  /* the chain are the same */

  if (FP_EQUAL(v->x, pchain[k][j].v.x) &&
      FP_EQUAL(v->y, pchain[k][j].v.y))
      first_posn = 0;
  else
    {
      pchain[nchain][0].v = *v;
      pchain[nchain][0].valid = TRUE;
      first_posn = 1;
    }

  for (i = first_posn; pchain[k][j].valid; j++, i++)
    {
      pchain[nchain][i] = pchain[k][j];
    }
  pchain[nchain][i].valid = FALSE;

  return nchain;
}


/* return the angle formed by the points (u0, u1, u2) in that order. */
/* Basically, it is a numerical value between 1 and -3. Used for */
/* comparison only */
static double intr_angle_between(point_typ *u0, point_typ *u1,
		    point_typ *u2)
{
  vector_t v0, v1;
  
  v0.x = u1->x - u0->x;
  v0.y = u1->y - u0->y;
  v1.x = u2->x - u1->x;
  v1.y = u2->y - u1->y;
  
  normalise(&v0);
  normalise(&v1);
  if (CROSS_VEC(v0, v1) >= 0.0)
    return  -DOT(v0, v1); /* 0: 1.0    180: -1.0 */
  else
    return (-2.0 + DOT(v0, v1));
}


/* Find the intersections of all the chaons with all theedges of the */
/* polygon */
int find_chain_intersections(int *nchains) 
{
  register int i, k;
  point_typ pp0, pp1, p0, p1, *v0, *v1, vv0, vv1;
  int nch, ich, nint, p, q, gnch, bflag, itype;
  double s0, s1, t0, t1;
  int pos, prev_edge, nedge, delta, hackflag;
  pchain_t *pc;
  int N = global.nseg;
  int is_first_last;

  init_tables();
  k = 0;
  while (k < *nchains)
    {
      p = 0;
      q = 1;
      pc = pchain[k];
      gnch = 0;
      
      while (pc[q].valid && pc[p].valid) /* not end-of-chain */
	{
	  /* for each segment in the chain do */
  	  
	  ich = 0;		/* index for #intersections */
	  bflag = hackflag = FALSE;
	  if (p == 0)
	    is_first_last = I_FIRSTPT;
	  else if (!pc[q+1].valid)
	    is_first_last = I_LASTPT;
	  else
	    is_first_last = I_NONEPT;

	  for (i = 0; i < global.nseg; i++)
	    {
	      v0 = &pc[p].v;
	      v1 = &pc[q].v;

	      if ((nint = 
		   shady_segments_intersect(&edge[i].v0, &edge[i].v1,
					  v0, v1, &itype, 
					  &s0, &s1, &t0, &t1,
					  &p0, &p1,
					  is_first_last)) == 0)
		{
		  /* segments do not intersect */
		  continue;
		}
	      else if (nint == 1)
		{
		  int cond;
		  /* insert information about this intrn. in the edgeinfo */
		  /* structure, and the ch-table */
		  
		  cond = FP_EQUAL(s0, 0.0) &&
		    (FP_EQUAL(t0, 0.0) || FP_EQUAL(t0, 1.0));

		  if ((FP_EQUAL(s0, 0.0) && cond)
				/* take care in next edge */
		      || 
		      (FP_EQUAL(s0, 1.0) && /* coincident with */
					    /* next edge */
		       intr_are_vectors_coincident(v0, v1, 
					   &edge[(i+1)%N].v0,
					   &edge[(i+1)%N].v1))
		      ||
		      ((s0 < 1.0) && !(cond) && FP_EQUAL(t0, 1.0) &&
		       (!intr_is_left_of(&edge[i].v0, &edge[i].v1,
				    v0)))
		      ||
		      (FP_EQUAL(t0, 0.0) && (s0 < 1.0) &&
		       (!cond)  &&
		       (!intr_is_left_of(&edge[i].v0, 
					 &edge[i].v1,
					 v1)))
		      ||
		      (FP_EQUAL(t0, 0.0) && FP_EQUAL(s0, 1.0) &&
		       !(intr_angle_between(&edge[i].v0, 
					    &edge[i].v1,
					    v1) > 
			 intr_angle_between(&edge[i].v0, 
					    &edge[i].v1,
					    &edge[(i+1)%N].v1)))) 
		    ;
		  else
		    {
		      if ((FP_EQUAL(s0, 0.0) || FP_EQUAL(s0, 1.0))
			  && !(FP_EQUAL(t0, 0.0) || 
			       FP_EQUAL(t0, 1.0)))
			{
			  int tp, h;
			  /* Check if the grazing edge is inside or */
			  /* outside */
			  if (FP_EQUAL(s0, 1.0))
			    {
			      tp = (i+1)%N;
			      vv0.x = edge[i].v1.x - edge[i].v0.x;
			      vv0.y = edge[i].v1.y - edge[i].v0.y;
				
			      vv1.x = edge[tp].v1.x - edge[tp].v0.x;
			      vv1.y = edge[tp].v1.y - edge[tp].v0.y;
			      
			      if (CROSS_VEC(vv0, vv1) > 0.0)
				continue;
			      else
				{
				  if (intr_is_left_of(&edge[i].v0,
						      &edge[i].v1,
						      v1))
				    h = 2;
				  else
				    h = 1;
				  insert_sinfo(ch, ich, &p0, i, s0,
					       t0, h);
				  ich++;
				  continue;
				}
			    }
			  else
			    {
			      tp = (i+N-1)%N;
			      vv0.x = edge[tp].v1.x - edge[tp].v0.x;
			      vv0.y = edge[tp].v1.y - edge[tp].v0.y;
			      
			      vv1.x = edge[i].v1.x - edge[i].v0.x;
			      vv1.y = edge[i].v1.y - edge[i].v0.y;
			      
			      if (CROSS_VEC(vv0, vv1) > 0.0)
				continue;
			      else
				{
				  if (intr_is_left_of(&edge[i].v0,
						      &edge[i].v1,
						      v1))
				    h = 2;
				  else
				    h = 1;
				  insert_sinfo(ch, ich, &p0, i, s0,
					       t0, h);
				  ich++;
				  continue;
				}
			    }
			}
		      insert_sinfo(ch, ich, &p0, i, s0, t0, -1);
		      ich++;
		    }
		}
	      else
		{
		  /* segments were parallel */
		  /* Now, insert the intersection points in the s_info */
		  /* structure only if the points are not (0, 1) for */
		  /* the chain. Because, if they are not 0, 1 then */
		  /* they s */		  
/*		  if (itype == I_PAR)
		    {
		      insert_sinfo(ch, ich, &p0, i, s0, t0, -1);
		      ich++;
		      insert_sinfo(ch, ich, &p1, i, s1, t1, -1);
		      ich++;
		    }
		  else
		    {
		      insert_sinfo(ch, ich, &p1, i, s1, t0, -1);
		      ich++;
		      insert_sinfo(ch, ich, &p0, i, s0, t1, -1);
		      ich++;
		    } */
		}
	    } /* end-forloop */
	  
	  /* Now, we have all the intersections for the segment into */
	  /* the table ch[]. Now scan this table and update the */
	  /* edge_info structures */

	  
	  /* Changes: (Oct 14). Try to reduce the number of */
	  /* poly_chains as far as possible by looking whether the */
	  /* subsequent segments in the chain too intersect with this */
	  /* same edge in the polygon. If so, skip over these */
	  /* remainign segments of this chain to avoid further */
	  /* un-necessary calls to make_new_polychain() */

	  prev_edge = -1;
	  delta = 0;
	  for (i = 0; i < ich; i++)
	    {
	      if ((nedge = ch[i].nedge) == prev_edge) 
				/* 2-intersections with */
				/* the same edge */
		{
		  make_new_polychain((*nchains)++, k, q, &ch[i].v);
		  bflag = TRUE;
		  break;
		}
	      else if (((i+1) < ich) && (ch[i+1].nedge == nedge))
		{		/* to prevent erroneous */
				/* insertion of the first intrn. */
		  make_new_polychain((*nchains)++, k, q+1, &ch[i+1].v);
		  bflag = TRUE;
		  break;		  
		}
				/* new intrn. is with a diff. edge */
	      else if ((prev_edge == (nedge + 1)%global.nseg) && 
		       (FP_EQUAL(ch[i].s, 1.0)))
		{
		  /* Change : (Oct 23) */
		  /* Even though the intersection is with a different */
		  /* edge, a case can arise where the new intersection */
		  /* is with the adjacent egde at the boundary where */
		  /* (s = 1) and so that it was not recorded in the */
		  /* current edge. So check whether the current */
		  /* intersection is with an edge at s = 1, s.t there */
		  /* was an intersection with an adjacent edge earlier */

		  make_new_polychain((*nchains)++, k, q, &ch[i].v);
		  bflag = TRUE;
		  break;		  
		  
		}
	      else		/* good diff-edge intersection */
		{
		  insert_iinfo(&edge[nedge], k, p, i+gnch-delta,
			       ch[i].s, ch[i].t);
		  prev_edge = nedge;
		}
	    } /* end-for */
	  
	  if (bflag)
	    {
	      bflag = FALSE;
	      break;
	    }
	  
	  p++;
	  q++;
	  gnch += ich - delta;;
	} /* end-while (for each segment) */
      
      k++;
    }
  
  return 0;
}
      
#define STACK_EMPTY (sp == 0)

static int stack_push(int edgenum, int einum, iinfo_t *ip)
{
  stack_t *s;
  
  ++sp;
  s = &stk[sp];
  s->chnum = ip->nch;
  s->segnum = ip->nseg;
  s->edgenum = edgenum;
  s->epos = edge[edgenum].epos;
  s->inum = ip->inum;
  s->einum = einum;
  s->s = ip->s;
  s->t = ip->t;
  s->nidx = 0;
  return 0;
}

static int stack_replace(int edgenum, int einum, iinfo_t *ip)
{
  stack_t *s = &stk[sp];

  s->chnum = ip->nch;
  s->segnum = ip->nseg;
  s->edgenum = edgenum;
  s->epos = edge[edgenum].epos;
  s->inum = ip->inum;
  s->einum = einum;
  s->s = ip->s;
  s->t = ip->t;
  return 0;
}

static stack_t *stack_pop()
{
  if (STACK_EMPTY)
    fprintf(stderr, "stack_pop: empty stack\n");
  sp--;
  return &stk[sp + 1];

}


/* Add the partition as another neighbour of stack-entry */
static int add_neighbour(int s_idx, int part_no)
{
  stack_t *s = &stk[s_idx];
  
  s->adj[s->nidx++] = part_no;
  return 0;
}


/* Copy the current neighbours of this partition from the stack into */
/* the partition structure, and set neighbouring info. in the */
/* neighbours */

static int set_neighbours(int s_idx, int part_no, part_t *p)
{
  register int i;
  stack_t *s = &stk[s_idx];
  int k, l;
  
  p[part_no].nnb = s->nidx;
  for (i = 0; i < s->nidx; i++)
    {
      k = p[part_no].adj[i] = s->adj[i];
      l = p[k].nnb++;
      p[k].adj[l] = part_no;
    }

  return 0;
}

/* return TRUE if the edge is entering the polygon at the point of */
/* intersection */
static int intr_entering(int edgenum, int chnum, int
			 segnum, double s, double t)
{
  if (FP_EQUAL(t, 0.0))		/* the segment begins here */
    return TRUE;

  if (FP_EQUAL(t, 1.0))
    return FALSE;
  
  return intr_is_left_of(&edge[edgenum].v0, &edge[edgenum].v1,
			   &pchain[chnum][segnum + 1].v);
}


static int new_polypoint(void)
{
  return poly_idx++;
}


/* After the stack is emptied, the last partition is available in the */
/* linked list poly[]. Output this */
static int intr_add_last_polygon(part_t *part, int oldpos)
{
  int i;
  int n;
  int pidx;
  
  if ((i = poly[oldpos].next) == oldpos)
    {
      part->npoints = 0;
      return 0;
    }

  part->pc = (point_typ *) new(MAX_PARTITION_SIZE*sizeof(point_typ));
  pidx = n = 0;
  part->pc[pidx++] = poly[oldpos].v;
  while (i != oldpos)
    {
      part->pc[pidx++] = poly[i].v;
      i = poly[i].next;
    }
  part->npoints = pidx;
  part->whichside = (whichside == S_LEFT) ? S_RIGHT : S_LEFT;

  return 1;
}

/* remove the spurious duplicate points in the polygon !! */

int remove_duplicate_points_in_partition(part_t *p)
{
  register int i, j;
  
  for (i = 0; i < p->npoints - 1; i++)
    if (FP_EQUAL(p->pc[i].x, p->pc[i+1].x) && 
	FP_EQUAL(p->pc[i].y, p->pc[i+1].y))
      {
	for (j = i; j < p->npoints - 1; j++)
	  p->pc[j] = p->pc[j+1];
	p->npoints--;
      }
  
  return 0;
}

/* A proper combination has been discovered on the stack. Construct a */
/* polygon using this information and modify the polygon chain */

static int intr_emit_polygon(part_t *part, stack_t *s, int
			     curedge, iinfo_t *ip, int ip_einum)
{
  register int i;
  int k, kl;
  int pidx, p, q, nelem;
  int N, dir, dirstep;
  int e, tmp;
  point_typ pi0, pi1;
  int cpos0, cpos1, ppos0, ppos1, epos;


  N = global.nseg;
  pidx = 0;
  e = s->edgenum;
  nelem = (curedge - e + N)%N +
    abs(ip->nseg - s->segnum) + 2;
  
  if ((ip->inum - s->inum) > 0)
    {
      dir = 1;
      dirstep = 1;		/* s to ip */
      
      /* New addition: (Oct 12). Include a 'which-side' flag in each */
      /* partition to determine which side of the chain it lies. Not */
      /* of use in partiton-polygon, but required for handling the */
      /* special planar intersection case where such partitions have */
      /* to be identified explicitly */

      whichside = S_RIGHT;	/* Outside B */
    }
  else
    {
      dir = 0;
      dirstep = -1;		/* ip to s */
      
      whichside = S_LEFT;	/* Inside B */
    }


  part->pc = (point_typ *) new(MAX_PARTITION_SIZE*sizeof(point_typ));
  part->whichside = whichside;

  /* Three possible cases at the first intersection */
  if (FP_EQUAL(s->s, 1.0))
    {
      if (FP_EQUAL(s->t, 0.0))	/* end-position for chain-loop */
	cpos1 = s->segnum + dirstep;
      else if (FP_EQUAL(s->t, 1.0))
	cpos1 = s->segnum + dir;
      else
	cpos1 = s->segnum + dir;
	    
      /* ppos0 = edge[(e + 1) % N].epos;*/ /* start posn for poly-loop */
      ppos0 = poly[edge[e].epos].next;
      epos = -1;
    }
  else if (FP_EQUAL(s->s, 0.0))
    {
      if (FP_EQUAL(s->t, 0.0))	/* end-position for chain-loop */
	cpos1 = s->segnum + dirstep;
      else if (FP_EQUAL(s->t, 1.0))
	cpos1 = s->segnum + dir;
      else
	cpos1 = s->segnum + dir;

      ppos0 = edge[e].epos;	/* will work !!! */
      epos = -1;      
    }
  else				/* usual case */
    {
      pi0.x = edge[e].v0.x + s->s*(edge[e].v1.x -  edge[e].v0.x);
      pi0.y = edge[e].v0.y + s->s*(edge[e].v1.y -  edge[e].v0.y);

#ifdef OLD_VERSION
      
      k = s->epos;		/* get the segment to break in the */
				/* polygon structure */
      if (((s->einum > 0) && 
	  (FP_EQUAL(edge[e].intr[s->einum - 1].s, s->s)) &&
	  (s->chnum == edge[e].intr[s->einum - 1].nch))) 
	{
	  ppos0 = k;
	  edge[e].epos = k;
	}
      else if (((s->einum + 1) < edge[e].nintr) && 
	       FP_EQUAL(edge[e].intr[s->einum + 1].s, s->s))
	{
	  ppos0 = poly[k].next;
	  edge[e].epos = k; 
	}
      else			/* single intersection .. not at */
				/* endpts .. good case */
	{
	  kl = new_polypoint();	
	  if (poly[k].next == (e+1)%N)
	    edge[e].epos = kl;
	  poly[kl].v = pi0;
	  poly[kl].next = poly[k].next;
	  poly[k].next = kl;
	  
	  ppos0 = kl;
	}
      
#else
      /* REMEMBER: edge number in ip is always >= edge number in s */

      k = s->epos;		/* get the segment to break in the */
				/* polygon structure */

      /* Case where there are two intersections at this point. */
      /* Consider which side to update the epos in */
      if (((s->einum > 0) && 
	  (FP_EQUAL(edge[e].intr[s->einum - 1].s, s->s)) &&
	   (s->chnum == edge[e].intr[s->einum - 1].nch))
          || (((s->einum + 1) < edge[e].nintr) && 
	       FP_EQUAL(edge[e].intr[s->einum + 1].s, s->s)))
	{
	  if (poly[k].next == (e+1)%N)
	    {			/* This is the first time we are */
				/* popping off this intersection */
				/* point. Hence, create a new point */
	      kl = new_polypoint();	
	      poly[kl].v = pi0;
	      poly[kl].next = poly[k].next;
	      poly[k].next = kl;
	      ppos0 = kl;
	      
	      edge[e].epos = k;
	    }
	  else
	    {			/* ont of the intersections at this */
				/* point has already been popped off */
	      ppos0 = poly[k].next;
	    } 
	}
      else			/* normal case */
	{  
	  kl = new_polypoint();	
	  if (poly[k].next == (e+1)%N)
	    edge[e].epos = kl;
	  poly[kl].v = pi0;
	  poly[kl].next = poly[k].next;
	  poly[k].next = kl;
	  ppos0 = kl;
	}

#endif
      
      if (FP_EQUAL(s->t, 0.0))	/* end-position for chain-loop */
	cpos1 = s->segnum + dirstep;
      else if (FP_EQUAL(s->t, 1.0))
	cpos1 = s->segnum + dir;
      else
	cpos1 = s->segnum + dir;
      epos = -1;
    }
  /* now take care of the other end (ip) */
  
  dir = 1 - dir;		/* dir = 0 : s to ip */
				/* dir = 1 : ip to s */
  if (FP_EQUAL(ip->s, 1.0))
    {
      if (FP_EQUAL(ip->t, 0.0))
	cpos0 = ip->nseg + dir; /* next pt. */
      else if (FP_EQUAL(ip->t, 1.0))
	cpos0 = ip->nseg + dir;
      else
	cpos0 = ip->nseg + dir;	

      ppos1 = poly[edge[curedge].epos].next;
      epos = -1;
    }
  else if (FP_EQUAL(ip->s, 0.0))
    {
      if (FP_EQUAL(ip->t, 0.0))
	cpos0 = ip->nseg + dir; /* next pt. */
      else if (FP_EQUAL(ip->t, 1.0))
	cpos0 = ip->nseg + dir;
      else
	cpos0 = ip->nseg + dir;	

      ppos1 = edge[curedge].epos;
      epos = -1;
    }
  else				/* usual case */
    {
      if (FP_EQUAL(ip->t, 0.0))
	cpos0 = ip->nseg + dir; /* next pt. */
      else if (FP_EQUAL(ip->t, 1.0))
	cpos0 = ip->nseg + dir;
      else
	cpos0 = ip->nseg + dir;	

      pi1.x = edge[curedge].v0.x + 
	ip->s*(edge[curedge].v1.x -  edge[curedge].v0.x);
      pi1.y = edge[curedge].v0.y + 
	ip->s*(edge[curedge].v1.y -  edge[curedge].v0.y);

#ifdef NEW_VERISON
      k = edge[curedge].epos;	/* get the segment to break in the */
				/* polygon structure */
      kl = new_polypoint();
      poly[kl].v = pi1;
      poly[kl].next = poly[k].next;
      poly[k].next = kl;
      
      edge[curedge].epos = kl;
      ppos1 = kl;
#else
      k = edge[curedge].epos;

      /* Case where there are two intersections at this point. */
      /* Consider which side to update the epos in */
      if (((ip_einum > 0) && 
	  (FP_EQUAL(edge[curedge].intr[ip_einum - 1].s, ip->s)) &&
	   (ip->nch == edge[curedge].intr[ip_einum - 1].nch))
          || (((ip_einum + 1) < edge[curedge].nintr) && 
	       FP_EQUAL(edge[curedge].intr[ip_einum + 1].s, ip->s)))
	{
	  if (poly[k].next == (curedge+1)%N)
	    {			/* This is the first time we are */
				/* popping off this intersection */
				/* point. Hence, create a new point */
	      kl = new_polypoint();	
	      poly[kl].v = pi1;
	      poly[kl].next = poly[k].next;
	      poly[k].next = kl;
	      ppos1 = kl;
	      
	      edge[curedge].epos = kl;
	    }
	  else
	    {			/* ont of the intersections at this */
				/* point has already been popped off */
	      ppos1 = poly[k].next;
	    } 				
	}
      else			/* single intersection .. not at */
				/* endpts .. good case */
	{
	  kl = new_polypoint();	
	  if (poly[k].next == (curedge+1)%N)
	    edge[curedge].epos = kl;
	  poly[kl].v = pi1;
	  poly[kl].next = poly[k].next;
	  poly[k].next = kl;
	  
	  ppos1 = kl;
	}
#endif

    }
  
  /* Now first, generate the partition polygon */
  
  for (i = ppos0; i != ppos1; i = poly[i].next)
    part->pc[pidx++] = poly[i].v;
  part->pc[pidx++] = poly[ppos1].v;
  
  if ((cpos0 - cpos1)*dirstep >= 0)
    {
      if (cpos1 > cpos0)
	tmp = 1;
      else
	tmp = -1;
      for (i = cpos0; i != cpos1; i += tmp)
	part->pc[pidx++] = pchain[ip->nch][i].v;
      part->pc[pidx++] = pchain[ip->nch][cpos1].v;
    }      
  
  /* Now mutilate the original polygon structure */

  if ((cpos0 - cpos1)*dirstep >= 0)
    {
      if (cpos0 > cpos1)
	tmp = 1;
      else
	tmp = -1;
      for (i = cpos1; i != cpos0; i += tmp)
	{
	  k = new_polypoint();
	  poly[k].v = pchain[ip->nch][i].v;
	  poly[ppos0].next = k;
	  ppos0 = k;
	}
      k = new_polypoint();
      poly[k].v = pchain[ip->nch][cpos0].v;
      poly[ppos0].next = k;
      ppos0 = k;

      poly[k].next = ppos1;
    }      
  else
    poly[ppos0].next = ppos1;

  part->npoints = pidx;

  /* Additional code to remove duplicate points generated (if at all) */
  /* in the chain!! Not the best approach, but otherwise, the code */
  /* will become too clumsy to handle */
  remove_duplicate_points_in_partition(part);

  return ppos0;
}



/* Return value: number of partitions generated */
/* Start from edge #0 to edge #(n-1). Consider all the intersections */
/* in sorted orde and use a stack-based algorithm to generate the */
/* partitions */
int trace_around_edges(part_t *part)
{
  register int i, j;
  einfo_t *ep;
  iinfo_t *ip;
  polychain_t *pp;
  int n = global.nseg, *tmp;
  int part_idx, oldpos, npts, ne;
  double chpoly[MAX_CHAIN_LENGTH][2], outerpoly[MAX_CHAIN_LENGTH][2];
  double val0, val1;

  /* first create a linked list of all the edges in the polygon */
  for (i = 0; i < n; i++)
    {
      poly[i].v = edge[i].v0;
      poly[i].edgenum = i;
      poly[i].next = i + 1;
      edge[i].epos = i;
    }
  poly[n - 1].next = 0;
  poly_idx = n;
  oldpos = 1;
  
  stk[0].nidx = 0;		/* # neighbours for last polygon */

  part_idx = sp = 0;			/* Stack pointer initialised */
  for (i = 0; i < n; i++)
    {
      ep = &edge[i];
      for (j = 0; j < ep->nintr; j++)
	{
	  stack_t *s = &stk[sp];
	  ip = &ep->intr[j];

/*	  
	  if (FP_EQUAL(ip->s, 1.0) && 
	      ((ne = (ip->nseg + 1)%n)
*/
	  if (STACK_EMPTY)	/* push if stack empty */
	    stack_push(i, j, ip);
	  else			/* not empty */
	    {
	      /* If the intersections are the same, replace the */
	      /* old one by the new one (if segno is bigger) */
	      /* CASE should not occur ideally!!!!!!!! */
	      if (FP_EQUAL(ip->s, s->s) && (s->edgenum == i) && 
		  (ip->inum > s->inum))
		stack_push(i, j, ip);
	      else 
		{
		  if ((ip->nch == s->chnum) && 
		      (abs(ip->inum - s->inum) == 1))
		    {
		      if (!intr_entering(i, ip->nch, ip->nseg,
				       ip->s, ip->t)
			&& intr_entering(s->edgenum, s->chnum, s->segnum,
					 s->s, s->t))
			{	/* always emit polygon */

			  /* Change Dec 28 *********************/
			  if (ip->inum > s->inum)
			    {
			      oldpos = intr_emit_polygon(&part[part_idx++],
							 s, i, ip, j);
			      add_neighbour(sp - 1, part_idx - 1);
			      set_neighbours(sp, part_idx - 1, part);
			      stack_pop();
			    }
			  else
			    stack_push(i, j, ip);
			}
		      else if (intr_entering(i, ip->nch, ip->nseg,
				     ip->s, ip->t) && 
			       !intr_entering(s->edgenum, s->chnum, 
					      s->segnum, s->s, s->t))
			{
			  if (ip->inum < s->inum)
			    {
			      oldpos = intr_emit_polygon(&part[part_idx++],
							 s, i, ip, j);
			      add_neighbour(sp - 1, part_idx - 1);
			      set_neighbours(sp, part_idx - 1, part);
			      stack_pop();
			    }			      
			  else
			    /* stack_pop();  */
			    /* Change: Oct 15 .. planar case ?? */
			    stack_push(i, j, ip);
			}
		      else
			stack_push(i, j, ip);
		    }
		  else
		    stack_push(i, j, ip);
		}
	    }
	} /* end inner for */
    } /* end outer-for */

  /* Now, the stack is empty. Hance, poly[] contains the last polygon. */
  /* Add it to the partition structure */
	
  if (!STACK_EMPTY)
    {
/*
      fprintf(stderr, "trace_around_edges:: error: stack not empty\n");
*/
      return -1;
      sleep(3);
    }
  
  intr_add_last_polygon(&part[part_idx++], oldpos);
  set_neighbours(0, part_idx - 1, part);

#ifndef DETECT_LOOP_PARTITIONS
  if (part_idx == 1)		/* no intersections were detected */
    {				/* check for holes... */
      int num_chpoints;
      int npoints;
      point_typ pA;

      npoints = part[0].npoints;

      /* Check whether any point on the chain in inside the polygon */
      for (i = 0; i < npoints; i++)
	{
	  outerpoly[i][0] = part[0].pc[i].x;
	  outerpoly[i][1] = part[0].pc[i].y;
	}
      outerpoly[npoints][0] = outerpoly[0][0]; /* The Polygon */
      outerpoly[npoints][1] = outerpoly[0][1];

      i = 0;
      while (pchain[0][i].valid)
	{
	  chpoly[i][0] = pchain[0][i].v.x;
	  chpoly[i][1] = pchain[0][i].v.y;
	  i++;
	}
      num_chpoints = i - 1;
      
      if (!(is_point_inside_polygon(npoints, outerpoly,
				   &pchain[0][0].v) &&
	    is_point_within_polygon(&pchain[0][0].v)))
	{			/* whole chain lies outside */
	  if (is_planar_overlap) 
	    {
	      /* find a point in A */
	      find_point_inside_polygon(npoints, outerpoly, &pA);
	      
	      /* chain is closed!! */
	      if (is_point_inside_polygon(num_chpoints, chpoly,
					  &pA))
		/* A completely lies inside the chain */
		part[0].whichside = S_LEFT;
	      else
		part[0].whichside = S_RIGHT;
	      part[0].nnb = 0;

	      return 1;
	    }
	}
      
      /* The complete chain lies inside the polygon. Check whether we */
      /* have more then 1 chains. If it does, we are doomed */
  
      /* check whether the chain is closed */
      if ((!(FP_EQUAL_PAR(chpoly[0][0], chpoly[num_chpoints][0], 1.0e-2) &&
	    FP_EQUAL_PAR(chpoly[0][1], chpoly[num_chpoints][1],
			 1.0e-2)))
	  || (num_chpoints <  3))
	{
#ifdef DEBUG_PRINT_POLY
	  printf ("Warning: chain inside the polygon not a loop\n");
#endif
	  part[0].nnb = 0;
	  return part_idx;
	}
      else
	{			/* ensure the endpoints are exactly */
				/* equal .. */
	  chpoly[num_chpoints][0] = chpoly[0][0];
	  chpoly[num_chpoints][1] = chpoly[0][1];
	}

      /* Orient the chpoly curve so that it is always anti-clockwise */

      if (!is_curve_anticlockwise(chpoly, num_chpoints))
	{
	  /* reverse the points */
	  for (i = 0; i < (num_chpoints+1)/2; i++)
	    {
	      val0 = chpoly[i][0];
	      val1 = chpoly[i][1];
	      chpoly[i][0] = chpoly[num_chpoints - i][0];
	      chpoly[i][1] = chpoly[num_chpoints - i][1];
	      chpoly[num_chpoints - i][0] = val0;
	      chpoly[num_chpoints - i][1] = val1;
	    }
	}

      split_into_two_partitions(part, outerpoly, npoints,
				chpoly, num_chpoints);

#ifdef OLD_VERSION      
      part_idx = 2;
#else
      /* We should return 3 partitions instead of two. We are */
      /* forgetting the original chain . Put that into part[3] */
      
      part[2].pc = 
	(point_typ *) new(MAX_PARTITION_SIZE*sizeof(point_typ));
      part[2].npoints = num_chpoints;
      for (i = 0; i < num_chpoints; i++)
	{
	  part[2].pc[i].x = chpoly[i][0];
	  part[2].pc[i].y = chpoly[i][1];
	}
      part[2].whichside = S_LEFT;

      part[0].nnb = part[1].nnb = 1; 
      part[2].nnb = 2;
      part[0].adj[0] = part[1].adj[0] = 2;
      part[2].adj[0] = 0; part[2].adj[1] = 1;

      part_idx = 3;
#endif

    }
  
#endif
  
  return part_idx;
}

/* Routine to detect whether the input chain has any */
/* self-intersections.. If yes, remove them. We do not perform an */
/* exuative search to test intersection of a segment with all other */
/* segments. Instead, assuming most of the kinks are formed when two */
/* different intersection-curves are combined, we need to check only */
/* the next-to-next segment for intersection */

int remove_kinks_from_chain(pchain_t *pch, int *npoints)
{
  register int i;
  double s0, s1, t0, t1;
  int j, itype;
  point_typ p0, p1;
  int nintr;

  for (i = 0; i < (*npoints - 3); i++)
    if ((nintr = segments_intersect(&pch[i].v, &pch[i+1].v, 
			   &pch[i+2].v, &pch[i+3].v, &itype,
			   &s0, &s1, &t0, &t1, &p0, &p1)) > 0)
      {
	/* there is an intersection. simply ignore pch[i+1]. i.e. */
	/* remove it from the list of points. Shift the remaining */
	/* points down by 1 unit */
		
	for (j = i+1; j < (*npoints - 1); j++)
	  pch[j] = pch[j+1];
	pch[*npoints - 1].valid = FALSE;
	(*npoints)--;

#ifdef DEBUG
	printf("remove_kinks_from_chain: detected spurious intersection\n"); 
#endif
	return 1;
      }

  return 0;
}


/* Remove all the collinear parts of the chain and replace them by a */
/* single segment. This is done to prevent detection of excessive */
/* number of segments in partition-polygon and to prevent breakdown of */
/* the tiangulation code */
int straighten_chain(pchain_t *pch, int *npoints)
{
  register int i, j;
  point_typ tmp0;
  int nA, delta;
  point_typ p0, p1, p2;
  double area;

  return 0;			/* Disblaed now!! Causes display */
				/* problems  */

  nA = *npoints;
  i = 0;
  while (i < (nA - 2))
    {
      p0 = pch[i].v;
      p1 = pch[i+1].v;

      j = i + 2;
      while (j <= (nA - 1))
	{
	  p2 = pch[j].v;
	
	  area = CROSS(p0, p1, p2);
	  if (!FP_EQUAL_STR(area, 0.0)) /* Not collinear */
	    break;
	  else
	    j++;		/* Collinear.. go to next point */
	}
      
      /* Now remove all the intermediate points */
      
      delta = j - (i + 2);
      if (delta > 0)
	{
	  nA -= delta;
	  for (j = i+1; j <= nA; j++)
	      pch[j].v = pch[j + delta].v;
	}

      i++;
   }

  *npoints = nA;
  pch[nA].valid = FALSE;
  return 0;
}


#ifdef OLD_DEFINITION

/* pchain[0] lies completely within the polygon part[0].. Hence spilt */
/* the single into 2 partitions by breaking the loop */
int split_into_two_partitions(part_t *part,
			      double opoly[][2], int npoints,
			      double cpoly[][2], int nchp)
{
  register int i, j;
  int k, l, itype, oposmin, oposmax, cposmax, cposmin;
  double s0, t0, s1, t1, tmin, tmax;
  point_typ v0, v1, p0, p1, q0, q1, q2, q3;
  point_typ cent;
  
  /* Find the centroid of the inner chain to get the vertical line to */
  /* use for intersections */

  part[1].pc = (point_typ *) new(MAX_PARTITION_SIZE*sizeof(point_typ));

  cent.x = cent.y = 0.0;
  for (i = 0; i < nchp; i++)
    {
      cent.x += cpoly[i][0];
      cent.y += cpoly[i][1];
    }
  cent.x /= nchp;
  cent.y /= nchp;
  
  /* Now find the two positions on the inner chain from where to */
  /* intersect the outer polygon */

#ifdef CENTROID_METHOD
  q0.x = 2.0;
  q2.x = 2.0;
  for (i = 0; i < nchp; i++)
    if ((cpoly[i][1] > cent.y) && 
	(fabs(cpoly[i][0] - cent.x) < fabs(q0.x - cent.x)))
      {
	q0.x = cpoly[i][0];
	q0.y = cpoly[i][1];
      }
    else if ((cpoly[i][1] < cent.y) && 
	     (fabs(cpoly[i][0] - cent.x) < fabs(q2.x - cent.x)))
      {
	q2.x = cpoly[i][0];
	q2.y = cpoly[i][1];
      }   
  
#else
  q3.x = cent.x;
  q3.y = -1.0;
  q1.x = cent.x;
  q1.y = 2.0;

  tmin = 2.0;
  tmax = -2.0;
  for (i = 0; i < nchp; i++)
    {
      v0.x = cpoly[i][0];
      v0.y = cpoly[i][1];
      v1.x = cpoly[i+1][0];
      v1.y = cpoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q3, &q1, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	{
	  if (t0 > tmax)
	    {
	      tmax = t0;
	      cposmax = i;
	    }
	  if (t0 < tmin)
	    {
	      tmin = t0;
	      cposmin = i;
	    }
	}
    }
  
  /* Get the two actual points on hte inner curve now */
  
  q0.x = cpoly[cposmax][0];
  q0.y = cpoly[cposmax][1];

  q2.x = cpoly[cposmin][0];
  q2.y = cpoly[cposmin][1];
  
#endif

  /* Now (hopefully), the upper endpoint is q0, and the lower andpoint */
  /* is q2 */
  
  q1.x = q0.x;
  q1.y = 2.0;
      
  tmin = 2.0;
  tmax = -2.0;
  
  /* Detect where this segment q0q1 intersects the outer polygon */
  for (i = 0; i < npoints; i++)
    {
      v0.x = opoly[i][0];
      v0.y = opoly[i][1];
      v1.x = opoly[i+1][0];
      v1.y = opoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q0, &q1, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	if (t0 < tmin)
	  {
	    tmin = t0;
	    oposmax = i;
	  }
    }

  q3.x = q2.x;
  q3.y = -2.0;
      
  tmin = 2.0;
  tmax = -2.0;

  /* Detect where this segment q0q1 intersects the outer polygon */
  for (i = 0; i < npoints; i++)
    {
      v0.x = opoly[i][0];
      v0.y = opoly[i][1];
      v1.x = opoly[i+1][0];
      v1.y = opoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q2, &q3, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	if (t0 < tmin)
	  {
	    tmin = t0;
	    oposmin = i;
	  }
    }

  /* now we have the values oposmax, oposmin, cposmax, cposmin..*/
  /* Generate the two partitions all over again */
  
  k = 0;
  for (i = oposmin; i != oposmax; i = (i+1)%npoints)
    {
      part[0].pc[k].x = opoly[i][0];
      part[0].pc[k].y = opoly[i][1];
      k++;
    }
  part[0].pc[k].x = opoly[oposmax][0];
  part[0].pc[k].y = opoly[oposmax][1];  
  k++;
  
  for (i = cposmax; i != cposmin; i = (i + nchp - 1)%nchp)
    {
      part[0].pc[k].x = cpoly[i][0];
      part[0].pc[k].y = cpoly[i][1];
      k++;
    }
  part[0].pc[k].x = cpoly[cposmin][0];
  part[0].pc[k].y = cpoly[cposmin][1];  
  k++;
  part[0].pc[k].x = part[0].pc[0].x;
  part[0].pc[k].y = part[0].pc[0].y;
  part[0].npoints = k;
  
  /* Now generate partition 2 */
  
  k = 0;
  for (i = oposmax; i != oposmin; i = (i+1)%npoints)
    {
      part[1].pc[k].x = opoly[i][0];
      part[1].pc[k].y = opoly[i][1];
      k++;
    }
  part[1].pc[k].x = opoly[oposmin][0];
  part[1].pc[k].y = opoly[oposmin][1];  
  k++;
  
  for (i = cposmin; i != cposmax; i = (i + nchp - 1)%nchp)
    {
      part[1].pc[k].x = cpoly[i][0];
      part[1].pc[k].y = cpoly[i][1];
      k++;
    }
  part[1].pc[k].x = cpoly[cposmax][0];
  part[1].pc[k].y = cpoly[cposmax][1];  
  k++;
  part[1].pc[k].x = part[1].pc[0].x;
  part[1].pc[k].y = part[1].pc[0].y;
  part[1].npoints = k;

  return 0;

}

#else  /* New partitioning method!! */

/* pchain[0] lies completely within the polygon part[0].. Hence spilt */
/* the single into 2 partitions by breaking the loop */
int split_into_two_partitions(part_t *part,
			      double opoly[][2], int npoints,
			      double cpoly[][2], int nchp)
{
  register int i, j;
  int k, l, itype, oposmin, oposmax, cposmax, cposmin;
  double s0, t0, s1, t1, tmin, tmax;
  point_typ v0, v1, p0, p1, q0, q1, q2, q3;
  point_typ cent, q0_sav, q1_sav, q2_sav, q3_sav;
  
  /* Find the centroid of the inner chain to get the vertical line to */
  /* use for intersections */

  part[1].pc = (point_typ *) new(MAX_PARTITION_SIZE*sizeof(point_typ));

  cent.x = cent.y = 0.0;
  for (i = 0; i < nchp; i++)
    {
      cent.x += cpoly[i][0];
      cent.y += cpoly[i][1];
    }
  cent.x /= nchp;
  cent.y /= nchp;
  
  /* Now find the two positions on the inner chain from where to */
  /* intersect the outer polygon */

#ifdef CENTROID_METHOD
  q0.x = 2.0;
  q2.x = 2.0;
  for (i = 0; i < nchp; i++)
    if ((cpoly[i][1] > cent.y) && 
	(fabs(cpoly[i][0] - cent.x) < fabs(q0.x - cent.x)))
      {
	q0.x = cpoly[i][0];
	q0.y = cpoly[i][1];
      }
    else if ((cpoly[i][1] < cent.y) && 
	     (fabs(cpoly[i][0] - cent.x) < fabs(q2.x - cent.x)))
      {
	q2.x = cpoly[i][0];
	q2.y = cpoly[i][1];
      }   
  
#else
  q3.x = cent.x;
  q3.y = -1.0;
  q1.x = cent.x;
  q1.y = 2.0;

  tmin = 2.0;
  tmax = -2.0;
  for (i = 0; i < nchp; i++)
    {
      v0.x = cpoly[i][0];
      v0.y = cpoly[i][1];
      v1.x = cpoly[i+1][0];
      v1.y = cpoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q3, &q1, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	{
	  if (t0 > tmax)
	    {
	      tmax = t0;
	      cposmax = i;
	    }
	  if (t0 < tmin)
	    {
	      tmin = t0;
	      cposmin = i;
	    }
	}
    }
  
  /* Get the two actual points on hte inner curve now */
  
  q0.x = (1 - tmax)*q3.x + tmax*q1.x;
  q0.y = (1 - tmax)*q3.y + tmax*q1.y;
  q2.x = (1 - tmin)*q3.x + tmin*q1.x;
  q2.y = (1 - tmin)*q3.y + tmin*q1.y;
  q0_sav = q0;
  q2_sav = q2;

#endif

  /* Now (hopefully), the upper endpoint is q0, and the lower andpoint */
  /* is q2 */
  
  q1.x = q0.x;
  q1.y = 2.0;
      
  tmin = 2.0;
  tmax = -2.0;
  
  /* Detect where this segment q0q1 intersects the outer polygon */
  for (i = 0; i < npoints; i++)
    {
      v0.x = opoly[i][0];
      v0.y = opoly[i][1];
      v1.x = opoly[i+1][0];
      v1.y = opoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q0, &q1, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	if (t0 < tmin)
	  {
	    tmin = t0;
	    oposmax = i;
	  }
    }

  q1_sav.x = (1 - tmin)*q0.x + tmin*q1.x;
  q1_sav.y = (1 - tmin)*q0.y + tmin*q1.y;


  q3.x = q2.x;
  q3.y = -2.0;
      
  tmin = 2.0;
  tmax = -2.0;

  /* Detect where this segment q2q3 intersects the outer polygon */
  for (i = 0; i < npoints; i++)
    {
      v0.x = opoly[i][0];
      v0.y = opoly[i][1];
      v1.x = opoly[i+1][0];
      v1.y = opoly[i+1][1];

      if (segments_intersect(&v0, &v1, &q2, &q3, &itype,
			     &s0, &s1, &t0, &t1,
			     &p0, &p1) > 0)
	if (t0 < tmin)
	  {
	    tmin = t0;
	    oposmin = i;
	  }
    }

  q3_sav.x = (1 - tmin)*q2.x + tmin*q3.x;
  q3_sav.y = (1 - tmin)*q2.y + tmin*q3.y;

  /* now we have the values oposmax, oposmin, cposmax, cposmin..*/
  /* Generate the two partitions all over again */
  
  k = 0;
  for (i = (oposmin+1)%npoints; i != oposmax; i = (i+1)%npoints)
    {
      part[0].pc[k].x = opoly[i][0];
      part[0].pc[k].y = opoly[i][1];
      k++;
    }
  part[0].pc[k].x = opoly[oposmax][0];
  part[0].pc[k].y = opoly[oposmax][1];  
  k++;

/* Add the new segment q1-q0 in between */

  part[0].pc[k] = q1_sav;
  k++;  
  part[0].pc[k] = q0_sav;
  k++;  

  for (i = cposmax; i != cposmin; i = (i + nchp - 1)%nchp)
    {
      part[0].pc[k].x = cpoly[i][0];
      part[0].pc[k].y = cpoly[i][1];
      k++;
    }
  part[0].pc[k] = q2_sav;
  k++;  
  part[0].pc[k] = q3_sav;
  k++;  
  part[0].npoints = k;
  
 
  /* Now generate partition 2 */
  
  k = 0;
  for (i = (oposmax+1)%npoints; i != oposmin; i = (i+1)%npoints)
    {
      part[1].pc[k].x = opoly[i][0];
      part[1].pc[k].y = opoly[i][1];
      k++;
    }
  part[1].pc[k].x = opoly[oposmin][0];
  part[1].pc[k].y = opoly[oposmin][1];  
  k++;
  part[1].pc[k++] = q3_sav;
  part[1].pc[k++] = q2_sav;

  for (i = cposmin; i != cposmax; i = (i + nchp - 1)%nchp)
    {
      part[1].pc[k].x = cpoly[i][0];
      part[1].pc[k].y = cpoly[i][1];
      k++;
    }
  part[1].pc[k++] = q0_sav;
  part[1].pc[k++] = q1_sav;
  /*  part[1].pc[k].x = part[1].pc[0].x;
      part[1].pc[k].y = part[1].pc[0].y;
  */
  part[1].npoints = k;

  remove_duplicate_points_in_partition(&part[0]);
  remove_duplicate_points_in_partition(&part[1]);

  part[0].whichside = S_RIGHT;
  part[1].whichside = S_RIGHT;

#ifdef DEBUG_PRINT_POLY
  printf("hole_removal: partition 0\n");
  for (i = 0; i < part[0].npoints; i++)
    printf("%lf %lf\n", part[0].pc[i].x, part[0].pc[i].y);

  printf("hole_removal: partition 1\n");
  for (i = 0; i < part[1].npoints; i++)
    printf("%lf %lf\n", part[1].pc[i].x, part[1].pc[i].y);

#endif
  
  return 0;

}

#endif

#ifdef DEBUG

int read_data()
{
  register int i, j;
  int n, m;
  
  scanf("%d", &n);
  global.nseg = n;
  for (i = 0; i < n; i++)
    {
      scanf("%lf%lf", &edge[i].v0.x, &edge[i].v0.y);
      if (i == 0)
        {
          edge[n-1].v1.x = edge[i].v0.x;
          edge[n-1].v1.y = edge[i].v0.y;
        }
      else
        {
          edge[i - 1].v1.x = edge[i].v0.x;
          edge[i - 1].v1.y = edge[i].v0.y;
        }       
    }

  scanf("%d", &n);		/* nchains */
  for (i = 0; i < n; i++)
    {
      scanf("%d", &m);
      pchain[i] = (pchain_t *)
	new(MAX_CHAIN_LENGTH*sizeof(pchain_t));
      for (j = 0; j < m; j++)
	{
	  scanf("%lf%lf", &pchain[i][j].v.x, &pchain[i][j].v.y);
	  pchain[i][j].valid = TRUE;
	}
      pchain[i][j].valid = FALSE;
    }
  
  return n;
}

#endif




