#include <intersect.h>
#include <math.h>

#define NORMALIZE(v, denom) v.x /= denom;\
                            v.y /= denom

#ifdef NEW_VERSION

static int intr_are_vectors_collinear(vector_t *v0, vector_t *v1)
{
  vector_t n0, n1;
  int denom0, denom1;
  
  n0.x = v0->x; n0.y = v0->y;
  n0.x = v1->x; n1.y = v1->y;
  denom0 = sqrt(DOT(n0, n0));
  denom1 = sqrt(DOT(n1, n1));

  NORMALIZE(n0, denom0);
  NORMALIZE(n1, denom1);

  
  if (FP_EQUAL_PAR(CROSS_VEC(n0, n1), 0.0, 1.0e-6)) /* collinear */
    return 1;
  else
    return 0;
}


static int intr_is_collinear(point_typ *p0,point_typ *p1,point_typ *p2)
{
  vector_t n0, n1;
  double denom0, denom1;
  
  n0.x = p1->x - p0->x;
  n0.y = p1->y - p0->y;
  n1.x = p2->x - p0->x;
  n1.y = p2->y - p0->y;
  denom0 = sqrt(DOT(n0, n0));
  denom1 = sqrt(DOT(n1, n1));

  NORMALIZE(n0, denom0);
  NORMALIZE(n1, denom1);

  return (FP_EQUAL_PAR(CROSS_VEC(n0, n1), 0.0, 1.0e-6)); /* collinear */
}

#else

static int intr_is_collinear(point_typ *p0,point_typ *p1,point_typ *p2)
{
  vector_t n0, n1;
  double denom0;
  double dist;

  n0.x = p1->x - p0->x;
  n0.y = p1->y - p0->y;
  n1.x = p2->x - p0->x;
  n1.y = p2->y - p0->y;

  denom0 = sqrt(DOT(n0, n0));
  NORMALIZE(n0, denom0);
  
  dist = CROSS_VEC(n0, n1);
  return (fabs(dist) <= 1.0e-7);
/*
  return (fabs(CROSS(*p0, *p1, *p2)) <=  I_TOLERANCE); ... collinear */
}


static int intr_are_vectors_collinear(point_typ *p0, point_typ *p1,
				      point_typ *q0, point_typ *q1)
{
  vector_t v0, v1;
  double denom0, denom1;

  if (intr_is_collinear(p0, p1, q0) &&
      intr_is_collinear(p0, p1, q1))
    return 1;
  
  v0.x = p1->x - p0->x;
  v0.y = p1->y - p0->y;
  v1.x = q1->x - q0->x;
  v1.y = q1->y - q0->y;

  denom0 = sqrt(DOT(v0, v0));
  denom1 = sqrt(DOT(v1, v1));

  NORMALIZE(v0, denom0);
  NORMALIZE(v1, denom1);

  if (fabs((double) CROSS_VEC(v0, v1)) <=  I_TOLERANCE) /* collinear */
    return 1;
  else
    return 0;
}

#endif

int intr_are_vectors_coincident(point_typ *p0, point_typ *p1, 
				   point_typ *q0, point_typ *q1)
{
  vector_t v0, v1;
  
  v0.x = p1->x - p0->x;
  v0.y = p1->y - p0->y;
  v1.x = q1->x - q0->x;
  v1.y = q1->y - q0->y;

  return (intr_are_vectors_collinear(p0, p1, q0, q1) &&
	  intr_is_collinear(p0, p1, q0));
}


/* routine to determine whether 2 segments intersect or not. return */
/* the s, t values of the 1/2 intersections. Also output the point(s) */
/* of intersection */
/* Return value: the number of intersections */
int segments_intersect(point_typ *u0, point_typ *u1, point_typ *v0, 
		       point_typ *v1, int *intersect_type,
		       double *ss0, double *ss1, 
		       double *tt0, double *tt1,
		       point_typ *p0, point_typ *p1)  
{
  double s, t, denom, s1, t1;
  double m, n;
  vector_t vec0, vec1;
  point_typ t0;

  vec0.x = u1->x - u0->x;
  vec0.y = u1->y - u0->y;
  
  vec1.x = v1->x - v0->x;
  vec1.y = v1->y - v0->y;
  
  if (intr_are_vectors_collinear(u0, u1, v0, v1)) /* collinear vectors */
    {
      if (intr_is_collinear(u0, u1, v0)) /* Coincident vectors */
	{
	  /* First, calculate the s values for each of the endpoints */
	  /* v0 and v1 w.r.t. the segment (u0, u1) */
	  

	  if (fabs(vec0.x) > fabs(vec0.y)) /* Non-vertical line segment */
	    {
	      s = (v0->x - u0->x)/(u1->x - u0->x);
	      s1 = (v1->x - u0->x)/(u1->x - u0->x);
	    }
	  else
	    {
	      s = (v0->y - u0->y)/(u1->y - u0->y);
	      s1 = (v1->y - u0->y)/(u1->y - u0->y);
	    }	    
	      
	  if (FP_EQUAL(s, 0.0))
	    s = 0.0;
	  else if (FP_EQUAL(s, 1.0))
	    s = 1.0;
	  if (FP_EQUAL(s1, 0.0))
	    s1 = 0.0;
	  else if (FP_EQUAL(s1, 1.0))
	    s1 = 1.0;

	  if (((s > 1.0) && (s1 > 1.0)) || ((s < 0.0) && (s1 < 0.0))) 
	    return 0;

	  /* Similarly, get the t-values corresponding to the */
	  /* intersection w.r.t. (v0, v1) */
	  
	  /* if (fabs(vec1.x) >= I_TOLERANCE) */
	  if (fabs(vec1.x) > fabs(vec1.y) ) /* Non-horizontal line segment */
	    {
	      t = (u0->x - v0->x)/(v1->x - v0->x);
	      t1 = (u1->x - v0->x)/(v1->x - v0->x);
	    }
	  else
	    {
	      t = (u0->y - v0->y)/(v1->y - v0->y);
	      t1 = (u1->y - v0->y)/(v1->y - v0->y);
	    }	    

	  if (FP_EQUAL(t, 0.0))
	    t = 0.0;
	  else if (FP_EQUAL(t, 1.0))
	    t = 1.0;
	  if (FP_EQUAL(t1, 0.0))
	    t1 = 0.0;
	  else if (FP_EQUAL(t1, 1.0))
	    t1 = 1.0;
	  
	  /* Now return proper endpoint values depending upon whether */
	  /* the vectors are parallel or antiparallel */

	  if (DOT(vec0, vec1) > 0.0) /* parallel vectors */
	    {
	      *intersect_type = I_PAR;
	      if (s < 0.0)
		{
		  if (s1 == 0.0)
		    {
		      *p0 = *u0;
		      *ss0 = 0.0;
		      *tt0 = 0.0;
		      return 1;
		    }
		  else if (s1 <= 1.0)	/* Partial overlap */
		    {
		      *p0 = *u0;
		      p1->x = u0->x + s1*(u1->x - u0->x);
		      p1->y = u0->y + s1*(u1->y - u0->y);
		      *ss0 = 0.0;
		      *ss1 = s1;
		      *tt0 = t;
		      *tt1 = 1;
		      return 2;
		    }
		  else		/* u0-u1 fully contained within v0-v1 */
		    {
		      *p0 = *u0;
		      *p1 = *u1;
		      *ss0 = 0.0;
		      *ss1 = 1.0;
		      *tt0 = t;
		      *tt1 = t1;
		      return 2;
		    } 		  
		}
	      else if ((s >= 0.0) && (s < 1.0))
		{		/* v0-v1 begins inside u0-u1 */
		  if (s1 > 1.0)	/* Partial overlap */
		    {
		      p0->x = u0->x + s*(u1->x - u0->x);
		      p0->y = u0->y + s*(u1->y - u0->y);
		      *p1 = *u1;
		      *ss0 = s;
		      *ss1 = 1.0;
		      *tt0 = 0.0;
		      *tt1 = t1;
		    }
		  else		/* v0-v1 fully contained within u0-u1 */
		    {
		      p0->x = u0->x + s*(u1->x - u0->x);
		      p0->y = u0->y + s*(u1->y - u0->y);		      
		      p1->x = u0->x + s1*(u1->x - u0->x);
		      p1->y = u0->y + s1*(u1->y - u0->y);
		      *ss0 = s;
		      *ss1 = s1;
		      *tt0 = 0.0;
		      *tt1 = 1.0;
		    } 
		  return 2;
		}
	      else if (s == 1.0)
		{
		  *p0 = *u1;
		  *ss0 = 1.0;
		  *tt0 = 0.0;
		  return 1;
		}
	      else
		fprintf(stderr, "segment_intersect: coll:error\n");
	    }
	  else			/* Anti-parallel vectors */
	    { 
	      *intersect_type = I_ANTIPAR;
	      if (s > 1.0)
		{
		  if (s1 == 1.0)
		    {
		      *p0 = *u1;
		      *ss0 = 1.0;
		      *tt0 = t1;
		      return 1;
		    }
		  else if (s1 >= 0.0)	/* Partial overlap */
		    {
		      p0->x = u0->x + s1*(u1->x - u0->x);
		      p0->y = u0->y + s1*(u1->y - u0->y);
		      *p1 = *u1;
		      *ss0 = s1;
		      *ss1 = 1.0;
		      *tt1 = 1.0;
		      *tt0 = t1;
		      return 2;
		    }
		  else		/* u0-u1 fully contained within v0-v1 */
		    {
		      *p0 = *u0;
		      *p1 = *u1;
		      *ss0 = 0.0;
		      *ss1 = 1.0;
		      *tt1 = t;
		      *tt0 = t1;
		      return 2;
		    } 		  
		}
	      else if (s > 0.0) /* v0-v1 begins inside u0-u1 */
		{
		  if (s1 < 0.0)	/* Partial overlap */
		    {
		      *p0 = *u0;
		      p1->x = u0->x + s*(u1->x - u0->x);
		      p1->y = u0->y + s*(u1->y - u0->y);
		      *ss0 = 0.0;
		      *ss1 = s;
		      *tt1 = t;
		      *tt0 = 0.0;
		    }
		  else		/* v0-v1 fully contained within u0-u1 */
		    {
		      p0->x = u0->x + s1*(u1->x - u0->x);
		      p0->y = u0->y + s1*(u1->y - u0->y);		      
		      p1->x = u0->x + s*(u1->x - u0->x);
		      p1->y = u0->y + s*(u1->y - u0->y);
		      *ss0 = s1;
		      *ss1 = s;
		      *tt1 = 1.0;
		      *tt0 = 0.0;
		    } 
		  return 2;
		}
	      else if (s == 0.0)
		{
		  *p0 = *u0;
		  *ss0 = 0.0;
		  *tt0 = 0.0;
		  return 1;
		}
	      else
		fprintf(stderr, "segment_intersect: coll:error\n");
	    }	      
	  
	}
      else			/* distinct parallel segments */
	return 0;
    }
  else				/* vectors are not collinear */
    {
      denom = -vec0.x * vec1.y + vec0.y * vec1.x;
      m = v0->x - u0->x;
      n = v0->y - u0->y;

      s = (-vec1.y*m + n*vec1.x)/denom;
      t = (n*vec0.x - vec0.y*m)/denom;
	
      if (FP_EQUAL_POINT(s, 0.0))
	s = 0.0;
      if (FP_EQUAL_POINT(t, 0.0))
	t = 0.0;
      if (FP_EQUAL_POINT(s, 1.0))
	s = 1.0;
      if (FP_EQUAL_POINT(t, 1.0))
	t = 1.0;

/*
      if (t < 0.0)
	{
	  t0.x = v0->x + t*(v1->x - v0->x);
	  t0.y = v0->y + t*(v1->y - v0->y);
	  
	  if (FP_EQUAL_POINT(t0.x, v0->x) &&
	      FP_EQUAL_POINT(t0.y, v0->y))
	    t = 0.0;
	}
      else if (t > 1.0)
	{
	  t0.x = v0->x + t*(v1->x - v0->x);
	  t0.y = v0->y + t*(v1->y - v0->y);
	  
	  if (FP_EQUAL_POINT(t0.x, v1->x) &&
	      FP_EQUAL_POINT(t0.y, v1->y))
	    t = 1.0;
	}

*/      
      if ((s > 1.0) || (s < 0.0 ) 
	   || (t > 1.0 ) || (t < 0.0 ))
	return 0;		/* do not intersect */
      
      *ss0 = s;
      *tt0 = t;
      if ((s == 0.0) || (s == 1.0)) /* V0 just touches V1 */
	{
	  if ((t == 0.0) || (t == 1.0))
	    *intersect_type = I_ST_ON;
	  else
	    *intersect_type = I_S_ON;
	}
      else
	{
	  if ((t == 0.0) || (t == 1.0))
	    *intersect_type = I_T_ON;
	  else
	    *intersect_type = I_NONE_ON;
	}	

      p0->x = u0->x + s*(u1->x - u0->x);
      p0->y = u0->y + s*(u1->y - u0->y);
      return 1;
    }
  
  return 0;
}


int shady_segments_intersect(point_typ *u0,point_typ *u1,point_typ *v0,
		       point_typ *v1, int *intersect_type,
		       double *ss0, double *ss1, 
		       double *tt0, double *tt1,
		       point_typ *p0, point_typ *p1,
		       int is_first_last)  
{
  double s, t, denom, s1, t1;
  double m, n;
  vector_t vec0, vec1;
  point_typ t0;
  double lu, lv;

  vec0.x = u1->x - u0->x;
  vec0.y = u1->y - u0->y;
  
  vec1.x = v1->x - v0->x;
  vec1.y = v1->y - v0->y;
  
  if (intr_are_vectors_collinear(u0, u1, v0, v1)) /* collinear vectors */
    {
      if (intr_is_collinear(u0, u1, v0)) /* Coincident vectors */
	{
	  /* First, calculate the s values for each of the endpoints */
	  /* v0 and v1 w.r.t. the segment (u0, u1) */
	  
	  /* if (fabs(vec0.x) >= I_TOLERANCE) */
	  if (fabs(vec0.x) > fabs(vec0.y)) /* Non-vertical line segment */
	    {
	      s = (v0->x - u0->x)/(u1->x - u0->x);
	      s1 = (v1->x - u0->x)/(u1->x - u0->x);
	    }
	  else
	    {
	      s = (v0->y - u0->y)/(u1->y - u0->y);
	      s1 = (v1->y - u0->y)/(u1->y - u0->y);
	    }	    
	      
	  if (FP_EQUAL(s, 0.0))
	    s = 0.0;
	  else if (FP_EQUAL(s, 1.0))
	    s = 1.0;
	  if (FP_EQUAL(s1, 0.0))
	    s1 = 0.0;
	  else if (FP_EQUAL(s1, 1.0))
	    s1 = 1.0;

	  if (((s > 1.0) && (s1 > 1.0)) || ((s < 0.0) && (s1 < 0.0))) 
	    return 0;

	  /* Similarly, get the t-values corresponding to the */
	  /* intersection w.r.t. (v0, v1) */
	  
	  /* if (fabs(vec1.x) >= I_TOLERANCE) */
	  if (fabs(vec1.x) > fabs(vec1.y) ) /* Non-horizontal line segment */
	    {
	      t = (u0->x - v0->x)/(v1->x - v0->x);
	      t1 = (u1->x - v0->x)/(v1->x - v0->x);
	    }
	  else
	    {
	      t = (u0->y - v0->y)/(v1->y - v0->y);
	      t1 = (u1->y - v0->y)/(v1->y - v0->y);
	    }	    

	  if (FP_EQUAL(t, 0.0))
	    t = 0.0;
	  else if (FP_EQUAL(t, 1.0))
	    t = 1.0;
	  if (FP_EQUAL(t1, 0.0))
	    t1 = 0.0;
	  else if (FP_EQUAL(t1, 1.0))
	    t1 = 1.0;
	  
	  /* Now return proper endpoint values depending upon whether */
	  /* the vectors are parallel or antiparallel */

	  if (DOT(vec0, vec1) > 0.0) /* parallel vectors */
	    {
	      *intersect_type = I_PAR;
	      if (s < 0.0)
		{
		  if (s1 == 0.0)
		    {
		      *p0 = *u0;
		      *ss0 = 0.0;
		      *tt0 = 0.0;
		      return 1;
		    }
		  else if (s1 <= 1.0)	/* Partial overlap */
		    {
		      *p0 = *u0;
		      p1->x = u0->x + s1*(u1->x - u0->x);
		      p1->y = u0->y + s1*(u1->y - u0->y);
		      *ss0 = 0.0;
		      *ss1 = s1;
		      *tt0 = t;
		      *tt1 = 1;
		      return 2;
		    }
		  else		/* u0-u1 fully contained within v0-v1 */
		    {
		      *p0 = *u0;
		      *p1 = *u1;
		      *ss0 = 0.0;
		      *ss1 = 1.0;
		      *tt0 = t;
		      *tt1 = t1;
		      return 2;
		    } 		  
		}
	      else if ((s >= 0.0) && (s < 1.0))
		{		/* v0-v1 begins inside u0-u1 */
		  if (s1 > 1.0)	/* Partial overlap */
		    {
		      p0->x = u0->x + s*(u1->x - u0->x);
		      p0->y = u0->y + s*(u1->y - u0->y);
		      *p1 = *u1;
		      *ss0 = s;
		      *ss1 = 1.0;
		      *tt0 = 0.0;
		      *tt1 = t1;
		    }
		  else		/* v0-v1 fully contained within u0-u1 */
		    {
		      p0->x = u0->x + s*(u1->x - u0->x);
		      p0->y = u0->y + s*(u1->y - u0->y);		      
		      p1->x = u0->x + s1*(u1->x - u0->x);
		      p1->y = u0->y + s1*(u1->y - u0->y);
		      *ss0 = s;
		      *ss1 = s1;
		      *tt0 = 0.0;
		      *tt1 = 1.0;
		    } 
		  return 2;
		}
	      else if (s == 1.0)
		{
		  *p0 = *u1;
		  *ss0 = 1.0;
		  *tt0 = 0.0;
		  return 1;
		}
	      else
		fprintf(stderr, "segment_intersect: coll:error\n");
	    }
	  else			/* Anti-parallel vectors */
	    {
	      *intersect_type = I_ANTIPAR;
	      if (s > 1.0)
		{
		  if (s1 == 1.0)
		    {
		      *p0 = *u1;
		      *ss0 = 1.0;
		      *tt0 = t1;
		      return 1;
		    }
		  else if (s1 >= 0.0)	/* Partial overlap */
		    {
		      p0->x = u0->x + s1*(u1->x - u0->x);
		      p0->y = u0->y + s1*(u1->y - u0->y);
		      *p1 = *u1;
		      *ss0 = s1;
		      *ss1 = 1.0;
		      *tt1 = 1.0;
		      *tt0 = t1;
		      return 2;
		    }
		  else		/* u0-u1 fully contained within v0-v1 */
		    {
		      *p0 = *u0;
		      *p1 = *u1;
		      *ss0 = 0.0;
		      *ss1 = 1.0;
		      *tt1 = t;
		      *tt0 = t1;
		      return 2;
		    } 		  
		}
	      else if (s > 0.0) /* v0-v1 begins inside u0-u1 */
		{
		  if (s1 < 0.0)	/* Partial overlap */
		    {
		      *p0 = *u0;
		      p1->x = u0->x + s*(u1->x - u0->x);
		      p1->y = u0->y + s*(u1->y - u0->y);
		      *ss0 = 0.0;
		      *ss1 = s;
		      *tt1 = t;
		      *tt0 = 0.0;
		    }
		  else		/* v0-v1 fully contained within u0-u1 */
		    {
		      p0->x = u0->x + s1*(u1->x - u0->x);
		      p0->y = u0->y + s1*(u1->y - u0->y);		      
		      p1->x = u0->x + s*(u1->x - u0->x);
		      p1->y = u0->y + s*(u1->y - u0->y);
		      *ss0 = s1;
		      *ss1 = s;
		      *tt1 = 1.0;
		      *tt0 = 0.0;
		    } 
		  return 2;
		}
	      else if (s == 0.0)
		{
		  *p0 = *u0;
		  *ss0 = 0.0;
		  *tt0 = 0.0;
		  return 1;
		}
	      else
		fprintf(stderr, "segment_intersect: coll:error\n");
	    }	      
	  
	}
      else			/* distinct parallel segments */
	return 0;
    }
  else				/* vectors are not collinear */
    {
      denom = -vec0.x * vec1.y + vec0.y * vec1.x;
      m = v0->x - u0->x;
      n = v0->y - u0->y;

      lu = sqrt(DOT(vec0, vec0));
      lv = sqrt(DOT(vec1, vec1));

      s = (-vec1.y*m + n*vec1.x)/denom;
      t = (n*vec0.x - vec0.y*m)/denom;
	
      if (FP_EQUAL_POINT(s, 0.0))
	s = 0.0;
      if (FP_EQUAL_POINT(t, 0.0))
	t = 0.0;
      if (FP_EQUAL_POINT(s, 1.0))
	s = 1.0;
      if (FP_EQUAL_POINT(t, 1.0))
	t = 1.0;

      if ((is_first_last == I_FIRSTPT) && (t < 0.0))
	{
	  t0.x = v0->x + t*(v1->x - v0->x);
	  t0.y = v0->y + t*(v1->y - v0->y);
	  
	  if (FP_EQUAL_POINT(t0.x, v0->x) &&
	      FP_EQUAL_POINT(t0.y, v0->y))
	    t = 0.0;
	}
      else if ((is_first_last == I_LASTPT) && (t > 1.0))
	{
	  t0.x = v0->x + t*(v1->x - v0->x);
	  t0.y = v0->y + t*(v1->y - v0->y);
	  
	  if (FP_EQUAL_POINT(t0.x, v1->x) &&
	      FP_EQUAL_POINT(t0.y, v1->y))
	    t = 1.0;
	}

      
      if ((s > 1.0) || (s < 0.0 ) 
	   || (t > 1.0 ) || (t < 0.0 ))
	return 0;		/* do not intersect */
      
      *ss0 = s;
      *tt0 = t;
      if ((s == 0.0) || (s == 1.0)) /* V0 just touches V1 */
	{
	  if ((t == 0.0) || (t == 1.0))
	    *intersect_type = I_ST_ON;
	  else
	    *intersect_type = I_S_ON;
	}
      else
	{
	  if ((t == 0.0) || (t == 1.0))
	    *intersect_type = I_T_ON;
	  else
	    *intersect_type = I_NONE_ON;
	}	

      p0->x = u0->x + s*(u1->x - u0->x);
      p0->y = u0->y + s*(u1->y - u0->y);
      return 1;
    }
  
  return 0;
}


