#include <defs.h>
#include <intersect.h>
#include <math.h>

#define PAR_EDGE 2
#define DISTANCE(p1, p2) (sqrt((double) (p2[0] - p1[0])*(p2[0] - p1[0])\
			       + (p2[1] - p1[1])*(p2[1] - p1[1])))


/*********** Global variables used  *******************/

extern node_t qs[];		/* Query structure */
extern trap_t tr[];		/* Trapezoid structure */
extern segment_t seg[SEGSIZE];	/* Segment table */

extern einfo_t edge[];
extern pchain_t *pchain[];
extern struct global_s global;
extern int is_planar_overlap;

static int initialise_in_plane(int nseg)
{
  register int i;

  for (i = 1; i <= nseg; i++)
    seg[i].is_inserted = FALSE;
  generate_random_ordering(nseg);
  
  return 0;
}

/* Intersect the polygon A and polygon B. Interpret B as the chain, */
/* and return the resulting partitions of A */

int intersect_polygonA_using_chainB(int nA, 
				    double polyA[][2],
				    int nB,
				    double chB[][2],
				    part_t *partn)
{
  register int i, j;
  int n, npartitions, nchain;
  LIST_POINT2D *lp;
  TRIM_PORTION_TYPE *trimp;
  point_typ p;


  global.nseg = n = nA;		/* #points */
  for (i = 0; i < n; i++)
    {
      edge[i].v0.x = polyA[i][0];
      edge[i].v0.y = polyA[i][1];
      if (i == 0)
        {
          edge[n-1].v1.x = edge[i].v0.x;
          edge[n-1].v1.y = edge[i].v0.y;
        }
      else
        {
          edge[i - 1].v1.x = edge[i].v0.x;
          edge[i - 1].v1.y = edge[i].v0.y;
        }       
#ifdef DEBUG_PRINT_POLY
      printf("%lf %lf\n", edge[i].v0.x, edge[i].v0.y);
      fflush(stdout);      
#endif
    }

#ifdef DEBUG_PRINT_POLY
      printf("iA_using_B: chain:\n");
      for (j = 0; j <= nB; j++)
	printf("%lf %lf\n", chB[j][0], chB[j][1]);
      fflush(stdout);
#endif

#ifdef CHECK_SIMPLICITY
  if (!is_simple_polygon(PAR_EDGE, n))
    {
      fprintf(stderr, "partition_polygon: Polygon not simple\n");
      return -1;
    }
#endif


  /* Initialise the chain data-structure */

  pchain[0] = (pchain_t *)
    malloc(MAX_CHAIN_LENGTH*sizeof(pchain_t));

  for (i = 0; i <= nB; i++)
    {
      pchain[0][i].v.x = chB[i][0];
      pchain[0][i].v.y = chB[i][1];
      pchain[0][i].valid = TRUE;
    }
  pchain[0][i].valid = FALSE;

  nchain = 1;
  find_chain_intersections(&nchain);
  is_planar_overlap = TRUE;
  npartitions = trace_around_edges(partn);
  is_planar_overlap = FALSE;

  /* We have got the partitions. Return promptly without the slightest */
  /* of hesitation and ged'outa he're */

/*  
  for (i = 0; i < npartitions; i++)
    remove_duplicate_points_in_partition(partn[i]);
*/  
  /* Free up the space held up by the chains */
  for (i = 0; i < nchain; i++)
    free(pchain[i]);
  
  return npartitions;
}



/* Transform the (u, v) point in patch B into a (s, t) point w.r.t. */
/* the patch A */

static int transform_uv_to_st(TENSOR_PATCH *tpA, TENSOR_PATCH *tpB,
		       double u, double v, double *s, double *t)
{
  double x, y, z;
  
  EvalPointOnPatch(*tpB, u, v, &x, &y, &z);
  special_plane_case(tpA, x, y, z, s, t);
}


/* Add this partition as an additional trim-curve for the patch A */
static int add_trim_portion(TRIM_PATCH_INFO *tpA, part_t *partn)
{
  int n, j;
  point_typ p;
  TRIM_PORTION_TYPE *trimp;

  n = (tpA->NumTrimPortions)++;
  trimp = &(tpA->TrimPortion[n]);
  
  trimp->PiecePortion.NumOfVertices = partn->npoints;
  for (j = 0; j < partn->npoints; j++)
    {
      trimp->PiecePortion.poly_vertex[j][0] = partn->pc[j].x;
      trimp->PiecePortion.poly_vertex[j][1] = partn->pc[j].y;
    }
  trimp->PiecePortion.poly_vertex[j][0] = partn->pc[0].x;
  trimp->PiecePortion.poly_vertex[j][1] = partn->pc[0].y;
  
  /* Now find point inside each partition */
  
  find_point_inside_polygon(partn->npoints,
			    trimp->PiecePortion.poly_vertex, &p);
  trimp->PiecePortion.ptInsideRegion.u = p.x;
  trimp->PiecePortion.ptInsideRegion.v = p.y;

  return 0;
}

/* Add this partition as an additional trim-curve for the patch A */
static int add_trim_portion_from_double_poly(TRIM_PATCH_INFO *tpA, 
					     int npoints, 
					     double polyA[][2])
{
  int n, j;
  point_typ p;
  TRIM_PORTION_TYPE *trimp;

  n = (tpA->NumTrimPortions)++;
  trimp = &(tpA->TrimPortion[n]);
  
  trimp->PiecePortion.NumOfVertices = npoints;
  for (j = 0; j < npoints; j++)
    {
      trimp->PiecePortion.poly_vertex[j][0] = polyA[j][0];
      trimp->PiecePortion.poly_vertex[j][1] = polyA[j][1];
    }
  trimp->PiecePortion.poly_vertex[j][0] = polyA[0][0];
  trimp->PiecePortion.poly_vertex[j][1] = polyA[0][1];
  
  /* Now find point inside each partition */
  
  find_point_inside_polygon(npoints,
			    trimp->PiecePortion.poly_vertex, &p);
  trimp->PiecePortion.ptInsideRegion.u = p.x;
  trimp->PiecePortion.ptInsideRegion.v = p.y;

  return 0;
}

/* Add this partition as an additional trim-curve for the patch A */
static int add_trim_portion_from_vector_table(TRIM_PATCH_INFO *tpA, 
					      int npoints,
					      vector4D_t spoly[])
{
  int n, j;
  point_typ p;
  TRIM_PORTION_TYPE *trimp;

  n = (tpA->NumTrimPortions)++;
  trimp = &(tpA->TrimPortion[n]);
  
  trimp->PiecePortion.NumOfVertices = npoints;
  for (j = 0; j < npoints; j++)
    {
      trimp->PiecePortion.poly_vertex[j][0] = spoly[j][0];
      trimp->PiecePortion.poly_vertex[j][1] = spoly[j][1];
    }
  trimp->PiecePortion.poly_vertex[j][0] = spoly[0][0];
  trimp->PiecePortion.poly_vertex[j][1] = spoly[0][1];
  
  /* Now find point inside each partition */
  
  find_point_inside_polygon(npoints,
			    trimp->PiecePortion.poly_vertex, &p);
  trimp->PiecePortion.ptInsideRegion.u = p.x;
  trimp->PiecePortion.ptInsideRegion.v = p.y;

  return 0;
}


/* Special routine to return the point inside the common intersection */
/* region */
static int get_point_inside_intersection(part_t *partn, point_typ *p_in)
{
  register int i, j;
  double polyt[MAX_VERTEX_COUNT][2];
  
  for (j = 0; j < partn->npoints; j++)
    {
      polyt[j][0] = partn->pc[j].x;
      polyt[j][1] = partn->pc[j].y;
    }
  polyt[j][0] = polyt[0][0];
  polyt[j][1] = polyt[0][1];

  find_point_inside_polygon(j, polyt, p_in);
  return 0;
}


/* Handle the special case of boolean operation on planar patches */
/* which overlap */

int handle_planar_patch_operations(int op,TRIM_PATCH_INFO *tpA,
				   TRIM_PATCH_INFO *tpB,
				   POINT2D *p2d)
{
  register int i, j, k;
  int nA, nB, nC, delta, was_in, is_in, pos0, npartitions;
  double polyA[MAX_VERTEX_COUNT][2], polyB[MAX_VERTEX_COUNT][2],
         polyC[MAX_VERTEX_COUNT][2]; 
  double tmp0[2], area, len, frac, s0, s1, t0, t1;
  part_t partn[20];
  point_typ p0, p1, p2, q0, q1, qq0, qq1, p_in;
  vector4D_t opoly[MAX_VERTEX_COUNT], cpoly[MAX_VERTEX_COUNT],
             spoly0[MAX_VERTEX_COUNT], spoly1[MAX_VERTEX_COUNT];
  int BinA, opoints, cpoints, nsteps, null_intrn, itype, n_intrns;
  int bflag;

  /* Initialize the two polygons */

  /* Allocate space to hold all the possible partitions */

  for (i = 0; i <= tpA->TrimPortion[0].PiecePortion.NumOfVertices; i++)
    {
      polyA[i][0] = tpA->TrimPortion[0].PiecePortion.poly_vertex[i][0];
      polyA[i][1] = tpA->TrimPortion[0].PiecePortion.poly_vertex[i][1];
    }
  nA = i - 1;

  for (i = 0; i <= tpB->TrimPortion[0].PiecePortion.NumOfVertices; i++)
    {
      tmp0[0] = tpB->TrimPortion[0].PiecePortion.poly_vertex[i][0];
      tmp0[1] = tpB->TrimPortion[0].PiecePortion.poly_vertex[i][1];
      
      /* Now obtain the corresponding (s, t) co-od from these (u, v) */
      /* co-od */
      
      transform_uv_to_st(&tpA->tp, &tpB->tp, tmp0[0], tmp0[1], 
			 &polyB[i][0], &polyB[i][1]);
    }
  nB = i - 1;

  /* First remove all the redundant points (i.e. 3 points collinear) */
  /* in both the polygons. This is not necessary, since the */
  /* partitioning routine takes care of these cases, but to conserve */
  /* on space and to prevent the blowing up of #subchains, do this !! */

  i = 0;
  while (i < (nA - 2))
    {
      p0.x = polyA[i][0];
      p0.y = polyA[i][1];

      p1.x = polyA[i+1][0];
      p1.y = polyA[i+1][1];

      j = i + 2;
      while (j < (nA - 1))
	{
	  p2.x = polyA[j][0];
	  p2.y = polyA[j][1];
	
	  area = CROSS(p0, p1, p2);
	  if (!FP_EQUAL(area, 0.0)) /* Not collinear */
	    break;
	  else
	    j++;		/* Collinear.. go to next point */
	}
      
      /* Now remove all the intermediate points */
      
      delta = j - (i + 2);
      if (delta > 0)
	{
	  nA -= delta;
	  for (j = i+1; j < nA; j++)
	    {
	      polyA[j][0] = polyA[j + delta][0];
	      polyA[j][1] = polyA[j + delta][1];
	    }
	}

      i++;
   }

  /* Perform the same procedure for the trim-curve in patch B */

  i = 0;
  while (i < (nB - 2))
    {
      p0.x = polyB[i][0];
      p0.y = polyB[i][1];

      p1.x = polyB[i+1][0];
      p1.y = polyB[i+1][1];

      j = i + 2;
      while (j < (nB - 1))
	{
	  p2.x = polyB[j][0];
	  p2.y = polyB[j][1];
	
	  area = CROSS(p0, p1, p2);
	  if (!FP_EQUAL(area, 0.0)) /* Not collinear */
	    break;
	  else
	    j++;		/* Collinear.. go to next point */
	}
      
      /* Now remove all the intermediate points */
      
      delta = j - (i + 2);
      if (delta > 0)
	{
	  nB -= delta;
	  for (j = i+1; j < nB; j++)
	    {
	      polyB[j][0] = polyB[j + delta][0];
	      polyB[j][1] = polyB[j + delta][1];
	    }
	}

      i++;
   }


  /* After transform (u, v) to (s, t).. the orientation of the */
  /* trim-curve might have changed!! Hence get the curve into a proper */
  /* orientation */
  
  if(!is_curve_anticlockwise(polyB, nB)) 
    {				/* curve is clockwise */
				/* reverse the order of points */
      for (i = 0, j = nB; i < j; i++, j--)
	{
	  tmp0[0] = polyB[i][0];
	  tmp0[1] = polyB[i][1];
	  
	  polyB[i][0] = polyB[j][0];
	  polyB[i][1] = polyB[j][1];

	  polyB[j][0] = tmp0[0];
	  polyB[j][1] = tmp0[1];
	}
    }


  /* Now partition polygon A considering polygon B as the partitioning */
  /* chain. For that, first find a point on polgon B which lies */
  /* outside A; this in turn requires that we obtain a trapezoidation */
  /* of A first. */

  memset((void *)seg, 0, sizeof(seg));
  for (i = 1; i <= nA; i++)
    {
      seg[i].is_inserted = FALSE;

      seg[i].v0.x = polyA[i][0]; /* x-cood */
      seg[i].v0.y = polyA[i][1]; /* y-cood */
      if (i == 1)
	{
	  seg[nA].v1.x = seg[i].v0.x;
	  seg[nA].v1.y = seg[i].v0.y;
	}
      else
	{
	  seg[i - 1].v1.x = seg[i].v0.x;
	  seg[i - 1].v1.y = seg[i].v0.y;
	}	
    }

  global.nseg = nA;
  initialise_in_plane(nA);
  construct_trapezoids(nA, seg);

  /* Check each point on polyB and stop when we detect a point */
  /* outside A. If such a point is not found, then B lies completely */
  /* inside/outside A */

  p0.x = polyB[0][0];
  p0.y = polyB[0][1];
  was_in = is_point_within_polygon(&p0);
  
  null_intrn = TRUE;
  i = 1;
  while (i < nB) 
    {
      p0.x = polyB[i][0];
      p0.y = polyB[i][1];
      
      is_in = is_point_within_polygon(&p0);
      
      if (was_in != is_in)	/* Aha!! found a transition */
	{
	  pos0 = (is_in) ? (i-1) : i;
	  null_intrn = FALSE;
	  break;
	}
      
      i++;
      was_in = is_in;
    }

  bflag = FALSE;
  if (null_intrn) 
    {				
      /* No intersection was detected in the previous loop. This means */
      /* that A and B do not intersect (most of the time). But verify */
      /* this by intersecting every segment of B with every segemnt of */
      /* A .. this should take care of pathetic cases */
      
      for (i = 0; i < nB; i++)
	{
	  p0.x = polyB[i][0];
	  p0.y = polyB[i][1];
	  p1.x = polyB[i+1][0];
	  p1.y = polyB[i+1][1];

	  for (j = 0; j < nA; j++)
	    {
	      q0.x = polyA[j][0];
	      q0.y = polyA[j][1];
	      q1.x = polyA[j+1][0];
	      q1.y = polyA[j+1][1];
	      
	      if ((n_intrns = 
		   segments_intersect(&p0, &p1, &q0, &q1, &itype,
				      &s0, &s1, &t0, &t1, 
				      &qq0, &qq1)) > 0)
		{
		  /* Found an intersection. Now check was_in. If all */
		  /* the points of B were outside, return any of these */
		  /* points as pos0. Else, if all points of B were */
		  /* inside A, introduce the intersection point as an */
		  /* additional point and re-form the chain */
		  
		  null_intrn = FALSE;
		  if (!was_in)	/* all points outside */
		    pos0 = i;
		  else
		    {
		      /* add in the intersection_pt */
		      
		      /* Change Dec 28.  for (k = nB; k > i ; k++) */
		      for (k = nB; k > i ; k--)
			{
			  polyB[k+1][0] = polyB[k][0];
			  polyB[k+1][1] = polyB[k][1];
			}
		      polyB[i+1][0] = qq0.x;
		      polyB[i+1][1] = qq0.y;
		      nB++;
		      pos0 = i+1;
		    }
		  bflag = TRUE;
		  break;	/* Ged'outa here */
		}
	    }
	  if (bflag)
	    break;
	}
    } /* End- intersection detection */


  if (null_intrn)
    {				/* polyB lies completely inside or */
				/* outside of A. */
      /* Take any point on polyB and test whether it lies inside or */
      /* outside of A */
      p0.x = polyB[0][0];
      p0.y = polyB[0][1];
      BinA = is_point_within_polygon(&p0);
      
      if (BinA)
	{
	  switch (op)
	    {
	    case DOUNION:
	    case DIFFERENCE:
	      {
		/* B drills a hole into A. Hence get a patch in A */
		/* minus the hole. Divide the resulting polygon into */
		/* two and remove the hole */

		tpA->NumTrimPortions = 0;
		for (i = 0; i <= nA; i++)
		  {
		    opoly[i][0] = polyA[i][0];
		    opoly[i][1] = polyA[i][1];
		  }
		opoints = nA;

		for (i = 0; i <= nB; i++)
		  {
		    cpoly[i][0] = polyB[i][0];
		    cpoly[i][1] = polyB[i][1];
		  }
		cpoints = nB;
		
		find_point_inside_polygon(nB, polyB, &p_in);
		p2d->u = p_in.x;
		p2d->v = p_in.y;

		get_two_curves(opoly, cpoly, 
			       opoints, cpoints,
			       spoly0, spoly1,
			       &nA, &nB, TRUE, TRUE);

		/* The two curves are now present in spoly0 and */
		/* spoly1. Put them into the data-structure for patch */
		/* A */
		add_trim_portion_from_vector_table(tpA, nA, spoly0);
		add_trim_portion_from_vector_table(tpA, nB, spoly1);
		break;
	      }
	    case INTERSECTION:
	      tpA->NumTrimPortions = 0;
	      add_trim_portion_from_double_poly(tpA, nB, polyB);
	      
	      find_point_inside_polygon(nB, polyB, &p_in);
	      p2d->u = p_in.x;
	      p2d->v = p_in.y;
	      
	      break;
	    default:
	      printf("Hee Haa Haa..You deserve this, you chimpanzee!!\n");
	      break;	      
	    }
	}
      else			/* A inside B */
	{
	  switch (op)
	    {
	    case DOUNION:
	    case DIFFERENCE:
	      tpA->NumTrimPortions = 0;

	      find_point_inside_polygon(nA, polyA, &p_in);
	      p2d->u = p_in.x;
	      p2d->v = p_in.y;

	      break;
	    case INTERSECTION:

	      find_point_inside_polygon(cpoints, cpoly, &p_in);
	      p2d->u = p_in.x;
	      p2d->v = p_in.y;
	      
	      break;		/* leave tpA unchanged!! */
	    default:
	      printf("Hee Haa Haa..You deserve this, you chimpanzee!!\n");
	      break;	 
	    }
	}
    } /* end - completely within one patch case */
  else
    {
      /* the chain B intersects A at some point. Invoke */
      /* partition_polygon using this information and select the */
      /* appropriate partitions */

      memcpy((void *) polyC, (void *) polyB, sizeof(polyC));
      
      /* Now fill in elements back into polyB such that the element in */
      /* pos0 gets to the first position */
      
      for (i = pos0, j = 0; i < nB; i++, j++)
	{
	  polyB[j][0] = polyC[i][0];
	  polyB[j][1] = polyC[i][1];
	}
      
      for (i = 0; i < pos0; i++, j++)
	{
	  polyB[j][0] = polyC[i][0];
	  polyB[j][1] = polyC[i][1];
	}
      polyB[nB][0] = polyC[pos0][0];
      polyB[nB][1] = polyC[pos0][1];
      
      npartitions = intersect_polygonA_using_chainB(nA, polyA,
						    nB, polyB, 
						    partn);
      
      /* Out of these partitions, the ones which are S_RIGHT belong to */
      /* A alone, and the remaining belong to both A and B */
      
      tpA->NumTrimPortions = 0;
      for (i = 0; i < npartitions; i++)
	{
	  switch (partn[i].whichside) 
	    {
	    case S_RIGHT:
	      switch (op)
		{
		case DOUNION:
		case DIFFERENCE:
		  add_trim_portion(tpA, &partn[i]);
		  break;
		case INTERSECTION:
		  break;
		default:
		  printf("Hee Haa Haa..You deserve this, you chimpanzee!!\n");
		  break;
		}
	      break;
	    case S_LEFT:
	      switch (op)
		{
		case DOUNION:
		case DIFFERENCE:
		  break;
		case INTERSECTION:
		  add_trim_portion(tpA, &partn[i]);
		  break;
		default:
		  printf("Hee Haa Haa..You deserve this, you chimpanzee!!\n");
		  break;
		}

	      /* Return the point inside this common region as */
	      /* required. For this, we have to copy out all the */
	      /* elements of the common partition into a temp array */
	      /* and find its inner point */
	      
	      get_point_inside_intersection(&partn[i], &p_in);
	      p2d->u = p_in.x;
	      p2d->v = p_in.y;

	      break;
	    }
	  
	} /* end-for each partition */

      /* Free up all the space used up by the partitions */
      for (i = 0; i < npartitions; i++)
	free(partn[i].pc);

    } /* End-proper intersection computation */

  return 0;
}

