#include "defs.h"
#include <stdio.h>

extern divide_dixon();

void expandDet();
void zeroTheDeterminant();
void printDetMatrix();


void implicitize( tp, ndet )
    TENSOR_PATCH   tp;
    NDETERMINANT   ndet;
{
    
    int iTerms, iCols;

    /* this routine assumes the ndet is alloced but not 
       zeroed */


    if ((ndet.cRows!=((2*tp.iDeg0 + 1)*(tp.iDeg1 + 1))) ||
	(ndet.cCols!=((2*tp.iDeg1 + 1)*(tp.iDeg0 + 1)))
       ) {
	fprintf(stderr, "\nimplicitize: determinant matrix is too small\n");
    }
    /* we can reduce the determinant to a sum of dets times the variables
       x, y, z and a "constant" array (of w) */
    /* using sederberg's review of dixon's resultant, we get 
       a resultant of s,t,a,b -- stored in a matrix which is
       2mn X 2mn, where s goes from 0 to m, t from 0 to 2n,
       b from 0 to n, a from 0 to 2m.  So, the matrix can be
       thought of as having a row vector on the left of a^i * b^j
       on the left, and a col vector of s^k * t^l, where the 
       exponents range as described above -- I will increment
       s and b more quickly than t and a in indexing the vector,
       and, accordingly the array (and I increase the powers
       as I go down the vector).  A macro is given to give
       the index into the 2d array given the a,b,s,t exponents. */

    /* determinants being expanded look like:

       | W(s, t)  -Y(s, t)  -Z(s, t)|
       |                            |
     x | W(a, t)  -Y(a, t)  -Z(a, t)|
       |                            |
       | W(a, b)  -Y(a, b)  -Z(a, b)|

       Where each of these is a polynomial stored in the tensor
       product coeffs array (s^i*t^j coeffs are in the POINT4D
       referenced by tp.coeffs[i][j].dCoord, and this one is for
       X coord since there is an X in the front (there is no 
       term of the determinant expansion with w in the front, 
       so the constant part is left there.),
       and this is expanded out into a 2d matrix with the ab vector
       on the left, the st vector on the right (matrix is 2mn square).

       This whole formulation is taken from sederberg's thesis, and it
       is just a determinant expansion and consequent removal of 
       zero determinants plus factoring out x, y, and z from the
       determinants to leave a linear expression in the form
       x * detx + y * dety + z * detz + detw = Det(a,b,s,t).
       Now, to get the entries for the 2mnX2mn matrix A, we 
       just do factorization by (s-a)(t-b) to get rid of the
       obvious a and b roots, and then the a^i b^j s^k t^l
       coeff in the A matrix is the sum of the corresponding
       entries x*detx(ROW(i,j,degj), COL(s,t,degt)) + for y +
       for z + for detw().

       NOTE: n = tp.iDeg1, m = tp.iDeg0.
    */

    zeroTheDeterminant( ndet );
     /* printDetMatrix(ndet); */

    for(iTerms=X;iTerms<=W;iTerms++)  {
	for(iCols=0;iCols<3;iCols++) {
	    expandDet( iTerms, iCols, tp, ndet );
	}
    }   

    divide_dixon(&ndet, tp.iDeg0, tp.iDeg1);
}



void expandDet( iInto, iAboutCol, tp, ndet )
    int           iInto, iAboutCol;
    TENSOR_PATCH  tp;
    NDETERMINANT  ndet;
{
    int    iRow0Rows, iRow0Cols, iRow1Rows, iRow1Cols, iRow2Rows, iRow2Cols;
    double dRow0Term, dRow1Term1, dRow1Term2, dRow2Term1, dRow2Term2;
    int    iCoord, iRowNDet, iColNDet, cAExp, cBExp, cSExp, cTExp;
    

    /* expands a 3 X 3 polynomial determinant about some column and
       writes the result into "plane" iInto */

    for(iRow0Rows=0;iRow0Rows<=tp.iDeg0;iRow0Rows++)
	for(iRow0Cols=0;iRow0Cols<=tp.iDeg1;iRow0Cols++) {

	    for(iRow1Rows=0;iRow1Rows<=tp.iDeg0;iRow1Rows++)
		for(iRow1Cols=0;iRow1Cols<=tp.iDeg1;iRow1Cols++) 
		    for(iRow2Rows=0;iRow2Rows<=tp.iDeg0;iRow2Rows++)
			for(iRow2Cols=0;iRow2Cols<=tp.iDeg1;iRow2Cols++) {

			    /* first term, ie, W(s,t)*(-Y(a,t))*(-Z(a,b)) */

			    dRow0Term = (iAboutCol==iInto) ?
				tp.coeffs[iRow0Rows][iRow0Cols].dCoord[W] :
				-tp.coeffs[iRow0Rows][iRow0Cols].dCoord[iAboutCol];

			    iCoord = (iAboutCol + 1) % 3;

			    dRow1Term1 = (iCoord==iInto) ?
				tp.coeffs[iRow1Rows][iRow1Cols].dCoord[W] :
			        -tp.coeffs[iRow1Rows][iRow1Cols].dCoord[iCoord];

			    iCoord = ((iAboutCol - 1) < 0) ? 2 : 
				iAboutCol - 1;

			    dRow2Term1 = (iCoord==iInto) ?
				tp.coeffs[iRow2Rows][iRow2Cols].dCoord[W] :
				-tp.coeffs[iRow2Rows][iRow2Cols].dCoord[iCoord];


			    /* second term, ie, W(s,t)*(-Z(a,t))*(-Y(a,b)) */

			    iCoord = ((iAboutCol - 1) < 0) ? 2 : 
				iAboutCol - 1;

			    dRow1Term2 = (iCoord==iInto) ?
				tp.coeffs[iRow1Rows][iRow1Cols].dCoord[W] :
			        -tp.coeffs[iRow1Rows][iRow1Cols].dCoord[iCoord];

			    iCoord = (iAboutCol + 1) % 3;

			    dRow2Term2 = (iCoord==iInto) ?
				tp.coeffs[iRow2Rows][iRow2Cols].dCoord[W] :
				-tp.coeffs[iRow2Rows][iRow2Cols].dCoord[iCoord];

			    cAExp = iRow1Rows + iRow2Rows;
			    cBExp = iRow2Cols;
			    cSExp = iRow0Rows;
			    cTExp = iRow0Cols + iRow1Cols;
			    iRowNDet = ROW_INDEX(cAExp,cBExp,tp.iDeg1);
			    iColNDet = COL_INDEX(cSExp,cTExp,tp.iDeg0);

			    ndet.coeffs[iRowNDet][iColNDet].dCoord[iInto] +=
				dRow0Term * 
				(dRow1Term1 * dRow2Term1 - 
				 dRow1Term2 * dRow2Term2);

			}
	}
}


void
allocDeterminant(pndet, m, n, fType)
NDETERMINANT  *pndet;
int            m, n;
int            fType;
{
    int iRows;

    /* final determinant is only 2mn X 2mn, but for the intermediate
       stage (before factoring out (s-a)(t-b)),
       we need (2m+1)(n+1) X (2n+1)(m+1)
    */
    if (fType==UNFACTORED) {
	pndet->cRows = (2*m + 1) * (n + 1);
	pndet->cCols = (2*n + 1) * (m + 1);
    }
    else if (fType==FINAL) {
	pndet->cRows = 2*m*n;
	pndet->cCols = 2*n*m;
    }
    else {
	fprintf(stderr,"\nallocDeterminant: unknown type of determinant.\n");
	exit(-2);
    }
    ALLOCN(pndet->coeffs, POINT4D*, pndet->cRows);
    for (iRows=0;iRows<pndet->cRows;iRows++)
	ALLOCN(pndet->coeffs[iRows], POINT4D, pndet->cCols);
}

void free_alloc_determ(pndet)
NDETERMINANT *pndet;
{
  int i;

  if (pndet == NULL) return;

  for (i=0; i<pndet->cRows; i++) {
    myfree(pndet->coeffs[i]);
    pndet->coeffs[i] = NULL;
  }

  myfree(pndet->coeffs);
  pndet->coeffs = NULL;
}

void zeroTheDeterminant( ndet )
    NDETERMINANT  ndet;
{
    int i,j,k;

    /* zero out the ndet */
    for(i=0;i<ndet.cRows;i++) 
	for(j=0;j<ndet.cCols;j++)
	    for(k=X;k<=W;k++)
		ndet.coeffs[i][j].dCoord[k]=0.0;
}

void
alloc_tensor_patch_1D(ptp1dpatch)
TENSOR_PATCH_1D 	*ptp1dpatch;
{
    int			iRows;
    int			m, n;

    m = ptp1dpatch->iDeg0;
    n = ptp1dpatch->iDeg1;

    ALLOCN(ptp1dpatch->coeffs, double*, m+1);
    for (iRows=0;iRows<=m;iRows++)
      ALLOCN(ptp1dpatch->coeffs[iRows], double, n+1);
}

void free_ten_p_1D(ptp1dpatch)
TENSOR_PATCH_1D         *ptp1dpatch;
{
  int i;

  if (ptp1dpatch == NULL) return;

  for (i=0; i<=ptp1dpatch->iDeg0; i++) {
    myfree(ptp1dpatch->coeffs[i]);
    ptp1dpatch->coeffs[i] = NULL;
  }

  myfree(ptp1dpatch->coeffs);
  ptp1dpatch->coeffs = NULL;
}

/* The matrix is of the order of m x n, and the bivariate polynomial
   stored at each element of the matrix has highest degrees p and q.
*/
void
allocDeterminant_1D(pndet, m, n, p, q)
NDETERMINANT_1D 	*pndet;
int            		m, n, p, q;
{
    int iRows, iCols;

    pndet->cRows = m;
    pndet->cCols = n;

    ALLOCN(pndet->poly, TENSOR_PATCH_1D*, pndet->cRows);
    for (iRows=0;iRows<pndet->cRows;iRows++)

    { 
	ALLOCN(pndet->poly[iRows], TENSOR_PATCH_1D, pndet->cCols);

	for(iCols=0;iCols<pndet->cCols;iCols++) {
	    pndet->poly[iRows][iCols].iDeg0 = p;
	    pndet->poly[iRows][iCols].iDeg1 = q;
	    alloc_tensor_patch_1D(&(pndet->poly[iRows][iCols]));
	}
    }
    
    ALLOCN(pndet->eval, double*, pndet->cRows);
    for (iRows=0;iRows<pndet->cRows;iRows++)
      ALLOCN(pndet->eval[iRows], double, pndet->cCols);
}

void free_alloc_det_1D(pndet)
NDETERMINANT_1D         *pndet;
{
  int i, j;

  if (pndet == NULL) return;

  for (i=0; i<pndet->cRows; i++) {
    myfree(pndet->eval[i]);
    pndet->eval[i] = NULL;
  }
  myfree(pndet->eval);
  pndet->eval = NULL;

  for (i=0; i<pndet->cRows; i++)
    for (j=0; j<pndet->cCols; j++)
      free_ten_p_1D(&(pndet->poly[i][j]));
      

  for (i=0; i<pndet->cRows; i++) {
    myfree(pndet->poly[i]);
    pndet->poly[i] = NULL;
  }
  myfree(pndet->poly);
  pndet->poly = NULL;
}
  
void  printDetMatrix( ndet )
    NDETERMINANT ndet;
{
    int i, j, k;

    fprintf(stdout, "M = {\n");

    for (i=0;i<ndet.cRows;i++) {
	fprintf(stdout, "{");
	for (j=0;j<ndet.cCols;j++) {
	    fprintf(stdout, "%2.1lf*x + %2.1lf*y + %2.1lf*z + %2.1lf",
		    ndet.coeffs[i][j].dCoord[X],
		    ndet.coeffs[i][j].dCoord[Y],
		    ndet.coeffs[i][j].dCoord[Z],
		    ndet.coeffs[i][j].dCoord[W]);
	    if (j!=(ndet.cCols-1))
		fprintf(stdout, ", \n");
	}
	fprintf(stdout, "}");
	if (i!=(ndet.cRows-1))
	    fprintf(stdout, ", \n");
    }
    fprintf(stdout, "}\n");

    for(k=X;k<=W;k++) {
	fprintf(stdout, "M%d = {\n", k);

	for (i=0;i<ndet.cRows;i++) {
	    fprintf(stdout, "{");
	    for (j=0;j<ndet.cCols;j++) {
		fprintf(stdout, "%2.1lf",
			ndet.coeffs[i][j].dCoord[k]);
		if (j!=(ndet.cCols-1))
		    fprintf(stdout, ", ");
	    }
	    fprintf(stdout, "}");
	    if (i!=(ndet.cRows-1))
		fprintf(stdout, ", \n");
	}
	fprintf(stdout, "}\n");
    }
}


void moveFactoredToFinalDeterminant( ndetTo, ndetFrom, m, n )
    NDETERMINANT   ndetTo, ndetFrom;
    int            m, n;
{
    int            a, b, s, t;

    for(a=0;a<2*m;a++)
	for(b=0;b<n;b++)
	    for(s=0;s<m;s++)
		for(t=0;t<2*n;t++)
		    ndetTo.coeffs[ROW_INDEX( a, b, n-1 )]
			[COL_INDEX( s, t, m-1 )]=
		    ndetFrom.coeffs[ROW_INDEX( a, b, n )]
			[COL_INDEX( s, t, m )];
}

/* assumes both determinants are properly allocated and copies 
   detB to detA
*/
void copyDeterminant_1D(ndet1D_A, ndet1D_B)
NDETERMINANT_1D		ndet1D_A, ndet1D_B;
{ 
  int			i, j, k, l;

   ndet1D_A.cRows = ndet1D_B.cRows;
   ndet1D_A.cCols = ndet1D_B.cCols;
   for(i = 0; i < ndet1D_B.cRows; i++)
   { for(j = 0; j < ndet1D_B.cCols; j++)
     { ndet1D_A.poly[i][j].iDeg0 = ndet1D_B.poly[i][j].iDeg0;
       ndet1D_A.poly[i][j].iDeg1 = ndet1D_B.poly[i][j].iDeg1;
       for(k = 0; k <= ndet1D_B.poly[i][j].iDeg0; k++)
         for(l = 0; l <= ndet1D_B.poly[i][j].iDeg1; l++)
           ndet1D_A.poly[i][j].coeffs[k][l] = ndet1D_B.poly[i][j].coeffs[k][l];
     }
   }
}
