#include <stdio.h>
#include "defs.h"
#include "kinter.h"

EXTERN_LINALG

/************************************************************************
  generate_Bernstein_polys
	Generates the nth degree Bernstein polynomials (there will be
  n+1 such polynomials).  Uses recursive definition of Bernstein 
  polynomials.  Must handle first and last polynomial separately.  Also
  must handle first and last coefficient (term) separately.  
************************************************************************/


void
generate_Bernstein_polys(B, n)
poly* B;
int n;
{
   int i,j;
   poly* Bold;

   if (n>0) {
      ALLOCN(Bold, poly, n);
      generate_Bernstein_polys(Bold, n-1);
      for (i=0;i<=n;i++) {
         if (i==0) {
            /* Bern[0] = (1-t)*BernOld[0] */
            B[i].deg = Bold[i].deg+1;
            ALLOCN(B[i].coef, double, B[i].deg+1);
            B[i].coef[0] = Bold[i].coef[0];
            for(j=1;j<(B[i].deg);j++) {
               B[i].coef[j] = Bold[i].coef[j] - Bold[i].coef[j-1];
            }
            B[i].coef[B[i].deg] = -Bold[i].coef[Bold[i].deg];
         } else if (i==n) {
            /* Bern[n] = t*BernOld[n-1] */
            B[i].deg = Bold[i-1].deg+1;
            ALLOCN(B[i].coef, double, B[i].deg+1);
            B[i].coef[0] = 0.0;
            for(j=1;j<=B[i].deg;j++) {
               B[i].coef[j] = Bold[i-1].coef[j-1];
            }
         } else {
            /* Bern[i] = (1-t)*Bern[i] + t*BernOld[i-1] */
            B[i].deg = Bold[i].deg+1;
            ALLOCN(B[i].coef, double, B[i].deg+1);
            B[i].coef[0] = Bold[i].coef[0];
            for (j=1;j<(B[i].deg);j++) {
               B[i].coef[j] = Bold[i].coef[j] - Bold[i].coef[j-1] +
                              Bold[i-1].coef[j-1];
            }
            B[i].coef[B[i].deg] = Bold[i-1].coef[Bold[i].deg] - 
                                    Bold[i].coef[Bold[i].deg];
         }
      }
      for (i=0;i<n;i++) {
         free(Bold[i].coef);
      }
      free(Bold);
   } else {
      B->deg = 0;
      ALLOCN(B->coef, double, 1);
      B->coef[0] = 1.0;
   }
}

/************************************************************************
  gen_polys_from_curve
	Given a curve (in u,v,w coordinates), generates the implicit
  polynomials corresponding to u, v, and w.
************************************************************************/

void
gen_polys_from_curve(deg, ctl, flag, u, v, w)
int deg, flag;
POINT3D* ctl;
poly* u;
poly* v;
poly* w;
{
   poly* Bern;
   int i,j;
   long tmp;

#ifdef DOBERNSTEIN
   ALLOCN(Bern, poly, deg+1);
   generate_Bernstein_polys(Bern, deg);
#endif

   u->deg = deg;
   ALLOCN(u->coef,double,deg+1);
   v->deg = deg;
   ALLOCN(v->coef,double,deg+1);
   w->deg = deg;
   ALLOCN(w->coef,double,deg+1);

#ifdef DOBERNSTEIN
   for(i=0;i<=deg;i++) {
      /* loop through terms in the polynomial */
      u->coef[i] = 0.0;
      v->coef[i] = 0.0;
      w->coef[i] = 0.0;
      for(j=0;j<=deg;j++) {
         /* loop through the Bernstein polynomials */
         u->coef[i] += ctl[j].dCoord[0]*Bern[j].coef[i];
         v->coef[i] += ctl[j].dCoord[1]*Bern[j].coef[i];
         w->coef[i] += ctl[j].dCoord[2]*Bern[j].coef[i];
      }
   }
   for (i=0;i<=deg;i++) {
      free(Bern[i].coef);
   }
   myfree(Bern);
   Bern = NULL;
#else
   tmp = 1;
   u->coef[0] = ctl[0].dCoord[0];
   v->coef[0] = ctl[0].dCoord[1];
   w->coef[0] = ctl[0].dCoord[2];
   for(i=1;i<=deg;i++) {
      u->coef[i] = ctl[i].dCoord[0];
      v->coef[i] = ctl[i].dCoord[1];
      w->coef[i] = ctl[i].dCoord[2];
      if (flag) { 
	tmp *= (deg-i+1);
	tmp = (long) (tmp/i);
	u->coef[i] *= tmp;
	v->coef[i] *= tmp;
	w->coef[i] *= tmp;
      }
   }
#endif
}

void InverseOper(c1, pt, s)
UV_SPLINE_TYPE *c1;
POINT2D *pt;
double *s;
{
  int i;
  double u1, v1, u2, v2, tmp;

  u1 = c1->CntrlPts[0].dCoord[0] / c1->CntrlPts[0].dCoord[2];
  v1 = c1->CntrlPts[0].dCoord[1] / c1->CntrlPts[0].dCoord[2];
  u2 = c1->CntrlPts[1].dCoord[0] / c1->CntrlPts[1].dCoord[2];
  v2 = c1->CntrlPts[1].dCoord[1] / c1->CntrlPts[1].dCoord[2];

  *s = -1.0;
  if (!FP_EQ_EPS(u1, u2, 1e-3)) {
    tmp = (pt->u - u1)/(u2 - u1);
    if (!FP_EQ_EPS(v1, v2, 1e-3) && (!FP_EQ_EPS(tmp,((pt->v-v1)/(v2-v1)),1e-3))) {
      printf("Error in InverseOper\n");
      exit(-1);
    }
    if ((tmp >= -0.001) && (tmp <= 1.001))
      *s = tmp;
  } else {
    if (FP_EQ_EPS(v1, v2, 1e-3)) {
      printf("(u1,v1) and (u2,v1) are the same\n");
      exit(-1);
    }
    tmp = (pt->v - v1)/(v2 - v1);
    if ((tmp >= -0.001) && (tmp <= 1.001))
      *s = tmp;
  }
}

void FindXsections(ndet1D_u, c1, c2, s, t, numinter, base)
NDETERMINANT_1D		*ndet1D_u;
UV_SPLINE_TYPE          *c1, *c2;
double                  *s, *t;
int                     *numinter, base;
{
  double		**rgrgA_plus_B, **rgrgB;
  double		*dEReal, *dEImag, *dEBeta, **rgrgEVectors;
  int			i, j, k, l, iSizeResultant;
  int			iDeg_Curve;
  int			iSize_C;
  POINT2D               p2d, p2d1;
  POINT2D               already_computed[20]; /* max. number of
						 intersections */

  iDeg_Curve = ndet1D_u->poly[0][0].iDeg1;
  iSizeResultant = ndet1D_u->cRows;

  iSize_C = iSizeResultant*iDeg_Curve;

  ALLOCMAT(rgrgB, iSize_C, iSize_C, double);
  ALLOCN(dEReal, double, iSize_C);
  ALLOCN(dEImag, double, iSize_C);
  ALLOCMAT(rgrgEVectors, iSize_C, iSize_C, double);
  ALLOCMAT(rgrgA_plus_B, iSize_C, iSize_C, double);
  
  if (form_power_companion_matrix((*ndet1D_u),rgrgB,rgrgA_plus_B,iSize_C)
      == -1) {
    s = t = NULL;
    *numinter = 0;
    return;
  }
  
  /* find the eigenvalues of C */
  *numinter = 0;
  ALLOCN(dEBeta, double, iSize_C);
  call_rgg(rgrgB, rgrgA_plus_B, iSize_C, iSize_C, dEReal, 
	   dEImag, dEBeta, rgrgEVectors, 0);
  for (i=0; i<iSize_C; i++) 
    if (!FP_EQ(dEBeta[i], 0.0)) {
      dEReal[i] /= dEBeta[i];
      dEImag[i] /= dEBeta[i];
    }
    else dEReal[i] = dEImag[i] = ERROR_VAL;
  for (i=j=0; i<iSize_C; i++) {
/*
    if (!FP_EQ(dEImag[i],0.0)) {
      printf("Symmetric matrix has complex eigenvalue\n");
      exit(-1);
    }
*/
    if (FP_EQ(dEImag[i],0.0) && (dEReal[i] >= -1e-3) && (dEReal[i] <= 1.001)) {
      t[(*numinter)] = dEReal[i];
      if (0) {
	s[(*numinter)] = rgrgEVectors[1][i]/rgrgEVectors[0][i];
	s[(*numinter)] = inverse_power_pt(s[(*numinter)]);
	Eval2DCurve(c2, t[(*numinter)], &p2d);
	Eval2DCurve(c1, s[(*numinter)], &p2d1);
	if ((s[(*numinter)] >= -0.001) && (s[(*numinter)] <= 1.001) &&
	    EQUAL2D_EPS(p2d, p2d1, 2e-6))
	  (*numinter)++;
      } else {
	/* when c1 is linear */
/*
	if (c1->Deg != 1) {
	  printf("Implicit reprn is order one, though Deg != 1\n");
	  exit(-1);
	}
*/
	Eval2DCurve(c2, t[(*numinter)], &p2d);
	for (k=0; k<(*numinter); k++)
	  if (EQUAL2D_EPS(already_computed[k], p2d, 1e-6)) break;
	if (k==(*numinter)) {
	  EvaluateParam(c1, &p2d, &s[(*numinter)]);
	  if ((s[(*numinter)] >= -0.001) && (s[(*numinter)] <= 1.001))
	    already_computed[(*numinter)++] = p2d;
	}
      }
    }
  }
  FREEMAT(rgrgB, iSize_C, iSize_C);
  FREEMAT(rgrgEVectors, iSize_C, iSize_C);
  FREEMAT(rgrgA_plus_B, iSize_C, iSize_C);
  myfree(dEReal);
  myfree(dEImag);
  myfree(dEBeta);
  dEReal = dEImag = dEBeta = NULL;
}

/************************************************************************
  intersect_plane_curves
        Given two planar curves (in u,v,w coordinates), determines
  the number of intersections between them and figures out the points
  of intersection.
************************************************************************/

void
intersect_plane_curves(c1, c2, s, t, numinter)
UV_SPLINE_TYPE* c1;
UV_SPLINE_TYPE* c2;
double* s;
double* t;
int* numinter;
{
#ifndef USERESULTANTS

   NDETERMINANT_1D thisdet;
   int degdet;
   int i,j,k,k2;
   int amt;
   poly u1;
   poly v1;
   poly w1;
   poly u2;
   poly v2;
   poly w2;
   TENSOR_PATCH_1D* e1;
   TENSOR_PATCH_1D* e2;
   TENSOR_PATCH_1D* m1;
   TENSOR_PATCH_1D* m2;
   NDETERMINANT ndet;
   int *row_perm, *col_perm;
   int base_case, rank_row_val;

   /* first get the polynomials for the parameters u, v, w */
   gen_polys_from_curve(c1->Deg, c1->CntrlPts, 1, &u1, &v1, &w1);
   gen_polys_from_curve(c2->Deg, c2->CntrlPts, 0, &u2, &v2, &w2);

#ifdef OLDWAY
   /* Determine the equations (two polynomials in s with coefficients
           which are polynomials in t */
   ALLOCN(e1, TENSOR_PATCH_1D, u1.deg+1);
   ALLOCN(e2, TENSOR_PATCH_1D, u1.deg+1);
   for (i=0;i<=u1.deg;i++) {
     e1[i].iDeg0=u2.deg;
     e1[i].iDeg1=0;
     ALLOCN(e1[i].coeffs, double*, u2.deg+1);
     for(j=0;j<=u2.deg;j++) {
       ALLOCN(e1[i].coeffs[j],double,1);
       e1[i].coeffs[j][0] = u1.coef[i]*w2.coef[j]-w1.coef[i]*u2.coef[j];
     }
     e2[i].iDeg0=u2.deg;
     e2[i].iDeg1=0;
     ALLOCN(e2[i].coeffs, double*, (u2.deg+1));
     for(j=0;j<=u2.deg;j++) {
       ALLOCN(e2[i].coeffs[j],double,1);
       e2[i].coeffs[j][0] = v1.coef[i]*w2.coef[j]-w1.coef[i]*v2.coef[j];
     }
   }

#define DONTDOSYLVESTER
#ifndef DONTDOSYLVESTER
   /* Now set up Sylvester-style (is this the right one?) resultant */
   degdet = u1.deg + u1.deg;
   thisdet.cRows = degdet;
   thisdet.cCols = degdet;
   ALLOCN(thisdet.poly, TENSOR_PATCH_1D*, degdet);
   /* build rows for first equation */
   for (i=0;i<u1.deg;i++) {
      ALLOCN(thisdet.poly[i], TENSOR_PATCH_1D, degdet);
      for (j=0;j<i;j++) {
         thisdet.poly[i][j].iDeg0 = 0;
         thisdet.poly[i][j].iDeg1 = 0;
         ALLOCN(thisdet.poly[i][j].coeffs, double*, 1);
         ALLOCN(thisdet.poly[i][j].coeffs[0], double, 1);
         thisdet.poly[i][j].coeffs[0][0] = 0.0;
      }
      for (j=0;j<=u1.deg;j++) {
         thisdet.poly[i][i+j].iDeg0 = e1[i].iDeg0;
         thisdet.poly[i][i+j].iDeg1 = e1[i].iDeg1;
         ALLOCN(thisdet.poly[i][i+j].coeffs, double*, e1[i].iDeg0+1);
         for(k=0;k<=e1[i].iDeg0;k++) {
            ALLOCN(thisdet.poly[i][i+j].coeffs[k], double, 1);
            thisdet.poly[i][i+j].coeffs[k][0] = e1[j].coeffs[k][0]
         }
      }
      for (j=(i+u1.deg+2);j<degdet;j++) {
         thisdet.poly[i][j].iDeg0 = 0; 
         thisdet.poly[i][j].iDeg1 = 0;
         ALLOCN(thisdet.poly[i][j].coeffs, double*, 1);
         ALLOCN(thisdet.poly[i][j].coeffs[0], double, 1);
         thisdet.poly[i][j].coeffs[0][0] = 0.0;
      }
   }
   for (i=0;i<u1.deg;i++) {
      ALLOCN(thisdet.poly[i+u1.deg], TENSOR_PATCH_1D, degdet);
      for (j=0;j<i;j++) {
         thisdet.poly[i+u1.deg][j].iDeg0 = 0; 
         thisdet.poly[i+u1.deg][j].iDeg1 = 0;
         ALLOCN(thisdet.poly[i+u1.deg][j].coeffs, double*, 1);
         ALLOCN(thisdet.poly[i+u1.deg][j].coeffs[0], double, 1);
         thisdet.poly[i+u1.deg][j].coeffs[0][0] = 0.0;
      }
      for (j=0;j<=u1.deg;j++) {
         thisdet.poly[i+u1.deg][i+j].iDeg0 = e1[i].iDeg0;
         thisdet.poly[i+u1.deg][i+j].iDeg1 = e1[i].iDeg1;
         ALLOCN(thisdet.poly[i+u1.deg][i+j].coeffs, double*, e1[i].iDeg0+1);
         for(k=0;k<=e1[i].iDeg0;k++) {
            ALLOCN(thisdet.poly[i+u1.deg][i+j].coeffs[k], double, 1);
            thisdet.poly[i+u1.deg][i+j].coeffs[k][0] = e2[j].coeffs[k][0]
         }
      }
      for (j=(i+u1.deg+2);j<degdet;j++) {
         thisdet.poly[i+u1.deg][j].iDeg0 = 0; 
         thisdet.poly[i+u1.deg][j].iDeg1 = 0;
         ALLOCN(thisdet.poly[i+u1.deg][j].coeffs, double*, 1);
         ALLOCN(thisdet.poly[i+u1.deg][j].coeffs[0], double, 1);
         thisdet.poly[i+u1.deg][j].coeffs[0][0] = 0.0;
      }
   }
#endif

   /* Now set up Cayley/Bezout resultant */
   /* multiply e1(s)*e2(alpha)-e2(alpha)*e1(s) */
   ALLOCN(m1,TENSOR_PATCH_1D,u1.deg+1);
   for (i=0;i<=u1.deg;i++) {
      ALLOCN(m1[i].coeffs, double*, u1.deg+1);
      for (j=0;j<=u1.deg;j++) {
         ALLOCN(m1[i].coeffs[j], double, (2*u2.deg)+1);
         for (k=0;k<=(2*u2.deg);k++) {
            m1[i].coeffs[j][k] = 0.0;
         }
      }
   }
   for (i=0;i<=u1.deg;i++) {
      for (j=0;j<=u1.deg;j++) {
         /* m1[i].coeffs[j] += e1[i].coeffs * e2[j].coeffs */
         /* m1[j].coeffs[i] -= e1[i].coeffs * e2[j].coeffs */
         for (k=0;k<=u2.deg;k++) {
            for (k2=0;k2<=u2.deg;k2++) {
               m1[i].coeffs[j][k+k2] += e1[i].coeffs[k][0]*e2[j].coeffs[k2][0];
               m1[j].coeffs[i][k+k2] -= e1[i].coeffs[k][0]*e2[j].coeffs[k2][0];
            }
         }
      }
   }
   /* now divide through by (s-alpha) */
   ALLOCN(m2,TENSOR_PATCH_1D,u1.deg+1);
   for (i=0;i<=u1.deg;i++) {
      ALLOCN(m2[i].coeffs, double*, u1.deg+1);
      for (j=0;j<=u1.deg;j++) {
         ALLOCN(m2[i].coeffs[j], double, (2*u2.deg)+1);
         for (k=0;k<=(2*u2.deg);k++) {
            m2[i].coeffs[j][k] = 0.0;
         }
      }
   }
   for (i=1;i<=u1.deg;i++) {
      for (j=0;j<i;j++) {
         amt = i-j-1;
         for (k=0;k<=amt;k++) {
            for (k2=0;k2<=(2*u2.deg);k2++) {
               m2[j+k].coeffs[j+amt-k][k2] += m1[i].coeffs[j][k2];
            }
         }
      }
   }

   /* Now fill in resultant matrix from m2 */
   thisdet.cRows = u1.deg;
   thisdet.cCols = u1.deg;
   ALLOCN(thisdet.poly, TENSOR_PATCH_1D*, u1.deg+1);
   for (i=0;i<=u1.deg;i++) {
      ALLOCN(thisdet.poly[i], TENSOR_PATCH_1D, u1.deg+1);
      for (j=0;j<=u1.deg;j++) {
         ALLOCN(thisdet.poly[i][j].coeffs, double*, (2*u2.deg)+1);
         thisdet.poly[i][j].iDeg0 = 2*u2.deg;
         thisdet.poly[i][j].iDeg1 = 0;
         for(k=0;k<=(2*u2.deg);k++) {
            ALLOCN(thisdet.poly[i][j].coeffs[k],double,1);
            thisdet.poly[i][j].coeffs[k][0] = m2[j].coeffs[i][k];
         }
      }
   }
#endif

#ifndef OLDWAY
   /* Now set up Cayley/Bezout resultant */
   /* multiply e1(s)*e2(alpha)-e2(alpha)*e1(s) */
   ALLOCN(m1,TENSOR_PATCH_1D,u1.deg+1);
   for (i=0;i<=u1.deg;i++) {
      ALLOCN(m1[i].coeffs, double*, u1.deg+1);
      for (j=0;j<=u1.deg;j++) {
         ALLOCN(m1[i].coeffs[j], double, 3);
         for (k=0;k<3;k++) {
            m1[i].coeffs[j][k] = 0.0;
         }
      }
   }
   for (i=0;i<=u1.deg;i++) {
      for (j=0;j<=u1.deg;j++) {
         m1[i].coeffs[j][0] += v1.coef[i]*w1.coef[j] - v1.coef[j]*w1.coef[i]; 
         m1[i].coeffs[j][1] += u1.coef[j]*w1.coef[i] - u1.coef[i]*w1.coef[j];
         m1[i].coeffs[j][2] += u1.coef[i]*v1.coef[j] - u1.coef[j]*v1.coef[i];
      }
   }
   /* now divide through by (s-alpha) */
   ALLOCN(m2,TENSOR_PATCH_1D,u1.deg);
   for (i=0;i<u1.deg;i++) {
      ALLOCN(m2[i].coeffs, double*, u1.deg);
      for (j=0;j<u1.deg;j++) {
         ALLOCN(m2[i].coeffs[j], double, 3);
         for (k=0;k<3;k++) {
            m2[i].coeffs[j][k] = 0.0;
         }
      }
   }
   for (i=1;i<=u1.deg;i++) {
      for (j=0;j<i;j++) {
         amt = i-j-1;
         for (k=0;k<=amt;k++) {
            for (k2=0;k2<3;k2++) {
               m2[j+k].coeffs[j+amt-k][k2] += m1[i].coeffs[j][k2];
            }
         }
      }
   }

   ndet.cRows = ndet.cCols = u1.deg;
   ALLOCN(ndet.coeffs, POINT4D *, ndet.cRows);
   for (i=0; i<ndet.cRows; i++)
     ALLOCN(ndet.coeffs[i], POINT4D, ndet.cCols);
   
   for (i=0; i<ndet.cRows; i++)
     for (j=0; j<ndet.cCols; j++) {
       ndet.coeffs[i][j].dCoord[X] = m2[i].coeffs[j][0];
       ndet.coeffs[i][j].dCoord[Y] = m2[i].coeffs[j][1];
       ndet.coeffs[i][j].dCoord[Z] = 0.0;
       ndet.coeffs[i][j].dCoord[W] = m2[i].coeffs[j][2];
     }
   
   base_case = check_for_base_points(ndet, ndet.cRows, ndet.cCols, &rank_row_val,
				     &row_perm, &col_perm);

   /* Now fill in resultant matrix from m2. thisdet is a symmetric matrix */
   if (base_case) {
     allocDeterminant_1D(&thisdet, rank_row_val, rank_row_val, 0, u2.deg);
     for (i=0; i<thisdet.cRows; i++)
       for (j=0; j<thisdet.cCols; j++)
	 for (k=0; k<=thisdet.poly[i][j].iDeg1; k++)
	   thisdet.poly[i][j].coeffs[0][k] = u2.coef[k]*m2[row_perm[i]].
	                                                coeffs[col_perm[j]][0] +
	                                     v2.coef[k]*m2[row_perm[i]].
                                                        coeffs[col_perm[j]][1] +
	                                     w2.coef[k]*m2[row_perm[i]].
                                                        coeffs[col_perm[j]][2];
   } else {   
     allocDeterminant_1D(&thisdet, u1.deg, u1.deg, 0, u2.deg);
     for (i=0; i<thisdet.cRows; i++)
       for (j=0; j<thisdet.cCols; j++)
	 for (k=0; k<=thisdet.poly[i][j].iDeg1; k++)
	   thisdet.poly[i][j].coeffs[0][k] = u2.coef[k]*m2[i].coeffs[j][0] +
	                                     v2.coef[k]*m2[i].coeffs[j][1] +
	                                     w2.coef[k]*m2[i].coeffs[j][2];
   }

   for (i=0; i<thisdet.cRows; i++)
     for (j=0; j<thisdet.cCols; j++)
       bernsteinToPowerBasis_1D(thisdet.poly[i][j]);

   FindXsections(&thisdet, c1, c2, s, t, numinter, base_case);

   myfree(u1.coef); u1.coef = NULL;
   myfree(v1.coef); v1.coef = NULL;
   myfree(w1.coef); w1.coef = NULL;
   myfree(u2.coef); u2.coef = NULL;
   myfree(v2.coef); v2.coef = NULL;
   myfree(w2.coef); w2.coef = NULL;

   for (i=0; i<u1.deg; i++) {
     for (j=0; j<u1.deg; j++) {
       myfree(m2[i].coeffs[j]);
       m2[i].coeffs[j] = NULL;
     }
     myfree(m2[i].coeffs);
     m2[i].coeffs = NULL;
   }
   myfree(m2); m2 = NULL;

   for (i=0; i<=u1.deg; i++) {
     for (j=0; j<=u1.deg; j++) {
       myfree(m1[i].coeffs[j]);
       m1[i].coeffs[j] = NULL;
     }
     myfree(m1[i].coeffs);
     m1[i].coeffs = NULL;
   }
   myfree(m1); m1 = NULL;
   for (i=0; i<ndet.cRows; i++) {
     myfree(ndet.coeffs[i]);
     ndet.coeffs[i] = NULL;
   }
   myfree(ndet.coeffs);
   ndet.coeffs = NULL;

   free_alloc_det_1D(&thisdet);
#endif

#else
   TENSOR_PATCH pat;
   XYZ_SPLINE_TYPE cur;
   double junk[50];
   double th[50];
   double sh[50];
   int nh;

/* we will use bezier clipping */

   /* convert c1 to tensor patch, c2 to xyz_spline */
   pat.iDeg0 = c1->Deg;
   pat.iDeg1 = 0;
   allocate_tensor_patch(&pat);
   for(i=0;i<=c1->Deg;i++) {
      pat.coeffs[i][0].dCoord[0] = c1->CntrlPts[i].dCoord[0];
      pat.coeffs[i][0].dCoord[1] = c1->CntrlPts[i].dCoord[1];
      pat.coeffs[i][0].dCoord[2] = 0.0;
      pat.coeffs[i][0].dCoord[3] = c1->CntrlPts[i].dCoord[2];
   }

   cur.Deg = c2->Deg;
   allocate_xyz_spline(&cur);
   for (i=0;i<=c2->Deg;i++) {
      cur.CntrlPts[i].dCoord[0] = c2->CntrlPts[i].dCoord[0];
      cur.CntrlPts[i].dCoord[1] = c2->CntrlPts[i].dCoord[1];
      cur.CntrlPts[i].dCoord[2] = 0.0;
      cur.CntrlPts[i].dCoord[3] = c2->CntrlPts[i].dCoord[2];
   }

   curve_surface_intersect(&pat,&cur,0.0001,0.0001,&nh,th,sh,junk);

if (!FP_EQ_EPS(nh,numinter[0],0.001)) printf("ERROR - # of hit points is different! %d vs. %d\n",nh,numinter[0]);
if (!FP_EQ_EPS(sh[0],s[0],0.001)) printf("ERROR - first S value is different! %lf vs. %lf\n",sh[0],s[0]);
if (!FP_EQ_EPS(th[0],t[0],0.001)) printf("ERROR - first T value is different! %lf vs. %lf\n",th[0],t[0]);


#endif
}
