#include <triangulate.h>
#include <defs.h>
#include <errno.h>

#define CROSS_SINE(v0, v1) ((v0).x * (v1).y - (v1).x * (v0).y)

#define NORMALIZE(v, denom) v.x /= denom;\
                            v.y /= denom

#define INPOLY_EPS 1.0e-3

/***********************************************************/
#define TOLERANCE_4D 0.5e-3
#define FP_EQUAL_4D(a, b)  ((((a) - (b)) < TOLERANCE_4D) &&  \
			    (((a) - (b)) > -TOLERANCE_4D)) 

#define EQUAL_4D(a, b) ((FP_EQUAL_4D((a).dCoord[0]/(a).dCoord[3],\
				     (b).dCoord[0]/(b).dCoord[3])) && \
			(FP_EQUAL_4D((a).dCoord[1]/(a).dCoord[3],\
				     (b).dCoord[1]/(b).dCoord[3])) && \
			(FP_EQUAL_4D((a).dCoord[2]/(a).dCoord[3],\
				     (b).dCoord[2]/(b).dCoord[3])))

/***********************************************************/

#define SQR_LENGTH(v0, v1) (((v0).x - (v1).x)*((v0).x - (v1).x) + \
                             ((v0).y - (v1).y)*((v0).y - (v1).y))

#define MAX_SUBCHAINS 4

#define PT_IN 1			/* do not change these values !! */
#define PT_OUT 0

static node_t qs2[2][QSIZE];		/* Query structure */
static trap_t tr2[2][TRSIZE];		/* Trapezoid structure */
static segment_t seg2[2][SEGSIZE];	/* Segment table */

static struct global_s global2[2];      /* global struct to hold */
					/* #segments */ 

static struct {
  int numpoints;
  POINT4D pc[SEGSIZE];			/* array to hold the */
} portion[MAX_SUBCHAINS];	/* sub-chains  */

static int initialise2(segment_t *seg, int nseg)
{
  register int i;

  for (i = 1; i <= nseg; i++)
    seg[i].is_inserted = FALSE;
  generate_random_ordering2(nseg);
  
  return 0;
}

static int get_3d_point(POINT3D *spt, POINT3D *ept, double t, POINT3D
		    *ret_pt)
{
  ret_pt->dCoord[0] = spt->dCoord[0] + t*(ept->dCoord[0] - spt->dCoord[0]);
  ret_pt->dCoord[1] = spt->dCoord[1] + t*(ept->dCoord[1] - spt->dCoord[1]);
  ret_pt->dCoord[2] = spt->dCoord[2] + t*(ept->dCoord[2] - spt->dCoord[2]);

  return 0;
}

static int get_intrn_point(TENSOR_PATCH *tp0, TENSOR_PATCH *tp1,
			   POINT4D *p0, POINT4D *p1, POINT4D *v,
			   double t, int wpatch)
{
  if (wpatch == W_UV)
    {
      /* calculate correct uv by interpolation */
      v->dCoord[0] = p0->dCoord[0] + t*(p1->dCoord[0] -
					p0->dCoord[0]);
      v->dCoord[1] = p0->dCoord[1] + t*(p1->dCoord[1] -
					p0->dCoord[1]);
      
      extract_st_val(*tp0, *tp1, v->dCoord[0], U, v->dCoord[1],
		     &(v->dCoord[2]), &(v->dCoord[3]));
    }
  else 
    {
      /* calculate correct st by interpolation */
      v->dCoord[2] = p0->dCoord[2] + t*(p1->dCoord[2] - p0->dCoord[2]);
      v->dCoord[3] = p0->dCoord[3] + t*(p1->dCoord[3] -
					p0->dCoord[3]);

      extract_st_val(*tp1, *tp0, v->dCoord[2], U, v->dCoord[3],
		     &(v->dCoord[0]), &(v->dCoord[1]));
    }
  return 0;
}

static int my_get_intrn_point(TENSOR_PATCH *tp0, TENSOR_PATCH *tp1,
			      POINT4D *p0, POINT4D *p1, POINT4D *v,
			      double t, int wpatch)
{
  if (wpatch == W_UV)
    {
      /* calculate correct uv by interpolation */
      v->dCoord[0] = p0->dCoord[0] + t*(p1->dCoord[0] -
					p0->dCoord[0]);
      v->dCoord[1] = p0->dCoord[1] + t*(p1->dCoord[1] -
					p0->dCoord[1]);
      
      v->dCoord[2] = p0->dCoord[2] + t*(p1->dCoord[2] -
					p0->dCoord[2]);
      v->dCoord[3] = p0->dCoord[3] + t*(p1->dCoord[3] -
					p0->dCoord[3]);
    }
  else 
    {
      /* calculate correct st by interpolation */
      v->dCoord[2] = p0->dCoord[2] + t*(p1->dCoord[2] - p0->dCoord[2]);
      v->dCoord[3] = p0->dCoord[3] + t*(p1->dCoord[3] -
					p0->dCoord[3]);

      v->dCoord[0] = p0->dCoord[0] + t*(p1->dCoord[0] -
					p0->dCoord[0]);
      v->dCoord[1] = p0->dCoord[1] + t*(p1->dCoord[1] -
					p0->dCoord[1]);
    }
  return 0;
}

/* Determine whether the polygon is simple or not */
int is_simple_polygon2(segment_t seg[], int npoints)
{
  register int i, j;
  point_typ p0, p1;
  double s0, s1, t0, t1;
  int itype;

  for (i = 0; i < npoints; i++)
    for (j = (i+2)%npoints; j != (i + npoints - 1)%npoints; 
	 j = (j + 1)%npoints)
      if (segments_intersect(&seg[i+1].v0, &seg[i+1].v1,
			     &seg[j+1].v0, &seg[j+1].v1, &itype,
			     &s0, &s1, &t0, &t1, &p0, &p1) > 0)
	return FALSE;
  
  return TRUE;
}


static int is_pt_inside_poly(segment_t *seg, trap_t *tr, node_t *qs,
			     int npoints, point_typ *v)
{
  register int i;
  double k1, k2, k3, denom;
  trap_t *t;
  vector_t u0, u1, vec;
  int rseg, trnum;
  
  /* first check if the point lies on any of the edges. If yes, return */
  /* TRUE. */
  
  for (i = 1; i <= npoints; i++)
    {
      u0.x = seg[i].v1.x - seg[i].v0.x;
      u0.y = seg[i].v1.y - seg[i].v0.y;
      
      denom = sqrt(DOT(u0, u0));
      NORMALIZE(u0, denom);

      vec.x = v->x - seg[i].v0.x;
      vec.y = v->y - seg[i].v0.y;

      denom = sqrt(DOT(vec, vec));
      NORMALIZE(vec, denom);

      /* if (fabs(CROSS(seg[i].v0, seg[i].v1, *v)) <= INPOLY_EPS) */
      if (fabs(CROSS_SINE(u0, vec)) <= INPOLY_EPS)
	{
	  k1 = SQR_LENGTH(seg[i].v0, seg[i].v1);
	  k2 = SQR_LENGTH(seg[i].v0, *v);
	  k3 = SQR_LENGTH(seg[i].v1,*v);
	  
	  if ((k2 <= k1) && (k3 <= k1))
	    return TRUE;
	}
    }
  trnum = locate_endpoint2(seg, tr, qs, v, v, 1);
  t = &tr[trnum];
  
  if (t->state == ST_INVALID)
    return 0;
  
  if ((t->lseg <= 0) || (t->rseg <= 0))
    return 0;
  rseg = t->rseg;
  return _greater_than_equal_to(&seg[rseg].v1, &seg[rseg].v0);
}


static int obtain_proper_trimcurve(TENSOR_PATCH *tp0, 
				   TENSOR_PATCH *tp1, 
				   segment_t seg0[], POINT3D *n3dsegs0,
				   segment_t seg1[], POINT3D *n3dsegs1,
				   int n0, int n1, POINT4D curve[],
				   int nc)
{
  register int i, j;
  int is_in0, is_in1, was_in0, was_in1;
  point_typ p0, p1, q0, q1, v0, v1;
  double s0, t0, s1, t1, maxt, mint;
  double s01, t01, s11, t11;
  int por_idx, pnum, nintr0, nintr1;
  POINT4D pt;
  int wpatch;

  pnum = 0;
  por_idx = 0;

  /* first triangulate each polygon */
  
  if (!is_simple_polygon2(seg0, n0))
    {
      fprintf(stderr, "obtain_proper_trimcurve: polygon not simple\n"); 
#ifdef DEBUG_PRINT_POLY
      for (i = 1; i <= n0; i++)
	printf("%lf %lf\n", seg0[i].v0.x, seg0[i].v0.y);
#endif
/*      exit(1); */
    }
  global2[0].nseg = n0;
  global2[1].nseg = n1;
  initialise2(seg0, n0);
  construct_trapezoids2(seg0, tr2[0], qs2[0], global2[0], n0);

  /* triangulate second poly */

  if (!is_simple_polygon2(seg1, n1))
    {
      fprintf(stderr, "obtain_proper_trimcurve: polygon not simple\n"); 
#ifdef DEBUG_PRINT_POLY
      for (i = 1; i <= n1; i++)
	printf("%lf %lf\n", seg1[i].v0.x, seg1[i].v0.y);
#endif
      exit(1);
    }
  initialise2(seg1, n1);
  construct_trapezoids2(seg1, tr2[1], qs2[1], global2[1], n1);

  
  /* Now trace out the curve performing the in-out step at each new */
  /* point */

  p0.x = curve[0].dCoord[2];
  p0.y = curve[0].dCoord[3];
  q0.x = curve[0].dCoord[0];
  q0.y = curve[0].dCoord[1];
  was_in0 = is_pt_inside_poly(seg0, tr2[0], qs2[0], n0, &p0);
  was_in1 = is_pt_inside_poly(seg1, tr2[1], qs2[1], n1, &q0);

  /* If both the points are in, add both to the intersection curve */ 
  if ((was_in0 == PT_IN) && (was_in1 == PT_IN))
    {
      pnum = 0;
      por_idx = 0;
      portion[por_idx].pc[pnum++] = curve[0];
    }
  else
    por_idx = -1;

  for (i = 1; i < nc; i++)
    {
      int first_index0, sec_index0;
      int first_index1, sec_index1;
      double maxs;
      POINT3D tmp3dpt;

      p1.x = curve[i].dCoord[2];
      p1.y = curve[i].dCoord[3];
      q1.x = curve[i].dCoord[0];
      q1.y = curve[i].dCoord[1];

      is_in0 = is_pt_inside_poly(seg0, tr2[0], qs2[0], n0, &p1);
      is_in1 = is_pt_inside_poly(seg1, tr2[1], qs2[1], n1, &q1);

      if ((nintr0 =
	   find_intersections_of_segment_with_polygon(&seg0[1], 
        	     global2[0], &p0, &p1, &s0, 
		     &t0, &first_index0, &s01, &t01, 
		     &sec_index0, was_in0, is_in0)) > 2)
	{
	  /* more than one intersection. Just flag this condition */
#ifdef DEBUG_PRINT_POLY
	  printf("obtain_proper_trimcurve: > 1 intersections detected: Using the last one \n"); 
#endif
	  printf("");
	}

      if ((nintr1 =
	   find_intersections_of_segment_with_polygon(&seg1[1], 
        	     global2[1], &q0, &q1, &s1, 
                     &t1, &first_index1, &s11, &t11, 
                     &sec_index1, was_in1, is_in1)) > 2)
	printf("");
#ifdef DEBUG_PRINT_POLY
	printf("obtain_proper_trimcurve: > 1 intersections detected: Using the last one \n"); 
#endif
      
      switch (was_in0)
	{
	case PT_OUT:		/* last point was OUT */
	  
	  switch (is_in0)
	    {
	    case PT_OUT:	/* out-out */
	      {
		if (nintr0 < 2)
		  break;
		else {
		  if ((was_in1 == PT_OUT) && (is_in1 == PT_OUT)) {
		    if (nintr1 < 2)
		      break;
		    maxt = MAX(t0, t1);
		    if (maxt == t0)
		      wpatch = W_ST;
		    else
		      wpatch = W_UV;

		    my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				       &pt, maxt, wpatch); 

/*
		    if (maxt == t0) {
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
		    } else {
		      maxs = s1;
		      get_3d_point(&n3dsegs1[first_index1],
				   &n3dsegs1[first_index1+1], maxs,
				   &tmp3dpt);
		    }
*/
		    /* initiate new chain */
		    pnum = 0;
		    por_idx++;
/*
		    start[por_idx] = tmp3dpt;
*/
		    portion[por_idx].pc[pnum++] = pt;
		    check_for_duplication(por_idx, &pnum);

		    maxt = MIN(t01, t11);
		    if (maxt == t01)
		      wpatch = W_ST;
		    else
		      wpatch = W_UV;

		    my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				       &pt, maxt, wpatch); 
/*
		    if (maxt == t01) {
		      maxs = s01;
		      get_3d_point(&n3dsegs0[sec_index0],
				   &n3dsegs0[sec_index0+1], maxs,
				   &tmp3dpt);
		    } else {
		      maxs = s11;
		      get_3d_point(&n3dsegs1[sec_index1],
				   &n3dsegs1[sec_index1+1], maxs,
				   &tmp3dpt);
		    }
*/
		    portion[por_idx].pc[pnum++] = pt;
/*
		    end[por_idx] = tmp3dpt;
*/
		    check_for_duplication(por_idx, &pnum);
		    portion[por_idx].numpoints = pnum;
		  } else if ((was_in1 == PT_OUT) && (is_in1 == PT_IN)) {
		    maxt = MAX(t0, t1);
		    if (maxt == t0)
		      wpatch = W_ST;
		    else
		      wpatch = W_UV;

		    my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				       &pt, maxt, wpatch); 
/*		    
		    if (maxt == t0) {
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
		    } else {
		      maxs = s1;
		      get_3d_point(&n3dsegs1[first_index1],
				   &n3dsegs1[first_index1+1], maxs,
				   &tmp3dpt);
		    }
*/
		    /* initiate new chain */
		    pnum = 0;
		    por_idx++;
/*
		    start[por_idx] = tmp3dpt;
*/
		    portion[por_idx].pc[pnum++] = pt;
		    check_for_duplication(por_idx, &pnum);

		    maxt = t01;
		    wpatch = W_ST;
/*
		    maxs = s01;
		    get_3d_point(&n3dsegs0[sec_index0],
				 &n3dsegs0[sec_index0+1], maxs,
				 &tmp3dpt);
*/
		    my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				       &pt, maxt, wpatch); 
/*
		    end[por_idx] = tmp3dpt;
*/
		    portion[por_idx].pc[pnum++] = pt;
		    check_for_duplication(por_idx, &pnum);
		    portion[por_idx].numpoints = pnum;
		  } else if ((was_in1 == PT_IN) && (is_in1 == PT_OUT)) {
		    maxt = MIN(t0, t1);
		    if (maxt == t0) {
		      wpatch = W_ST;
		      
/*
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
*/
		      get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				      &pt, maxt, wpatch); 
		      
		      /* initiate new chain */
		      pnum = 0;
		      por_idx++;
/*
		      start[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      check_for_duplication(por_idx, &pnum);
		      
		      maxt = MIN(t01, t1);
		      if (maxt == t01)
			wpatch = W_ST;
		      else
			wpatch = W_UV;
		      
/*
		      if (maxt == t01) {
			maxs = s01;
			get_3d_point(&n3dsegs0[sec_index0],
				     &n3dsegs0[sec_index0+1], maxs,
				     &tmp3dpt);
		      } else {
			maxs = s11;
			get_3d_point(&n3dsegs1[sec_index1],
				     &n3dsegs1[sec_index1+1], maxs,
				     &tmp3dpt);
		      }
*/
		      get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				      &pt, maxt, wpatch); 
		      
/*
		      end[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      check_for_duplication(por_idx, &pnum);
		      portion[por_idx].numpoints = pnum;
		    }
		  } else if ((was_in1 == PT_IN) && (is_in1 == PT_IN)) {
		    if (nintr1 < 2) {
		      maxt = t0;
		      wpatch = W_ST;

/*
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
*/
		      my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					 &pt, maxt, wpatch);

		      /* initiate new chain */
		      pnum = 0;
		      por_idx++;
/*
		      start[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      check_for_duplication(por_idx, &pnum);

		      maxt = t01;
		      wpatch = W_ST;

/*
		      maxs = s01;
		      get_3d_point(&n3dsegs0[sec_index0],
				   &n3dsegs0[sec_index0+1], maxs,
				   &tmp3dpt);
*/
		      my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					 &pt, maxt, wpatch);

/*
		      end[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      check_for_duplication(por_idx, &pnum);
		      portion[por_idx].numpoints = pnum;
		    } else {
		      maxt = MIN(t0, t1);
		      if (maxt == t0) {
			wpatch = W_ST;
		      
/*
			maxs = s0;
			get_3d_point(&n3dsegs0[first_index0],
				     &n3dsegs0[first_index0+1], maxs,
				     &tmp3dpt);
*/
			my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					   &pt, maxt, wpatch); 
		      
			/* initiate new chain */
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);

			maxt = MIN(t01, t1);
			if (maxt == t01) {
			  wpatch = W_ST;
/*
			  maxs = s01;
			  get_3d_point(&n3dsegs0[sec_index0],
				       &n3dsegs0[sec_index0+1], maxs,
				       &tmp3dpt);
*/
			  my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					     &pt, maxt, wpatch);
			  
/*
			  end[por_idx] = tmp3dpt;
*/
			  portion[por_idx].pc[pnum++] = pt;
			  check_for_duplication(por_idx, &pnum);
			  portion[por_idx].numpoints = pnum;
			} else {
			  wpatch = W_UV;
/*
			  maxs = s1;
			  get_3d_point(&n3dsegs1[first_index1],
				       &n3dsegs1[first_index1+1], maxs,
				       &tmp3dpt);
*/
			  my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					     &pt, maxt, wpatch);
			  
/*
			  end[por_idx] = tmp3dpt;
*/
			  portion[por_idx].pc[pnum++] = pt;
			  check_for_duplication(por_idx, &pnum);
			  portion[por_idx].numpoints = pnum;

			  maxt = MIN(t01, t11);
			  if (maxt == t11) {
			    wpatch = W_UV;
/*
			    maxs = s11;
			    get_3d_point(&n3dsegs1[sec_index1],
					 &n3dsegs1[sec_index1+1], maxs,
					 &tmp3dpt);
*/
			    my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					       &pt, maxt, wpatch); 
			    
			    /* initiate new chain */
			    pnum = 0;
			    por_idx++;
/*
			    start[por_idx] = tmp3dpt;
*/
			    portion[por_idx].pc[pnum++] = pt;
			    check_for_duplication(por_idx, &pnum);

			    maxt = t01;
/*
			    maxs = s01;
			    get_3d_point(&n3dsegs0[sec_index0],
					 &n3dsegs0[sec_index0+1], maxs,
					 &tmp3dpt);
*/
			    wpatch = W_ST;
			    my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					       &pt, maxt, wpatch);
			    
/*
			    end[por_idx] = tmp3dpt;
*/
			    portion[por_idx].pc[pnum++] = pt;
			    check_for_duplication(por_idx, &pnum);
			    portion[por_idx].numpoints = pnum;
			  }
			}
		      } else {
			maxt = MAX(t0, t11);
			if (maxt == t0)
			  wpatch = W_ST;
			else
			  wpatch = W_UV;

/*
			if (maxt == t0) {
			  maxs = s0;
			  get_3d_point(&n3dsegs0[first_index0],
				       &n3dsegs0[first_index0+1], maxs,
				       &tmp3dpt);
			} else {
			  maxs = s11;
			  get_3d_point(&n3dsegs1[sec_index1],
				       &n3dsegs1[sec_index1+1], maxs,
				       &tmp3dpt);
			}
*/
			my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					   &pt, maxt, wpatch); 
			
			/* initiate new chain */
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);

			maxt = t01;
			wpatch = W_ST;
/*
			maxs = s01;
			get_3d_point(&n3dsegs0[sec_index0],
				     &n3dsegs0[sec_index0+1], maxs,
				     &tmp3dpt);
*/
			my_get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					   &pt, maxt, wpatch);
			
/*
			end[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);
			portion[por_idx].numpoints = pnum;
		      }
		    }
		  }
		}
	      }
	      break;		
	    case PT_IN:		/* out-in */
	      {			/* start a new chain if possible */
		if ((was_in1 == PT_OUT) && (is_in1 == PT_OUT)) {
		  if (nintr1 < 2)
		    break;
		  maxt = MAX(t0, t1);
		  if (maxt == t0)
		    wpatch = W_ST;
		  else
		    wpatch = W_UV;
		  
/*
		  if (maxt == t0) {
		    maxs = s0;
		    get_3d_point(&n3dsegs0[first_index0],
				 &n3dsegs0[first_index0+1], maxs,
				 &tmp3dpt);
		  } else {
		    maxs = s1;
		    get_3d_point(&n3dsegs1[first_index1],
				 &n3dsegs1[first_index1+1], maxs,
				 &tmp3dpt);
		  }
*/
		  get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				  &pt, maxt, wpatch); 
		  
		  /* initiate new chain */
		  pnum = 0;
		  por_idx++;
/*
		  start[por_idx] = tmp3dpt;
*/
		  portion[por_idx].pc[pnum++] = pt;
		  check_for_duplication(por_idx, &pnum);
		  
		  maxt = t11;
		  wpatch = W_UV;
/*
		  maxs = s11;
		  get_3d_point(&n3dsegs1[sec_index1],
			       &n3dsegs1[sec_index1+1], maxs,
			       &tmp3dpt);
*/
		  get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				  &pt, maxt, wpatch); 
/*
		  end[por_idx] = tmp3dpt;
*/
		  portion[por_idx].pc[pnum++] = pt;
		  check_for_duplication(por_idx, &pnum);
		  portion[por_idx].numpoints = pnum;
		} else if ((was_in1 == PT_OUT) && (is_in1 == PT_IN))
		  {
		    /* use the greater parameter point */
		    /* for intersection */
		    maxt = MAX(t0, t1);
		    if (maxt == t0)
		      wpatch = W_ST;
		    else
		      wpatch = W_UV;

/*
		    if (maxt == t0) {
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
		    } else {
		      maxs = s1;
		      get_3d_point(&n3dsegs1[first_index1],
				   &n3dsegs1[first_index1+1], maxs,
				   &tmp3dpt);
		    }
*/
		    get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				    &pt, maxt, wpatch); 
		    
		    /* initiate new chain */
		    pnum = 0;
		    por_idx++;
/*
		    start[por_idx] = tmp3dpt;
*/
		    portion[por_idx].pc[pnum++] = pt;
		    portion[por_idx].pc[pnum++] = curve[i];
		    check_for_duplication(por_idx, &pnum);
		  }
		else if ((was_in1 == PT_IN) && (is_in1 == PT_OUT))
		  {
		    if (t0 < t1) /* make-and-terminate-small chain */
		      {
			wpatch = W_ST;
/*
			maxs = s0;
			get_3d_point(&n3dsegs0[first_index0],
				     &n3dsegs0[first_index0+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1],
					&curve[i], &pt, t0, wpatch);
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;

			wpatch = W_UV;
/*
			maxs = s1;
			get_3d_point(&n3dsegs1[first_index1],
				     &n3dsegs1[first_index1+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1],
					&curve[i], &pt, t1, wpatch);
/*
			end[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);
			portion[por_idx].numpoints = pnum;
#ifdef DEBUG1
			printf("proper_trim: 2 point chain\n");
			printf("(%lf %lf %lf %lf) (%lf %lf %lf %lf)\n",
			       portion[por_idx].pc[0].dCoord[0],
			       portion[por_idx].pc[0].dCoord[1],
			       portion[por_idx].pc[0].dCoord[2],
			       portion[por_idx].pc[0].dCoord[3],
			       portion[por_idx].pc[1].dCoord[0],
			       portion[por_idx].pc[1].dCoord[1],
			       portion[por_idx].pc[1].dCoord[2],
			       portion[por_idx].pc[1].dCoord[3]);
#endif
		      }
		  }
		else if ((was_in1 == PT_IN) && (is_in1 == PT_IN))
		  {
		    if (nintr1 < 2) {
		      /* Initiate new chain */
		      wpatch = W_ST;
/*
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
*/
		      get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				      &pt, t0, wpatch);
		      pnum = 0;
		      por_idx++;
/*
		      start[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      portion[por_idx].pc[pnum++] = curve[i];
		      check_for_duplication(por_idx, &pnum);
		    } else {
		      maxt = MIN(t0, t1);
		      if (maxt == t0) {
			wpatch = W_ST;
/*
			maxs = s0;
			get_3d_point(&n3dsegs0[first_index0],
				     &n3dsegs0[first_index0+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					&pt, maxt, wpatch);
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);

			wpatch = W_UV;
/*
			maxs = s1;
			get_3d_point(&n3dsegs1[first_index1],
				     &n3dsegs1[first_index1+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1],
					&curve[i], &pt, t1, wpatch);
/*
			end[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);
			portion[por_idx].numpoints = pnum;

			wpatch = W_UV;
/*
			maxs = s11;
			get_3d_point(&n3dsegs1[sec_index1],
				     &n3dsegs1[sec_index1+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					&pt, t11, wpatch);
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			portion[por_idx].pc[pnum++] = curve[i];
			check_for_duplication(por_idx, &pnum);
		      } else {
			maxt = MAX(t0, t11);
			if (maxt == t0) 
			  wpatch = W_ST;
			else
			  wpatch = W_UV;
			
/*
			if (maxt == t0) {
			  maxs = s0;
			  get_3d_point(&n3dsegs0[first_index0],
				       &n3dsegs0[first_index0+1], maxs,
				       &tmp3dpt);
			} else {
			  maxs = s11;
			  get_3d_point(&n3dsegs1[sec_index1],
				       &n3dsegs1[sec_index1+1], maxs,
				       &tmp3dpt);
			}
*/
			get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
					&pt, maxt, wpatch);
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			portion[por_idx].pc[pnum++] = curve[i];
			check_for_duplication(por_idx, &pnum);
		      }
		    }
		  }
		break;
	      }
	    } /* end-inner switch */
	  break;
	  
	case PT_IN:		/* last_point was IN */
	  
	  switch (is_in0)
	    {
	    case PT_OUT:	/* in-out */
	      {
		if ((was_in1 == PT_OUT) && (is_in1 == PT_OUT)) {
		  if (nintr1 < 2)
		    break;
		  maxt = MIN(t0, t1);
		  if (maxt == t1) {
		    wpatch = W_UV;
		    
/*
		    maxs = s1;
		    get_3d_point(&n3dsegs1[first_index1],
				 &n3dsegs1[first_index1+1], maxs,
				 &tmp3dpt);
*/
		    get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				    &pt, maxt, wpatch); 
		    
		    /* initiate new chain */
		    pnum = 0;
		    por_idx++;
/*
		    start[por_idx] = tmp3dpt;
*/
		    portion[por_idx].pc[pnum++] = pt;
		    check_for_duplication(por_idx, &pnum);
		    
		    maxt = MIN(t0, t11);
		    if (maxt == t0)
		      wpatch = W_ST;
		    else
		      wpatch = W_UV;
		    
/*
		    if (maxt == t0) {
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
		    } else {
		      maxs = s11;
		      get_3d_point(&n3dsegs1[sec_index1],
				   &n3dsegs1[sec_index1+1], maxs,
				   &tmp3dpt);
		    }
*/
		    get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				    &pt, maxt, wpatch); 
		    
/*
		    end[por_idx] = tmp3dpt;
*/
		    portion[por_idx].pc[pnum++] = pt;
		    check_for_duplication(por_idx, &pnum);
		    portion[por_idx].numpoints = pnum;
		  }
		} else if ((was_in1 == PT_OUT) && (is_in1 == PT_IN))
		  {
		    if (t0 > t1) /* make-and-terminate-small chain */
		      {
			wpatch = W_UV;
/*
			maxs = s1;
			get_3d_point(&n3dsegs1[first_index1],
				     &n3dsegs1[first_index1+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1],
					&curve[i], &pt, t1, wpatch);
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;

			wpatch = W_ST;
/*
			maxs = s0;
			get_3d_point(&n3dsegs0[first_index0],
				     &n3dsegs0[first_index0+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1],
					&curve[i], &pt, t0, wpatch);
/*
			end[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);
			portion[por_idx].numpoints = pnum;
#ifdef DEBUG1
			printf("proper_trim: 2 point chain\n");
			printf("(%lf %lf %lf %lf) (%lf %lf %lf %lf)\n",
			       portion[por_idx].pc[0].dCoord[0],
			       portion[por_idx].pc[0].dCoord[1],
			       portion[por_idx].pc[0].dCoord[2],
			       portion[por_idx].pc[0].dCoord[3],
			       portion[por_idx].pc[1].dCoord[0],
			       portion[por_idx].pc[1].dCoord[1],
			       portion[por_idx].pc[1].dCoord[2],
			       portion[por_idx].pc[1].dCoord[3]);
#endif
		      }
		    
		  }
		else if ((was_in1 == PT_IN) && (is_in1 == PT_OUT))
		  {

		    /* use the smaller parameter point */
		    /* for intersection */
		    mint = MIN(t0, t1);
		    if (mint == t0)
		      wpatch = W_ST;
		    else
		      wpatch = W_UV;

/*
		    if (maxt == t0) {
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
		    } else {
		      maxs = s1;
		      get_3d_point(&n3dsegs1[first_index1],
				   &n3dsegs1[first_index1+1], maxs,
				   &tmp3dpt);
		    }
*/
		    get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				    &pt, mint, wpatch); 
		    
		    /* terminate the existing chain */
/*
		    end[por_idx] = tmp3dpt;
*/
		    portion[por_idx].pc[pnum++] = pt;		    
		    check_for_duplication(por_idx, &pnum);
		    portion[por_idx].numpoints = pnum;
		  }
		else if ((was_in1 == PT_IN) && (is_in1 == PT_IN))
		  {
		    if (nintr1 < 2) {
		      wpatch = W_ST;
/*
		      maxs = s0;
		      get_3d_point(&n3dsegs0[first_index0],
				   &n3dsegs0[first_index0+1], maxs,
				   &tmp3dpt);
*/
		      get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				      &pt, t0, wpatch);
		      /* terminate the chain */
		      
/*
		      end[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      check_for_duplication(por_idx, &pnum);
		      portion[por_idx].numpoints = pnum;
		    } else {
		      maxt = MIN(t0, t1);
		      if (maxt == t0)
			wpatch = W_ST;
		      else
			wpatch = W_UV;

/*
		      if (maxt == t0) {
			maxs = s0;
			get_3d_point(&n3dsegs0[first_index0],
				     &n3dsegs0[first_index0+1], maxs,
				     &tmp3dpt);
		      } else {
			maxs = s1;
			get_3d_point(&n3dsegs1[first_index1],
				     &n3dsegs1[first_index1+1], maxs,
				     &tmp3dpt);
		      }
*/
		      get_intrn_point(tp0, tp1, &curve[i-1], &curve[i],
				      &pt, maxt, wpatch);
		      /* terminate the chain */
/*
		      end[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      check_for_duplication(por_idx, &pnum);
		      portion[por_idx].numpoints = pnum;

		      if (t11 < t0) {
			wpatch = W_UV;
/*
			maxs = s11;
			get_3d_point(&n3dsegs1[sec_index1],
				     &n3dsegs1[sec_index1+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1],
					&curve[i], &pt, t11, wpatch);
			pnum = 0;
			por_idx++;
/*
			start[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);

			wpatch = W_ST;
/*
			maxs = s0;
			get_3d_point(&n3dsegs0[first_index0],
				     &n3dsegs0[first_index0+1], maxs,
				     &tmp3dpt);
*/
			get_intrn_point(tp0, tp1, &curve[i-1],
					&curve[i], &pt, t0, wpatch);
/*
			end[por_idx] = tmp3dpt;
*/
			portion[por_idx].pc[pnum++] = pt;
			check_for_duplication(por_idx, &pnum);
			portion[por_idx].numpoints = pnum;
		      }
		    }
		  }
		
		break;	      
		
	      case PT_IN:		/* in-in */
		{			/* start a new chain if possible */
		  double maxt;
		  
		  if ((was_in1 == PT_OUT) && (is_in1 == PT_OUT))
		    break;
		  else if ((was_in1 == PT_OUT) && (is_in1 == PT_IN))
		    {
		      /* begin new chain */
		      wpatch = W_UV;
/*
		      maxs = s1;
		      get_3d_point(&n3dsegs1[first_index1],
				   &n3dsegs1[first_index1+1], maxs,
				   &tmp3dpt);
*/
		      get_intrn_point(tp0, tp1, &curve[i-1],
				      &curve[i], &pt, t1, wpatch); 
		      pnum = 0;
		      por_idx++;
/*
		      start[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		    }
		  else if ((was_in1 == PT_IN) && (is_in1 == PT_OUT))
		    {
		      /* terminate existing chain */
		      wpatch = W_UV;		      
/*
		      maxs = s1;
		      get_3d_point(&n3dsegs1[first_index1],
				   &n3dsegs1[first_index1+1], maxs,
				   &tmp3dpt);
*/
		      get_intrn_point(tp0, tp1, &curve[i-1],
				      &curve[i], &pt, t1, wpatch); 
/*
		      end[por_idx] = tmp3dpt;
*/
		      portion[por_idx].pc[pnum++] = pt;
		      check_for_duplication(por_idx, &pnum);
		      portion[por_idx].numpoints = pnum;
		    }
		  else if ((was_in1 == PT_IN) && (is_in1 == PT_IN))
		    {
		      /* Simply add the new point into the existing */
		      /* chain */ 
		      portion[por_idx].pc[pnum++] = curve[i];
		      check_for_duplication(por_idx, &pnum);
		    }
		  
		  break;
		}
	      }
	    }
	  break;
	}  /* end outer switch */

      p0 = p1;
      q0 = q1;
      was_in0 = is_in0;
      was_in1 = is_in1;

    } /* end of for loop */
  
  if ((is_in0 == PT_IN) && (is_in1 == PT_IN))
    portion[por_idx].numpoints = pnum;
  
  return (por_idx + 1);
}


int check_for_duplication(int por_idx, int *pnum)
{
  /* Check if pc[pnum-1] and pc[pnum-2] are almost same. If yes, */
  /* decrement pnum */
  if (EQUAL_4D(portion[por_idx].pc[*pnum - 1], 
	       portion[por_idx].pc[*pnum - 2]))
    (*pnum)--;
  
  return 0;
}


#ifdef TEST_PROPER

int read_data2(POINT4D curve[])
{
  register int i, j;
  int n0, n1, m;
  
  scanf("%d", &n0);
  global2[0].nseg = n0;
  for (i = 1; i <= n0; i++)
    {
      scanf("%lf%lf", &seg2[0][i].v0.x, &seg2[0][i].v0.y);
      if (i == 1)
        {
          seg2[0][n0].v1.x = seg2[0][1].v0.x;
          seg2[0][n0].v1.y = seg2[0][1].v0.y;
        }
      else
        {
          seg2[0][i - 1].v1.x = seg2[0][i].v0.x;
          seg2[0][i - 1].v1.y = seg2[0][i].v0.y;
        }       
      seg2[0][i].is_inserted = FALSE;
    }

  scanf("%d", &n0);
  global2[1].nseg = n0;
  for (i = 1; i <= n0; i++)
    {
      scanf("%lf%lf", &seg2[1][i].v0.x, &seg2[1][i].v0.y);
      if (i == 1)
        {
          seg2[1][n0].v1.x = seg2[1][1].v0.x;
          seg2[1][n0].v1.y = seg2[1][1].v0.y;
        }
      else
        {
          seg2[1][i - 1].v1.x = seg2[1][i].v0.x;
          seg2[1][i - 1].v1.y = seg2[1][i].v0.y;
        }       
      seg2[1][i].is_inserted = FALSE;
    }

  scanf("%d", &m);		/* npoints */
  for (i = 0; i < m; i++)
    {
      scanf("%lf%lf%lf%lf", &curve[i].dCoord[0], &curve[i].dCoord[1],
	    &curve[i].dCoord[2], &curve[i].dCoord[3]);
    }
  
  return m;
}

void *new(int size)
{
  void *p;

  if ((p = malloc(size)) == NULL)
    perror("malloc");
  
  return p;
}


main()
{
  POINT4D curve[SEGSIZE];
  int m, n;
  
  m = read_data2(curve);
  obtain_proper_trimcurve(seg2[0], seg2[1], global2[0].nseg, 
			   global2[1].nseg, curve, m);
  
  return 0;
}

#endif



#ifdef CONTINUE

/* Return value: the number curves generated  as  */
int obtain_correct_trimcurve(TRIM_PATCH_INFO *tp0, 
			     TRIM_PATCH_INFO *tp1, 
			     INT_CURVE_STRUCT *incurve,
			     INT_CURVE_STRUCT **outcurve)
{
  register int i, j;
  int n0, n1, nc, k;
  POINT4D curve[SEGSIZE];
  POINT3D n3dsegs0[SEGSIZE], n3dsegs1[SEGSIZE];
  POINT2D pt;
  LIST_POINT4D *p, *q;
  int nportions, actual_nportions;
  int num_pts;
  double step_size;
  UV_SPLINE_TYPE *ic;
  XYZ_SPLINE_TYPE *xyzc;
  
  if ((tp0->NumTrimPortions != 1) || (tp1->NumTrimPortions != 1))
    {
      fprintf(stderr, "obtain_correct_trimcurve: > 1 portions\n");
      return -1;
    }

  if (!tp0->SplineReprn) {
    n0  = tp0->TrimPortion[0].PiecePortion.NumOfVertices;
    for (i = 1; i <= n0; i++)
      {
	seg2[0][i].v0.x =  tp0->TrimPortion[0].PiecePortion.poly_vertex[i][0];
	seg2[0][i].v0.y =  tp0->TrimPortion[0].PiecePortion.poly_vertex[i][1];
	if (i == 1)
	  {
	    seg2[0][n0].v1.x = seg2[0][1].v0.x;
	    seg2[0][n0].v1.y = seg2[0][1].v0.y;
	  }
	else
	  {
	    seg2[0][i - 1].v1.x = seg2[0][i].v0.x;
	    seg2[0][i - 1].v1.y = seg2[0][i].v0.y;
	  }       
	seg2[0][i].is_inserted = FALSE;
      }  
  } else { /* tp0->SplineReprn == 1 */
    n0 = tp0->TrimPortion[0].SplinePortion.NumOfCurves;
    for (k=1, i=0; i<n0; i++) {
      ic = &tp0->TrimPortion[0].SplinePortion.UV_Splines[i];
      xyzc = &tp0->TrimPortion[0].SplinePortion.XYZ_Splines[i];
      num_pts = NumPointsOnCurve(ic, 0.03);
      step_size = 1.0 / num_pts;
      for (j=0; j<num_pts; j++) {
	Eval3DCurve(xyzc, j*step_size, &n3dsegs0[k]);
	Eval2DCurve(ic, j*step_size, &pt);
	seg2[0][k].v0.x = pt.u;
	seg2[0][k].v0.y = pt.v;
	if (k != 1) {
	  seg2[0][k-1].v1.x = pt.u;
	  seg2[0][k-1].v1.y = pt.v;
	}
	seg2[0][k].is_inserted = FALSE;
	k++;
      }
    }
    n0 = k-1; /* use n0 itself to denote number of segments in seg2 */
    /* Now to fill seg2[0][n0].v1 with that of first point */
    seg2[0][n0].v1.x = seg2[0][1].v0.x;
    seg2[0][n0].v1.y = seg2[0][1].v0.y;
  }
    
  if (!tp1->SplineReprn) {
    n1  = tp1->TrimPortion[0].PiecePortion.NumOfVertices;
    for (i = 1; i <= n1; i++)
      {
	seg2[1][i].v0.x =  tp1->TrimPortion[0].PiecePortion.poly_vertex[i][0];
	seg2[1][i].v0.y =  tp1->TrimPortion[0].PiecePortion.poly_vertex[i][1];
	if (i == 1)
	  {
	    seg2[1][n1].v1.x = seg2[1][1].v0.x;
	    seg2[1][n1].v1.y = seg2[1][1].v0.y;
	  }
	else
	  {
	    seg2[1][i - 1].v1.x = seg2[1][i].v0.x;
	    seg2[1][i - 1].v1.y = seg2[1][i].v0.y;
	  }       
	seg2[1][i].is_inserted = FALSE;
      }  
  } else { /* tp1->SplineReprn == 1 */
    n1 = tp1->TrimPortion[0].SplinePortion.NumOfCurves;
    for (k=1, i=0; i<n1; i++) {
      ic = &tp1->TrimPortion[0].SplinePortion.UV_Splines[i];
      xyzc = &tp1->TrimPortion[0].SplinePortion.XYZ_Splines[i];
      num_pts = NumPointsOnCurve(ic, 0.03);
      step_size = 1.0 / num_pts;
      for (j=0; j<num_pts; j++) {
	Eval3DCurve(xyzc, j*step_size, &n3dsegs1[k]);
	Eval2DCurve(ic, j*step_size, &pt);
	seg2[1][k].v0.x = pt.u;
	seg2[1][k].v0.y = pt.v;
	if (k != 1) {
	  seg2[1][k-1].v1.x = pt.u;
	  seg2[1][k-1].v1.y = pt.v;
	}
	seg2[1][k].is_inserted = FALSE;
	k++;
      }
    }
    n1 = k-1; /* use n1 itself to denote number of segments in seg2 */
    /* Now to fill seg2[1][n1].v1 with that of first point */
    seg2[1][n1].v1.x = seg2[1][1].v0.x;
    seg2[1][n1].v1.y = seg2[1][1].v0.y;
  }

  nc = 0;
  p = (LIST_POINT4D *) incurve->Curve;
  while (p != NULL)
    {
      curve[nc++] = p->p4D;
      p = p->plp4DNext;
    }
  
  nportions = obtain_proper_trimcurve(&tp0->tp, &tp1->tp, seg2[0], n3dsegs0,
				      seg2[1], n3dsegs1, n0, n1, curve,
				      nc);
  verify_intersection_curves(nportions, &actual_nportions);
  
  if (actual_nportions == 0)
    {
      *outcurve = NULL;
      return 0;
    }

  *outcurve = (INT_CURVE_STRUCT *)
    new(actual_nportions*sizeof(INT_CURVE_STRUCT)); 
  
  k = 0;
  for (i = 0; i < nportions; i++)
    {
      if (portion[i].numpoints < 2)
	continue;

      q = (*outcurve)[k].Curve = (LIST_POINT4D *) new(sizeof(LIST_POINT4D));
      q->p4D = portion[i].pc[0];
      q->plp4DNext = NULL;
      (*outcurve)[k].endPoints[1] = q->p4D;
      
      for (j = 1; j < portion[i].numpoints; j++)
	{
	  p = (LIST_POINT4D *) new(sizeof(LIST_POINT4D));
	  p->p4D = portion[i].pc[j];
	  p->plp4DNext = NULL;
	  q->plp4DNext = p;
	  q = p;
	}
      
      (*outcurve)[k].endPoints[0] = q->p4D;
      k++;
    }
  
  return nportions;
}
      

int verify_intersection_curves(int nportions, int *actual_nportions)
{
  register int i;

  *actual_nportions = nportions;
  for (i = 0; i < nportions; i++)
    if (portion[i].numpoints < 2)
      (*actual_nportions)--;
}
#endif
